/*****************************************************************************\
* (c) Copyright 2000-2018 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/
#ifndef DAVINCIKERNEL_GUARDS_H
#define DAVINCIKERNEL_GUARDS_H 1

#include "GaudiKernel/DataObject.h"
#include "Kernel/Functions.h"

/** @namespace DaVinci::Utils Kernel/Guards.h
 *
 *  General helper functions and classes for use in DaVinci.
 *  These should only know about Gaudi classes and patterns.
 *  No knowledge of DaVinci specific code should be assumed.
 *
 *  @author Juan PALACIOS palacios@uzh.uzh.ch
 *  @date   2010-09-21
 */

namespace DaVinci {

  namespace Utils {

    /**
     *
     * Guard for container of newed pointers. Delete each item in a container of
     * newed pointers.
     *
     * T must be a const_iterator defined, begin(), end() and clear() methods,
     * and holding pointers.
     *
     * @author Juan Palacios juan.palacios@nikhef.nl
     * @date 31/08/2009
     *
     **/
    template <class T>
    class PointerContainerGuard {
    public:
      explicit PointerContainerGuard( T& container ) : m_container( container ) {}

      ~PointerContainerGuard() {
        for ( auto* obj : m_container ) delete obj;
        m_container.clear();
      }

    private:
      T& m_container;
    };

    /**
     *
     * Guard for container of newed pointers. Delete each item in a container of
     * newed pointers unless they are also in the TES.
     *
     * T must be a const_iterator defined, begin(), end() and clear() methods,
     * and holding pointers.
     *
     *
     * @author Juan Palacios juan.palacios@nikhef.nl
     * @date 01/09/2009
     *
     **/
    template <class T>
    class OrphanPointerContainerGuard {
    public:
      explicit OrphanPointerContainerGuard( T& container ) : m_container( container ) {}

      ~OrphanPointerContainerGuard() {
        for ( auto* obj : m_container ) {
          if ( !DaVinci::Utils::inTES( obj ) ) delete obj;
        }
        m_container.clear();
      }

    private:
      T& m_container;
    };

    /**
     * Call the container's clear() method when going out of scope.
     *
     * @author Juan Palacios palacios@physik.uzh.ch
     * @date 2011-01-07
     *
     **/
    template <class T>
    class ClearContainerGuard {
    public:
      explicit ClearContainerGuard( T& container ) : m_container( container ) {}

      ~ClearContainerGuard() { m_container.clear(); }

    private:
      T& m_container;
    };

  } // namespace Utils

} // namespace DaVinci

#endif // KERNEL_DAVINCIGUARDS_H
