/*****************************************************************************\
* (c) Copyright 2000-2018 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/
// ============================================================================
// STD
// ============================================================================
#include <algorithm>
// ============================================================================
// local
// ============================================================================
#include "MessagingBase.h"
// ============================================================================
// GaudiKernel
// ============================================================================
#include "GaudiKernel/IAlgorithm.h"
#include "GaudiKernel/SmartIF.h"
// ============================================================================
/** @file
 *  Implementation file for class : MessagingBase
 *  @see LoKi::MessagingBase
 *  @date 2014-10-21
 *  @author Chris Jones
 */
// ============================================================================
// Standard constructor, initializes variables
// ============================================================================
LoKi::MessagingBase::MessagingBase( const std::string& type, const std::string& name, const IInterface* parent )
    : GaudiTool( type, name, parent ) {
  //
  declareProperty( "PrintMyAlg", m_printMyAlg, "Print the name of ``associated'' algorithm" );
  declareProperty( "MaxPrints", m_prints, "Maximum number of prints" );
  declareProperty( "MeasureCPUPerformance", m_timing, "Measure CPU perormance" );
  //
  if ( parent ) {
    SmartIF<IAlgorithm> alg( const_cast<IInterface*>( parent ) );
    if ( alg.isValid() ) { m_myAlg = alg->name(); }
  }
}
// ============================================================================
// Initialize
// ============================================================================
StatusCode LoKi::MessagingBase::initialize() {
  /// initialize the base
  const StatusCode sc = GaudiTool::initialize();
  if ( sc.isFailure() ) { return sc; }
  //
  if ( msgLevel( MSG::DEBUG ) && 0 == m_prints ) {
    m_prints = 10;
    warning() << "Redefine 'MaxPrints' property to " << m_prints << endmsg;
  }
  //
  if ( parent() ) {
    SmartIF<IAlgorithm> alg( const_cast<IInterface*>( parent() ) );
    if ( alg.isValid() ) { m_myAlg = alg->name(); }
  }
  //
  // is message level is low enough, enforce CPU performance measurements
  if ( msgLevel( MSG::DEBUG ) ) { m_timing = true; }
  //
  return sc;
}
// ============================================================================

// ============================================================================
// get the correct algorithm context
// ============================================================================
bool LoKi::MessagingBase::getMyAlg() const {
  m_myAlg.clear();
  const IAlgContextSvc* asvc = contextSvc();
  if ( !asvc ) { return false; }
  const IAlgorithm* current = asvc->currentAlg();
  if ( !current ) { return false; }
  m_myAlg = " [" + current->name() + "]";
  return true;
}

// ============================================================================
// The END
// ============================================================================
