/*****************************************************************************\
* (c) Copyright 2000-2018 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/

#include <algorithm>
#include <memory>
#include <sstream>

#include "Event/Particle.h"
#include "Event/Vertex.h"

#include "Kernel/IParticlePropertySvc.h"
#include "Kernel/IParticleTransporter.h"
#include "Kernel/IVertexFit.h"
#include "Kernel/ParticleProperty.h"

#include "LHCbMath/Kinematics.h"
#include "LHCbMath/LHCbMath.h"
#include "LHCbMath/MatrixTransforms.h"
#include "LHCbMath/MatrixUtils.h"

#include "LoKi/Child.h"
#include "LoKi/DecayChain.h"
#include "LoKi/Trees.h"

#include "Math/Functions.h"

#include "KalmanFilter/VertexFit.h"

#include "CaloFutureUtils/CaloFutureParticle.h"
#include "CaloFutureUtils/CaloMomentum.h"

#include "VertexFitter.h"

/** @file
 *
 *  This file is a part of
 *  <a href="http://cern.ch/lhcb-comp/Analysis/LoKi/index.html">LoKi project:</a>
 *  ``C++ ToolKit for Smart and Friendly Physics Analysis''
 *
 *  This file is a part of LoKi project -
 *    "C++ ToolKit  for Smart and Friendly Physics Analysis"
 *
 *  The package has been designed with the kind help from
 *  Galina PAKHLOVA and Sergey BARSUK.  Many bright ideas,
 *  contributions and advices from G.Raven, J.van Tilburg,
 *  A.Golutvin, P.Koppenburg have been used in the design.
 *
 *  @author Vanya BELYAEV ibelyaev@physics.syr.edu
 *  @date 2006-05-24
 */

/// anonymous namespace to hide few technical constants
namespace {

  const double s_scale   = Gaudi::Units::perCent;
  const double s_scale2  = s_scale * s_scale;
  const double s_small   = 0.1 * Gaudi::Units::micrometer;
  const double s_small2  = s_small * s_small;
  const double s_middle  = 1 * Gaudi::Units::centimeter;
  const double s_middle2 = s_middle * s_middle;
  const double s_large   = 10 * Gaudi::Units::centimeter;
  const double s_large2  = s_large * s_large;

  /** @var s_equal
   *  equality comparison
   *  @see LHCb::Math::Equal_To
   */
  const LHCb::Math::Equal_To<double> s_equal; // equality comparison

} // namespace

// ============================================================================
// load the data from the daughter particles into the internal structures
// ============================================================================
StatusCode LoKi::VertexFitter::_load( const LHCb::Particle::ConstVector& ds, FitData& data ) const {
  StatusCode sc = StatusCode::SUCCESS;
  data.m_entries.clear();
  data.m_entries.reserve( ds.size() );
  //
  for ( LHCb::Particle::ConstVector::const_iterator c = ds.begin(); ds.end() != c && sc.isSuccess(); ++c ) {
    sc = _load( *c, data );
  }
  //
  if ( sc.isFailure() ) {
    ++m_load_error;
    return sc;
  } // RETURN
  if ( data.m_entries.empty() ) {
    ++m_load_no_valid_data;
    return StatusCode{ InvalidData };
  } // RETURN
  //
  if ( !LoKi::KalmanFilter::okForVertex( data.m_entries ) &&
       ( !m_use_rho_like_branch || !LoKi::KalmanFilter::rhoPlusLike( data.m_entries ) ) ) {
    ++m_not_vertexed;
    if ( msgLevel( MSG::DEBUG ) ) {
      LoKi::DecayChain printer;
      error() << "Problem to vertex:" << endmsg << printer.print_( ds ) << endmsg;
    }
    return StatusCode{ InvalidData };
  }
  //
  return StatusCode::SUCCESS;
}
// ============================================================================
// load the data into internal representation
// ============================================================================
StatusCode LoKi::VertexFitter::_load( const LHCb::Particle* particle, FitData& data ) const {
  //
  if ( !particle ) {
    ++m_load_no_particle;
    return StatusCode{ InvalidParticle };
  } // RETURN
  //
  const LoKi::KalmanFilter::ParticleType t = particleType_( particle );
  // regular particles
  if ( LoKi::KalmanFilter::JetLikeParticle != t ) {
    data.m_entries.push_back( Entry() );
    StatusCode sc = LoKi::KalmanFilter::load( *particle, t, data.m_entries.back() );
    if ( sc.isFailure() ) {
      std::ostringstream ostr;
      ostr << " Failure to load particle ID " << particle->particleID().abspid() << " as particle type " << t;
      return Error( ostr.str(), sc ); // RETURN
    }
    return sc; // RETURN
  }
  // for jets, load all components...
  typedef SmartRefVector<LHCb::Particle> DAUGHTERS;
  const DAUGHTERS&                       daughters = particle->daughters();
  for ( DAUGHTERS::const_iterator idau = daughters.begin(); daughters.end() != idau; ++idau ) {
    StatusCode sc = _load( *idau, data );
    if ( sc.isFailure() ) { return sc; } // RETURN
  }
  //
  return StatusCode::SUCCESS;
}
// ============================================================================
// add one particle at the end of the queue
// ============================================================================
StatusCode LoKi::VertexFitter::_add( LHCb::Particle const* child, Gaudi::XYZPoint const& point,
                                     IGeometryInfo const& geometry, FitData& data ) const {
  StatusCode sc = _load( child, data );
  if ( sc.isFailure() ) ++m_add_failure_ignore;
  sc = _transport( data.m_entries.back(), point, geometry );
  if ( sc.isFailure() ) ++m_add_transport_ignore;
  return StatusCode::SUCCESS;
}
// ============================================================================
// stopping criteria for iterations
// ============================================================================
bool LoKi::VertexFitter::stop_iter( const Gaudi::Vector3& dx, const Gaudi::SymMatrix3x3& ci, const size_t iter,
                                    FitData& data ) const {
  // distance in the absolute position
  const double d1 = ROOT::Math::Mag( dx );
  // distance in the chi2 units
  double d2 = ROOT::Math::Similarity( dx, ci );
  //
  if ( d2 < 0 ) {
    if ( std::abs( d2 ) < std::min( m_DistanceChi2, 0.01 ) || s_equal( d2, 0 ) ) {
      d2 = 0;
      ++m_stop_iter_sn_chi2;
    } else {
      ++m_stop_iter_ln_chi2;
    }
  }
  //
  // termination conditions:
  //
  //  (1) STOP if the distance is sufficiently small
  //       - either the absolute distance
  //       - or chi2 distance (if at least one iteration is performed)
  //
  const bool ok1 = d1 < m_DistanceMax;
  const bool ok2 = 1 < iter && 0 <= d2 && d2 < m_DistanceChi2;
  if ( ok1 || ok2 ) {
    //
    // kalman smooth : could it be moved just before evalCov ?
    // @todo study the possibility to move smoothig out of iteration loop
    StatusCode sc = LoKi::KalmanFilter::smooth( data.m_entries );
    if ( sc.isFailure() ) {
      ++m_stop_iter_smoother;
      return false; // make more iterations ..
    }
    //
    sc = LoKi::KalmanFilter::evalCov( data.m_entries );
    if ( sc.isFailure() ) {
      ++m_stop_iter_cov;
      return false; // make more iterations ..
    }
    //
    if ( ok1 ) { m_nIterations1 += iter; }
    if ( ok2 ) { m_nIterations2 += iter; }
    //
    return true;
  }
  //
  return false;
}
// ============================================================================

// ============================================================================
// make optimised Kalman filter iterations
// ============================================================================
StatusCode LoKi::VertexFitter::_iterate_opt( size_t nMax, IGeometryInfo const& geometry, FitData& data ) const {
  // timing measurements
  Chrono chrono( timing() ? chronoSvc() : nullptr, timing() ? name() + ":optimized" : std::string() );
  //
  Gaudi::Vector3 x0;
  for ( size_t iIter = 0; iIter <= nMax + 1; ++iIter ) {
    //
    StatusCode sc = LoKi::KalmanFilter::steps( data.m_entries, 0 );
    if ( sc.isFailure() ) ++m_iterate_steps;
    //
    const Gaudi::Vector3&      x  = data.m_entries[0].m_x;
    const Gaudi::SymMatrix3x3& ci = data.m_entries[0].m_ci;
    //
    if ( sc.isSuccess() && 1 <= iIter && stop_iter( x - x0, ci, iIter, data ) ) {
      m_nIterationsOpt += ( iIter - 1 ); // the first one acts as seeding
      return StatusCode::SUCCESS;
    }
    //
    Gaudi::XYZPoint point;
    Gaudi::Math::la2geo( x, point );
    sc = _transport( point, geometry, data );
    if ( sc.isFailure() ) ++m_iterate_transport;
    //
    x0 = x;
  }
  //
  ++m_iterate_no_convergence;
  return StatusCode{ NoConvergency }; // RETURN
}
// ============================================================================
// make few kalman iterations
// ============================================================================
StatusCode LoKi::VertexFitter::_iterate( size_t nIterMax, Gaudi::Vector3 const& _x, IGeometryInfo const& geometry,
                                         FitData& data ) const {
  // timing measurements
  Chrono chrono( timing() ? chronoSvc() : nullptr, timing() ? name() + ":iterate" : std::string() );
  //
  // initial position
  const Gaudi::Vector3* x = &_x;
  // chi2
  const double  _chi2 = 0;
  const double* chi2  = &_chi2;
  // inverse covariance matrix for the position
  const Gaudi::SymMatrix3x3* ci = &data.m_seedci;
  //
  const bool rho_like = m_use_rho_like_branch && LoKi::KalmanFilter::rhoPlusLike( data.m_entries );
  //
  for ( size_t iIter = 1; iIter <= nIterMax; ++iIter ) {
    // make a proper transportation
    Gaudi::XYZPoint point;
    Gaudi::Math::la2geo( *x, point );
    StatusCode sc = _transport( point, geometry, data );
    if ( sc.isFailure() ) ++m_iterate_transport2;
    // initialize the covariance matrix
    if ( 0 != iIter ) { data.m_seedci = ( *ci ) * s_scale2; }
    ci = &data.m_seedci;
    // initialize the chi2
    chi2 = &_chi2;
    const Gaudi::Vector3 x0( *x );
    // D) use ``rho+-Like'' branch
    if ( rho_like ) {
      sc = LoKi::KalmanFilter::stepRho( data.m_entries, *chi2 );
      if ( sc.isFailure() ) {
        ++m_rho_kalman;
      } else {
        // update the parameters
        const LoKi::KalmanFilter::Entry& last = data.m_entries.back();
        ci                                    = &last.m_ci;
        x                                     = &last.m_x;
        chi2                                  = &last.m_chi2;
      }
    }
    //
    // D) general case (or failure specialization )
    //
    if ( !rho_like || sc.isFailure() ) {
      for ( EIT entry = data.m_entries.begin(); data.m_entries.end() != entry; ++entry ) {
        // make one Kalman step
        sc = LoKi::KalmanFilter::step( *entry, *x, *ci, *chi2 );
        // skip on Failure
        if ( sc.isFailure() ) {
          ++m_kalman;
          continue; // CONTINUE
        }
        // update the parameters
        ci   = &entry->m_ci;
        x    = &entry->m_x;
        chi2 = &entry->m_chi2;
      }
    }
    //
    if ( sc.isSuccess() && stop_iter( *x - x0, *ci, iIter, data ) ) {
      m_nbIterGenCounter += iIter;
      return StatusCode::SUCCESS;
    }
    //
  } // end of iterations
  //
  ++m_no_convergence;
  return StatusCode{ NoConvergency }; // RETURN
}
// ============================================================================
// make few kalman iterations
// ============================================================================
StatusCode LoKi::VertexFitter::_iterate_rho( size_t nIterMax, IGeometryInfo const& geometry, FitData& data ) const {
  // timing measurements
  Chrono chrono( timing() ? chronoSvc() : nullptr, timing() ? name() + ":rho" : std::string() );
  //
  // initial position
  const Gaudi::Vector3*      x  = 0;
  const Gaudi::SymMatrix3x3* ci = 0;
  //
  Entries::const_iterator igood = data.m_entries.begin();
  for ( Entries::const_iterator ientry = data.m_entries.begin(); data.m_entries.end() != ientry; ++ientry ) {
    if ( LoKi::KalmanFilter::LongLivedParticle == ientry->m_type ||
         LoKi::KalmanFilter::ShortLivedParticle == ientry->m_type ) {
      igood = ientry;
      break;
    }
  }
  //
  x = &igood->m_x;
  // iterate
  for ( size_t iIter = 1; iIter <= nIterMax; ++iIter ) {
    // make a proper transportation
    Gaudi::XYZPoint point;
    Gaudi::Math::la2geo( *x, point );
    StatusCode sc = _transport( point, geometry, data );
    if ( sc.isFailure() ) ++m_rho_iterate_transport;
    //
    const Gaudi::Vector3 x0( *x );
    //
    // use ``rho+-Like'' branch
    sc = LoKi::KalmanFilter::stepRho( data.m_entries, 0 );
    if ( sc.isFailure() ) {
      ++m_rho_kalman;
    } else {
      // update the parameters
      const LoKi::KalmanFilter::Entry& last = data.m_entries.back();
      ci                                    = &last.m_ci;
      x                                     = &last.m_x;
    }
    //
    if ( sc.isSuccess() && 0 != ci && stop_iter( *x - x0, *ci, iIter, data ) ) {
      m_nbIterRhoCounter += iIter;
      return StatusCode::SUCCESS;
    }
    //
  } // end of iterations
  //
  ++m_no_convergency;
  return StatusCode{ NoConvergency }; // RETURN
}
// ============================================================================
// make a seed
// ============================================================================
StatusCode LoKi::VertexFitter::_seed( LHCb::Vertex const* vertex, IGeometryInfo const& geometry, FitData& data ) const {
  // timing measurements
  Chrono chrono( timing() ? chronoSvc() : nullptr, timing() ? name() + ":seed" : std::string() );
  //
  // check if vertex could be used as a seed
  const Gaudi::XYZPoint&     p = vertex->position();
  const Gaudi::SymMatrix3x3& c = vertex->covMatrix();

  //
  if ( seedOK( p ) && 0 < c( 0, 0 ) && c( 0, 0 ) < 1 * Gaudi::Units::cm2 && 0 < c( 1, 1 ) &&
       c( 1, 1 ) < 1 * Gaudi::Units::cm2 && 0 < c( 2, 2 ) && c( 2, 2 ) < 100 * Gaudi::Units::cm2 &&
       s_small2 < Gaudi::Math::min_diagonal( c ) && s_large2 > Gaudi::Math::max_diagonal( c ) ) {
    // use the vertex parameters as the proper seed
    Gaudi::Math::geo2LA( p, data.m_seed );
    int ifail     = 0;
    data.m_seedci = c.Inverse( ifail );
    if ( !ifail ) {
      // properly scale the seed matrix
      Gaudi::Math::scale( data.m_seedci, s_scale2 );
      ++m_nbSeedCase0Counter;
      return StatusCode::SUCCESS; // RETURN
    }
  }
  //
  if ( m_use_shortlived_as_seed ) {
    for ( EIT entry = data.m_entries.begin(); data.m_entries.end() != entry; ++entry ) {
      if ( LoKi::KalmanFilter::ShortLivedParticle == entry->m_type ) {
        data.m_seed   = entry->m_parx;
        data.m_seedci = entry->m_vxi;
        Gaudi::Math::scale( data.m_seedci, s_scale2 );
        ++m_nbSeedCase2Counter;
        return StatusCode::SUCCESS; // RETURN
      }
    }
  }
  //
  StatusCode sc      = StatusCode::FAILURE;
  const bool massage = forMassage( data.m_entries );
  //
  // the simple case with no-massage
  if ( !massage && okForVertex( data.m_entries ) ) {
    sc = LoKi::KalmanFilter::steps( data.m_entries, 0 );
    if ( sc.isSuccess() ) {
      data.m_seed   = data.m_entries[0].m_x;
      data.m_seedci = data.m_entries[0].m_ci;
      Gaudi::Math::scale( data.m_seedci, s_scale2 );
      ++m_nbSeedCase3Counter;
      return StatusCode::SUCCESS; // RETURN
    }
  }
  //
  // specific cases: seeds for 2,3,4-body decays
  //
  if ( 2 == data.m_entries.size() && LoKi::KalmanFilter::twoProngs( data.m_entries ) ) {
    sc = LoKi::KalmanFilter::step( data.m_entries[0], data.m_entries[1], 0 );
  } else if ( 3 == data.m_entries.size() && LoKi::KalmanFilter::threeProngs( data.m_entries ) ) {
    sc = LoKi::KalmanFilter::step( data.m_entries[0], data.m_entries[1], data.m_entries[2], 0 );
  } else if ( 4 == data.m_entries.size() && LoKi::KalmanFilter::fourProngs( data.m_entries ) ) {
    sc = LoKi::KalmanFilter::step( data.m_entries[0], data.m_entries[1], data.m_entries[2], data.m_entries[3], 0 );
  }
  //
  if ( sc.isSuccess() ) {
    data.m_seed   = data.m_entries[0].m_x;
    data.m_seedci = data.m_entries[0].m_ci;
    Gaudi::Math::scale( data.m_seedci, s_scale2 );
    ++m_nbSeedCase1Counter;
    return StatusCode::SUCCESS; // RETURN
  }
  //
  // simple case: make seed ignoring massaged stuff
  const Entries* entries = 0;
  //
  Entries e_good;
  if ( massage ) {
    e_good.reserve( data.m_entries.size() );
    // pick up good entries:
    for ( EIT entry = data.m_entries.begin(); data.m_entries.end() != entry; ++entry ) {
      if ( entry->regular() && !forMassage( entry->m_p0 ) ) { e_good.push_back( *entry ); }
    }
    //
    entries = &e_good;
  } else {
    entries = &data.m_entries;
  }
  //
  // relatively obvious case:
  //
  if ( massage && LoKi::KalmanFilter::okForVertex( *entries ) ) {
    // estimate "seed" using only good components
    sc = LoKi::KalmanFilter::seed( *entries, data.m_seed, data.m_seedci, s_scale2 );
    if ( sc.isSuccess() && seedOK( data.m_seed ) ) {
      sc = _transport( data.m_seed, geometry, data );
      if ( sc.isSuccess() ) {
        // adjust seed using all entries
        sc = LoKi::KalmanFilter::seed( data.m_entries, data.m_seed, data.m_seedci, s_scale2 );
        if ( sc.isSuccess() && seedOK( data.m_seed ) ) {
          ++m_nbSeedCase4Counter;
          return StatusCode::SUCCESS;
        }
      }
    }
  }
  //
  // less trivial case: perform some "massage" action
  //
  if ( massage && !m_massage.empty() ) {
    //
    Gaudi::Vector3      p_seed;
    Gaudi::SymMatrix3x3 ci_seed;
    unsigned short      n_seed = 0;
    //
    for ( const auto& z : m_massage ) {
      //
      if ( z > m_seedZmax || z < m_seedZmin ) { continue; }
      // transport all? - think a bit more...
      StatusCode sc = _transport( Gaudi::XYZPoint( 0, 0, z ), geometry, data );
      if ( sc.isFailure() ) { continue; }
      sc = LoKi::KalmanFilter::seed( data.m_entries, data.m_seed, data.m_seedci, s_scale2 );
      if ( sc.isFailure() ) { continue; }
      if ( !seedOK( data.m_seed ) ) { continue; }
      //
      ci_seed += data.m_seedci;
      p_seed += data.m_seedci * data.m_seed;
      ++n_seed;
      //
    }
    //
    m_nbSeedMassageCounter += n_seed;
    //
    int                 ifail = 0;
    Gaudi::SymMatrix3x3 c     = ci_seed.Inverse( ifail );
    if ( 0 == ifail ) {
      data.m_seed = c * p_seed;
      if ( seedOK( data.m_seed ) ) {
        // transport all ?
        StatusCode sc = _transport( data.m_seed, geometry, data );
        if ( sc.isSuccess() ) {
          // make proper seed from all components
          sc = LoKi::KalmanFilter::seed( data.m_entries, data.m_seed, data.m_seedci, s_scale2 );
          if ( sc.isSuccess() && seedOK( data.m_seed ) ) {
            ++m_nbSeedCase5Counter;
            return StatusCode::SUCCESS;
          }
        }
      }
    }
  }
  //
  // the basic case
  //
  sc = LoKi::KalmanFilter::seed( data.m_entries, data.m_seed, data.m_seedci, s_scale2 );
  if ( sc.isSuccess() && seedOK( data.m_seed ) ) {
    ++m_nbSeedCase6Counter;
    return StatusCode::SUCCESS;
  }
  //
  data.m_seed[0] = 0;
  data.m_seed[1] = 0;
  data.m_seed[2] = 100 * Gaudi::Units::mm;
  //
  Gaudi::Math::setToUnit( data.m_seedci, 1.0 / s_middle2 );
  data.m_seedci( 2, 2 ) = 1.0 / s_large2;
  //
  ++m_nbSeedCase7Counter;
  //
  ++m_no_proper_seed;
  return StatusCode::SUCCESS;
}
// ============================================================================
// The vertex fitting method without creation of a Particle
// ============================================================================
StatusCode LoKi::VertexFitter::fit( LHCb::Vertex& vertex, LHCb::Particle::ConstVector const& daughters,
                                    IGeometryInfo const& geometry ) const {
  FitData data;
  return _fit( vertex, daughters, geometry, data );
}

StatusCode LoKi::VertexFitter::_fit( LHCb::Vertex& vertex, LHCb::Particle::ConstVector const& daughters,
                                     IGeometryInfo const& geometry, FitData& data ) const {
  // timing measurements
  Chrono chrono( timing() ? chronoSvc() : nullptr, timing() ? name() + ":fit" : std::string() );

  //
  // load the data
  StatusCode sc = _load( daughters, data );
  if ( sc.isFailure() ) {
    ++m_fit_load;
    return sc;
  } // RETURN
  //
  bool special = false;
  // special case:  use optimized branch when possible
  if ( m_use_optimized && okForVertex( data.m_entries ) && !forMassage( data.m_entries ) ) {
    sc      = _iterate_opt( m_nIterMaxI, geometry, data );
    special = true;
  }
  // special case: rho+-like branch
  else if ( m_use_rho_like_branch && LoKi::KalmanFilter::rhoPlusLike( data.m_entries ) ) {
    sc      = _iterate_rho( m_nIterMaxI, geometry, data );
    special = true;
  }
  //
  // regular case
  if ( !special || sc.isFailure() ) {
    /// make a seed
    sc = _seed( &vertex, geometry, data );
    if ( sc.isFailure() ) ++m_fit_seed_failure;
    // make "m_nIterMax" iterations
    sc = _iterate( m_nIterMaxI, data.m_seed, geometry, data );
    if ( sc.isFailure() ) {
      ++m_fit_failure_iterator;
      return sc;
    } // RETURN
  }
  // get the data from filter
  const Entry&               entry = data.m_entries.back();
  const Gaudi::Vector3&      x     = entry.m_x;
  const Gaudi::SymMatrix3x3& c     = entry.m_c;
  // const double               chi2  = entry.m_chi2      ;
  // update the vertex parameters:
  vertex.setPosition( Gaudi::XYZPoint( x( 0 ), x( 1 ), x( 2 ) ) );
  vertex.setCovMatrix( c );
  //
  double chi2 = LoKi::KalmanFilter::chi2( data.m_entries );
  //
  if ( chi2 < 0 ) {
    if ( std::abs( chi2 ) < std::min( m_DistanceChi2, 0.01 ) || s_equal( chi2, 0 ) ) {
      chi2 = 0;
      ++m_fit_small_negative_chi2;
    } else {
      ++m_fit_large_negative_chi2;
    }
  }
  //
  vertex.setChi2( chi2 );
  vertex.setNDoF( LoKi::KalmanFilter::nDoF( data.m_entries ) );
  vertex.setTechnique( LHCb::Vertex::VertexFitter );
  // fill the vertex
  vertex.clearOutgoingParticles();
  for ( Entries::const_iterator ientry = data.m_entries.begin(); data.m_entries.end() != ientry; ++ientry ) {
    if ( ientry->regular() ) { vertex.addToOutgoingParticles( ientry->m_p0 ); }
  }
  // keep for future tracing
  data.m_vertex = &vertex;
  if ( !seedOK( vertex.position() ) ) ++m_fit_outside_fiducial;
  return sc;
}
// ============================================================================
// The vertex fitting method with the creation of a Particle
// ============================================================================
StatusCode LoKi::VertexFitter::fit( LHCb::Particle::ConstVector const& daughters, LHCb::Vertex& vertex,
                                    LHCb::Particle& particle, IGeometryInfo const& geometry ) const {
  FitData data;
  return _fit( daughters, vertex, particle, geometry, data );
}
StatusCode LoKi::VertexFitter::_fit( LHCb::Particle::ConstVector const& daughters, LHCb::Vertex& vertex,
                                     LHCb::Particle& particle, IGeometryInfo const& geometry, FitData& data ) const {
  //
  using namespace ROOT::Math;
  //
  // play a bit with extra-info
  if ( particle.hasInfo( LHCb::Particle::Chi2OfVertexConstrainedFit ) ) {
    particle.eraseInfo( LHCb::Particle::Chi2OfVertexConstrainedFit );
  }

  // make a vertex fit
  StatusCode sc = _fit( vertex, daughters, geometry, data );
  if ( sc.isFailure() ) {
    ++m_fit_failure;
    return sc;
  }
  // links:

  particle.clearDaughters();
  for ( const auto& dau : daughters ) particle.addToDaughters( dau );
  //
  particle.setEndVertex( &vertex );
  //
  // fill the particle with the proper information
  // 1) reference point and proper covariance matrices
  particle.setReferencePoint( vertex.position() );
  particle.setPosCovMatrix( vertex.covMatrix() );
  // 2) momentum, its covariance matrix and covariance with position
  // reset existing matrices
  Gaudi::Math::setToScalar( data.m_cmom, 0.0 );
  Gaudi::Math::setToScalar( data.m_cmom1, 0.0 );
  Gaudi::Math::setToScalar( data.m_mpcov, 0.0 );
  Gaudi::Math::setToScalar( data.m_mm_c, 0.0 );
  //
  // 1. for true kinematics:
  Gaudi::LorentzVector       vct;
  const Gaudi::SymMatrix3x3& pos = particle.posCovMatrix();
  // 2. for "measured-mass":
  Gaudi::LorentzVector mm_v;

  for ( EIT i = data.m_entries.begin(); data.m_entries.end() != i; ++i ) {
    //
    Gaudi::Math::add( vct, i->m_q );
    data.m_cmom += i->m_d;
    data.m_mpcov += i->m_e;
    //
    // for measured mass: blind sum of 4-momenta for extrapolated daughters
    //
    mm_v += i->m_p.momentum();
    data.m_mm_c += i->m_p.momCovMatrix();
    //
    if ( i->special() ) { continue; } // gamma & digamma
    //
    for ( EIT j = i + 1; data.m_entries.end() != j; ++j ) {
      data.m_cmom1 += i->m_f * pos * Transpose( j->m_f );
      data.m_cmom1 += j->m_f * pos * Transpose( i->m_f );
    }
  }
  Gaudi::Math::add( data.m_cmom, data.m_cmom1 );
  particle.setMomentum( vct );
  particle.setMomCovMatrix( data.m_cmom );
  particle.setPosMomCovMatrix( data.m_mpcov );
  //
  // measured mass & error in measured mass
  //
  double mmerr = Gaudi::Math::sigmamass( mm_v, data.m_mm_c );
  if ( 0 >= mmerr ) {
    if ( mmerr > -0.1 * Gaudi::Units::MeV || s_equal( mmerr, 0 ) ) {
      mmerr = Gaudi::Units::MeV;
      ++m_fit_mmas_e_bit_np;
    } else if ( mmerr > -1.0 * Gaudi::Units::MeV ) {
      mmerr = Gaudi::Units::MeV;
      ++m_fit_mmas_e_slight_np;
    } else {
      m_badMassError += mmerr;
      mmerr = Gaudi::Units::MeV;
      ++m_fit_mmas_e_np;
    }
  }
  //
  double mmass = mm_v.M();
  if ( 0 > mmass ) {
    if ( mmass > -0.1 * Gaudi::Units::MeV || s_equal( mmerr, 0 ) ) {
      mmass = Gaudi::Units::MeV;
      ++m_fit_mmas_bit_np;
    } else if ( mmass > -0.5 * mmerr ) {
      mmass = Gaudi::Units::MeV;
      ++m_fit_mmas_slight_np;
    } else {
      m_badMass += mmass;
      mmass = Gaudi::Units::MeV;
      ++m_fit_mmas_np;
    }
  }
  //
  particle.setMeasuredMass( mmass );
  particle.setMeasuredMassErr( mmerr );
  //
  // play a bit with extra-info:
  particle.addInfo( LHCb::Particle::Chi2OfVertexConstrainedFit, vertex.chi2() );
  //
  return StatusCode::SUCCESS;
}
// ============================================================================
// add the particle to the vertex and refit
// ============================================================================
StatusCode LoKi::VertexFitter::add( LHCb::Particle const* particle, LHCb::Vertex& vertex,
                                    IGeometryInfo const& geometry ) const {
  if ( !particle ) {
    ++m_add_no_particle;
    return StatusCode{ InvalidParticle };
  }
  //
  switch ( particleType_( particle ) ) {
  case LoKi::KalmanFilter::GammaLikeParticle:
    ++m_add_no_gamma;
    return StatusCode{ InvalidParticle };
  case LoKi::KalmanFilter::DiGammaLikeParticle:
    ++m_add_no_digamma;
    return StatusCode{ InvalidParticle };
  case LoKi::KalmanFilter::MergedPi0LikeParticle:
    ++m_add_no_mergedpi0;
    return StatusCode{ InvalidParticle };
  default:
    break;
  }

  FitData data;
  //
  // first need to fit it!
  // FIXME: As this class does not save state, we assume the vertex has not been fit before.
  // If this assumption causes performance problems, can be solved by following suggestion here:
  // https://gitlab.cern.ch/lhcb/Rec/-/merge_requests/2803#note_5391100
  // i.e. store FitData after fit() call, retrieve it here and check if the passed vertex has been
  // fit in the past.
  StatusCode sc =
      _fit( vertex, LHCb::Particle::ConstVector( vertex.outgoingParticles().begin(), vertex.outgoingParticles().end() ),
            geometry, data );
  if ( sc.isFailure() ) {
    ++m_add_fit;
    return sc;
  }

  sc = _add( particle, vertex.position(), geometry, data );
  if ( sc.isFailure() ) ++m_add_failure;
  //
  Entry&       entry = data.m_entries.back();
  const Entry& e0    = *( data.m_entries.end() - 2 );
  // make one Kalman step using the previos parameters as estimate
  sc = LoKi::KalmanFilter::step( entry, e0.m_x, e0.m_ci, e0.m_chi2 );
  if ( sc.isFailure() ) ++m_add_step;
  // smooth it!
  sc = LoKi::KalmanFilter::smooth( data.m_entries );
  if ( sc.isFailure() ) ++m_add_smooth;
  // make few more full iterations
  data.m_seedci = entry.m_ci;
  Gaudi::Math::scale( data.m_seedci, s_scale2 );
  sc = _iterate( m_nIterMaxII, entry.m_x, geometry, data );
  if ( sc.isFailure() ) ++m_add_iterate;
  //
  const Gaudi::Vector3&      x = entry.m_x;
  const Gaudi::SymMatrix3x3& c = entry.m_c;
  // const double               chi2  = entry.m_chi2      ;
  // update the vertex parameters:
  vertex.setPosition( Gaudi::XYZPoint( x( 0 ), x( 1 ), x( 2 ) ) );
  vertex.setCovMatrix( c );
  vertex.setChi2( LoKi::KalmanFilter::chi2( data.m_entries ) );
  vertex.setNDoF( LoKi::KalmanFilter::nDoF( data.m_entries ) );
  // fill the vertex
  vertex.addToOutgoingParticles( particle );
  // keep for future tracing
  data.m_vertex = &vertex;
  // check for positions
  if ( !seedOK( vertex.position() ) ) ++m_fit_outside_fiducial;
  //
  return StatusCode::SUCCESS;
}
// ============================================================================
// remove the particle from  the vertex and refit
// ============================================================================
StatusCode LoKi::VertexFitter::remove( [[maybe_unused]] LHCb::Particle const* particle,
                                       [[maybe_unused]] LHCb::Vertex&         vertex,
                                       [[maybe_unused]] IGeometryInfo const&  geometry ) const {
  ++m_no_remove;
  return StatusCode{ NotImplemented }; // ATTENTION!!!
  //
  // const StatusCode OK = StatusCode::SUCCESS;
  // if ( !particle ) { return Warning( "remove: nothing to remove", OK ); } // RETURN
  // if ( &vertex != data.m_vertex ) {
  //   // first need to fit it !
  //   const IVertexFit* vFit = this;
  //   StatusCode sc = vFit->fit( vertex, vertex.outgoingParticles().begin(), vertex.outgoingParticles().end(), geometry
  //   ); if ( sc.isFailure() ) { return Warning( "remove: error from 'fit'", sc ); } // RETURN
  // }
  // // remove the particle from the vertex
  // vertex.removeFromOutgoingParticles( particle );
  // find the corresponding entry

  //   EIT ifind = data.m_entries.end() ;
  //   for ( EIT entry = data.m_entries.begin() ; data.m_entries.end() != entry ; ++entry )
  //   { if ( particle == entry->m_p0 ) { ifind = entry ; break ; } }
  //   if ( data.m_entries.end() == ifind )
  //   { return Warning ("remove: particle is not in the fit", OK ) ; } // RETURN

  //   const Entry& back = data.m_entries.back() ;
  //   // FIX ME!!!
  //   const Gaudi::SymMatrix3x3 nci= ( back.m_ci  /* + ... FIX ME !!!! */ ) ;
  //   int iFail = 0 ;
  //   const Gaudi::SymMatrix3x3 nc = nci.Sinverse( iFail ) ;
  //   if ( 0 != iFail ) { iFail  = 0 ; Error ("remove: invaid matrix inversion") ;}
  //   // FIX ME!!!
  //   const Gaudi::Vector3 x     = nc*back.m_ci*back.m_x /** - ... FIX ME */ ;
  //   // FIX ME !!!
  //   const double         nchi2 = back.m_chi2 /** - ... FIX ME!!! */  ;

  //   // remove the entry
  //   data.m_entries.erase ( ifind ) ;

  //   // make few more full iterations
  //   StatusCode sc  = _iterate ( m_nIterMaxIII , x ) ;

  // return StatusCode::SUCCESS;
}
// ============================================================================
// standard constructor
// ============================================================================
LoKi::VertexFitter::VertexFitter( const std::string& type, const std::string& name, const IInterface* parent )
    : base_class( type, name, parent )
    /// maximal number of iteration for vertex fit
    , m_nIterMaxI( 10 )  // maximal number of iteration for vertex fit
                         /// maximal number of iteration for "add"
    , m_nIterMaxII( 5 )  // maximal number of iteration for "add"
                         /// maximal number of iteration for "remove"
    , m_nIterMaxIII( 5 ) // maximal number of iteration for "remove"
    , m_DistanceMax( 1.0 * Gaudi::Units::micrometer )
    , m_DistanceChi2( 1.0 * Gaudi::Units::perCent )
    , m_transporterName( "DaVinci::ParticleTransporter:PUBLIC" )
    , m_transporter( 0 )
    , m_seedZmin( -1.0 * Gaudi::Units::meter )
    , m_seedZmax( 3.0 * Gaudi::Units::meter )
    , m_seedRhoZmax( 80.0 * Gaudi::Units::centimeter )
    , m_seedRhoZmin( 20.0 * Gaudi::Units::centimeter )
    /// Allow "rho+"-like particles ?
    , m_use_rho_like_branch( true ) // allow "rho+"-like particles ?
                                    // Use short-lived particles as seeds
    , m_use_shortlived_as_seed( true )
    /// The transport tolerance
    , m_transport_tolerance( 10 * Gaudi::Units::micrometer )
    /// Use optimized algorithm ?
    , m_use_optimized( true ) // ATTENTION !!!
                              /// Recursive refit?
    , m_recursive_refit( true )
    ///
    , m_massage() {
  //
  if ( std::string::npos != name.find( "Massage" ) || std::string::npos != name.find( "Downstream" ) ||
       std::string::npos != name.find( "DD" ) || std::string::npos != name.find( "V0" ) ||
       std::string::npos != name.find( "KS" ) || std::string::npos != name.find( "Ks" ) ||
       std::string::npos != name.find( "K0S" ) || std::string::npos != name.find( "Lam" ) ||
       std::string::npos != name.find( "L0" ) || std::string::npos != name.find( "Gam" ) ||
       std::string::npos != name.find( "State" ) || std::string::npos != name.find( "Master" ) ) {
    m_massage.push_back( 180 * Gaudi::Units::cm );
    m_massage.push_back( 120 * Gaudi::Units::cm );
    m_massage.push_back( 60 * Gaudi::Units::cm );
    m_massage.push_back( 10 * Gaudi::Units::cm );
    m_transporterName = "DaVinci::ParticleTransporter:PUBLIC";
  }
  //
  if ( std::string::npos != name.find( "Opt" ) || std::string::npos != name.find( "OPT" ) ) { m_use_optimized = true; }

  declareProperty( "MaxIterations", m_nIterMaxI, "Maximal number of iterations" );
  declareProperty( "MaxIterForAdd", m_nIterMaxII, "Maximal number of iterations for 'Add'-method" );
  declareProperty( "MaxIterForRemove", m_nIterMaxIII, "Maximal number of iterations for 'Remove'-method" );
  declareProperty( "DeltaDistance", m_DistanceMax, "Delta-distance as convergency criterion" );
  declareProperty( "DeltaChi2", m_DistanceChi2, "Delta-chi2     as convergency criterion" );
  declareProperty( "AllowRhoPlusLikeParticle", m_use_rho_like_branch, "Allow ``rho+''-like particles?" );
  declareProperty( "USeShortLivedParticlesAsSeed", m_use_shortlived_as_seed, "Use shortlived particles as seed" );
  declareProperty( "TransportTolerance", m_transport_tolerance, "The tolerance for particle transport" );
  declareProperty( "UseOptimizedAlgorithm", m_use_optimized, "Use optimised algorithm" );

  declareProperty( "Massage", m_massage, "Z-coordinates for downstream track massage" );

  declareProperty( "SeedZmin", m_seedZmin, "Allowed Z-min   for seed-vertex " );
  declareProperty( "SeedZmax", m_seedZmax, "Allowed Z-max   for seed-vertex " );
  declareProperty( "SeedRhoZmax", m_seedRhoZmax, "Allowed Rho (at Zmax) for seed-vertex " );
  declareProperty( "SeedRhoZmin", m_seedRhoZmin, "Allowed Rho (at Zmin) for seed-vertex " );
  declareProperty( "Transporter", m_transporterName, "The typename of transporter to be used" );

  declareProperty( "RecursiveReFit", m_recursive_refit, "Refit recursively?" );
}
// ============================================================================
// the standard initialization of the tool
// ============================================================================
StatusCode LoKi::VertexFitter::initialize() {
  /// initialize the base class
  StatusCode sc = LoKi::ParticleClassificator::initialize();
  if ( sc.isFailure() ) { return sc; }
  //
  svc<IService>( "LoKiSvc", true );
  //
  if ( m_DistanceChi2 < 0.0005 ) {
    m_DistanceChi2 = 0.0005;
    info() << "Redefine 'DeltaChi2' property to " << m_DistanceChi2 << endmsg;
  }
  //
  if ( m_DistanceMax < 0.001 * Gaudi::Units::micrometer ) {
    m_DistanceMax = 0.001 * Gaudi::Units::micrometer;
    info() << "Redefine 'DeltaDistance' property to " << m_DistanceMax << endmsg;
  }
  //
  if ( m_nIterMaxI > 100 ) {
    m_nIterMaxI = 100;
    info() << "Redefine 'MaxIterations' property to " << m_nIterMaxI << endmsg;
  }
  //
  if ( m_nIterMaxII > 100 ) {
    m_nIterMaxII = 100;
    info() << "Redefine 'MaxIterForAdd' property to " << m_nIterMaxII << endmsg;
  }
  //
  if ( m_transport_tolerance < 0.1 * Gaudi::Units::micrometer ) {
    m_transport_tolerance = 0.1 * Gaudi::Units::micrometer;
    info() << "Redefine 'TransportTolerance' property to " << m_transport_tolerance << endmsg;
  }
  //
  if ( !m_use_optimized ) { info() << "Option for Optimised Kalman Filter fit is de-activated " << endmsg; }
  //
  if ( !m_massage.empty() && std::string::npos == m_transporterName.find( "DaVinci::" ) ) {
    warning() << "The transporter '" << m_transporterName << "'is not optimal for activated ``massage''-option"
              << " ``Massage'' is OFF" << endmsg;
    m_massage.clear();
  }
  //
  if ( !m_massage.empty() ) { info() << "``Massage''-option is ON" << endmsg; }
  //
  return StatusCode::SUCCESS;
}
// ========================================================================
// the standard finalization of the tool
// ========================================================================
StatusCode LoKi::VertexFitter::finalize() { return LoKi::ParticleClassificator::finalize(); }
// ============================================================================
// transport the data to a certain position
// ============================================================================
StatusCode LoKi::VertexFitter::_transport( LoKi::VertexFitter::Entry& entry, Gaudi::XYZPoint const& point,
                                           IGeometryInfo const& geometry ) const {
  const double refPointZ = entry.m_p.referencePoint().Z();
  const double parZ      = entry.m_parx[2];
  //
  // for short-lived particles one needs to transport them into their decay vertex
  //
  if ( LoKi::KalmanFilter::ShortLivedParticle == entry.m_type ) {
    //
    // get the end-vertex
    //
    const LHCb::VertexBase* vertex = entry.m_p0->endVertex();
    if ( 0 != vertex ) {
      const double vertexZ = vertex->position().Z();
      // already properly transported ?
      if ( std::fabs( refPointZ - vertexZ ) < m_transport_tolerance &&
           std::fabs( parZ - vertexZ ) < m_transport_tolerance ) {
        return StatusCode::SUCCESS;
      } // RETURN
      // else : transport them into own vertex
      return LoKi::KalmanFilter::transport( entry, vertex->position(), transporter(), geometry ); // RETURN
    }
  }
  /// special "transport" for rho+-like particles
  else if ( LoKi::KalmanFilter::RhoPlusLikeParticle == entry.m_type ) {
    return _transportRhoPlusLike( entry, point, geometry );
  }
  //
  // no need for transport?
  //
  if ( std::fabs( refPointZ - point.Z() ) < m_transport_tolerance &&
       std::fabs( parZ - point.Z() ) < m_transport_tolerance ) {
    return StatusCode::SUCCESS;
  } // RETURN
  //
  // finally: transport it!
  //
  return LoKi::KalmanFilter::transport( entry, point, transporter(), geometry );
}
// ============================================================================
// transport the data to a certain position
// ============================================================================
StatusCode LoKi::VertexFitter::_transportRhoPlusLike( LoKi::VertexFitter::Entry& entry, Gaudi::XYZPoint const& point,
                                                      IGeometryInfo const& geometry ) const {
  const double refPointZ = entry.m_p.referencePoint().Z();
  const double parZ      = entry.m_parx[2];
  //
  // no need for transport?
  //
  if ( std::fabs( refPointZ - point.Z() ) < m_transport_tolerance &&
       std::fabs( parZ - point.Z() ) < m_transport_tolerance ) {
    return StatusCode::SUCCESS;
  } // RETURN
  //
  const LHCb::Particle* lparticle = longFromRhoPlusLike_( entry.m_p0 );
  if ( !lparticle ) {
    ++m_rho_wrong_structure;
    return StatusCode::FAILURE;
  }
  // temporary particle
  LHCb::Particle temp( *lparticle );
  StatusCode     sc = transporter()->transportAndProject( lparticle, point.Z(), temp, geometry );
  if ( sc.isFailure() ) {
    ++m_rho_transport;
    return sc;
  }
  //
  const Gaudi::SymMatrix3x3& _pmcov = temp.posCovMatrix();
  Gaudi::SymMatrix2x2        cixy;
  // basic particle? : use some tricks to
  cixy( 0, 0 ) = _pmcov( 0, 0 );
  cixy( 0, 1 ) = _pmcov( 0, 1 );
  cixy( 1, 1 ) = _pmcov( 1, 1 );
  // Invert matrix: use Cholesky's decomposition; if it fails, use regular inversion
  if ( !cixy.InvertChol() && !cixy.Invert() ) {
    ++m_rho_bad_inverse;
    return StatusCode::FAILURE;
  } // RETURN
  //
  // The most tricky part I
  entry.m_vxi( 0, 0 ) = cixy( 0, 0 );
  entry.m_vxi( 0, 1 ) = cixy( 0, 1 );
  entry.m_vxi( 1, 1 ) = cixy( 1, 1 );
  //
  // The most tricky part II
  const Gaudi::LorentzVector& mom = temp.momentum();
  const Gaudi::Vector2        slopes( mom.Px() / mom.Pz(), mom.Py() / mom.Pz() );
  const Gaudi::Vector2        cslope( cixy * slopes );
  entry.m_vxi( 0, 2 ) = -cslope( 0 );
  entry.m_vxi( 1, 2 ) = -cslope( 1 );
  entry.m_vxi( 2, 2 ) = ROOT::Math::Similarity( slopes, cixy );
  //
  // transported particle:
  //
  LHCb::Particle& tp = entry.m_p;
  //
  tp.setMomentum( temp.momentum() );
  tp.setReferencePoint( temp.referencePoint() );
  //
  tp.setMomCovMatrix( temp.momCovMatrix() );
  tp.setPosCovMatrix( temp.posCovMatrix() );
  tp.setPosMomCovMatrix( temp.posMomCovMatrix() );
  //
  bool                                   found = false;
  typedef SmartRefVector<LHCb::Particle> DAUGHTERS;
  const DAUGHTERS&                       daughters = entry.m_p0->daughters();
  //
  for ( DAUGHTERS::const_iterator idau = daughters.begin(); daughters.end() != idau; ++idau ) {
    const LHCb::Particle* dau = *idau;
    if ( dau == lparticle ) {
      found = true;
      continue;
    } // CONTINUE
    //
    StatusCode sc = _transportCalo( dau, temp, entry.m_p.referencePoint() );
    if ( sc.isFailure() ) ++m_rho_transport_calo;
    //
    const_cast<Gaudi::LorentzVector&>( tp.momentum() ) += temp.momentum();
    const_cast<Gaudi::SymMatrix4x4&>( tp.momCovMatrix() ) += temp.momCovMatrix();
    const_cast<Gaudi::SymMatrix3x3&>( tp.posCovMatrix() ) += temp.posCovMatrix();
    const_cast<Gaudi::Matrix4x3&>( tp.posMomCovMatrix() ) += temp.posMomCovMatrix();
  }
  if ( !found ) {
    ++m_rho_wrong_structure;
    return StatusCode::FAILURE;
  }
  //
  return StatusCode::SUCCESS;
}
// ============================================================================
// transport gamma & digmma & merged pi0 particles
// ============================================================================
StatusCode LoKi::VertexFitter::_transportCalo( const LHCb::Particle* gamma, LHCb::Particle& transported,
                                               const Gaudi::XYZPoint&     point,
                                               const Gaudi::SymMatrix3x3* pointCov2 ) const {
  if ( !gamma ) { return StatusCode::FAILURE; }
  // assign
  if ( gamma != &transported ) { transported = *gamma; }
  // the actual type:
  const LoKi::KalmanFilter::ParticleType t = particleType_( gamma );
  if ( LoKi::KalmanFilter::MergedPi0LikeParticle == t ) {
    int status = 0;
    if ( !pointCov2 ) {
      LHCb::Calo::CaloParticle calo( &transported, point );
      calo.updateParticle();
      status = calo.status();
    } else {
      LHCb::Calo::CaloParticle calo( &transported, point, *pointCov2 );
      calo.updateParticle();
      status = calo.status();
    }
    //
    if ( 0 != status ) ++m_rho_transport_pi0;
    //
    // mass and mass uncertainties
    transported.setMeasuredMass( gamma->measuredMass() ); // corrected mass
    transported.setMeasuredMassErr( gamma->measuredMassErr() );
    //
    return StatusCode::SUCCESS;
  }
  //
  // Gamma-like & Digamma-like entry
  LHCb::Calo::Momentum calo;
  //
  calo.setReferencePoint( point );
  if ( 0 != pointCov2 ) { calo.setPosCovMatrix( *pointCov2 ); }
  //
  if ( LoKi::KalmanFilter::GammaLikeParticle == t ) {
    const LHCb::ProtoParticle* proto = gamma->proto();
    if ( !proto ) {
      ++m_calo_wrong_proto;
      return StatusCode::FAILURE;
    }
    //
    calo.addCaloPosition( proto );
  } else if ( LoKi::KalmanFilter::DiGammaLikeParticle == t ) {
    // the first gamma  :
    const LHCb::Particle* gamma1 = LoKi::Child::child( gamma, 1 );
    if ( !gamma1 ) {
      ++m_calo_invalid_first_gamma;
      return StatusCode::FAILURE;
    }
    const LHCb::ProtoParticle* proto1 = gamma1->proto();
    if ( !proto1 ) {
      ++m_calo_invalid_proto_1;
      return StatusCode::FAILURE;
    }
    // the second gamma :
    const LHCb::Particle* gamma2 = LoKi::Child::child( gamma, 2 );
    if ( !gamma2 ) {
      ++m_calo_invalid_second_gamma;
      return StatusCode::FAILURE;
    }
    const LHCb::ProtoParticle* proto2 = gamma2->proto();
    if ( !proto1 ) {
      ++m_calo_invalid_proto_2;
      return StatusCode::FAILURE;
    }
    //
    calo.addCaloPosition( proto1 );
    calo.addCaloPosition( proto2 );
    //
  } else {
    ++m_calo_illegal;
    return StatusCode::FAILURE;
  } // RETURN
  //
  const bool ok = calo.evaluate();
  if ( !ok ) {
    ++m_calo_invalid;
    return StatusCode::FAILURE;
  }
  //
  // extract the values:
  transported.setReferencePoint( point );
  transported.setMomentum( calo.momentum() );
  transported.setMomCovMatrix( calo.momCovMatrix() );
  //
  if ( 0 != pointCov2 ) {
    transported.setPosCovMatrix( calo.pointCovMatrix() );
    transported.setPosMomCovMatrix( calo.momPointCovMatrix() );
  } else {
    Gaudi::Math::setToScalar( const_cast<Gaudi::SymMatrix3x3&>( transported.posCovMatrix() ), 0.0 );
    Gaudi::Math::setToScalar( const_cast<Gaudi::Matrix4x3&>( transported.posMomCovMatrix() ), 0.0 );
  }
  //
  return StatusCode::SUCCESS;
}

DECLARE_COMPONENT( LoKi::VertexFitter )
