/*****************************************************************************\
* (c) Copyright 2000-2018 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/
#pragma once

#include <algorithm>
#include <set>

#include "GaudiKernel/Chrono.h"

#include "Event/Particle.h"
#include "Event/Vertex.h"

#include "Kernel/IParticleTransporter.h"
#include "Kernel/IVertexFit.h"

#include "LHCbMath/MatrixTransforms.h"
#include "LHCbMath/MatrixUtils.h"

#include "Math/Functions.h"

#include "KalmanFilter/VertexFit.h"

#include "ParticleClassificator.h"

/** @file
 *
 *  This file is a part of
 *  <a href="http://cern.ch/lhcb-comp/Analysis/LoKi/index.html">LoKi project:</a>
 *  ``C++ ToolKit for Smart and Friendly Physics Analysis''
 *
 *  This file is a part of LoKi project -
 *    "C++ ToolKit  for Smart and Friendly Physics Analysis"
 *
 *  The package has been designed with the kind help from
 *  Galina PAKHLOVA and Sergey BARSUK.  Many bright ideas,
 *  contributions and advices from G.Raven, J.van Tilburg,
 *  A.Golutvin, P.Koppenburg have been used in the design.
 *
 *  @author Vanya BELYAEV ibelyaev@physics.syr.edu
 *  @date 2006-05-24
 */
namespace LoKi {

  /** @class VertexFitter
   *  the most trivial implementation of Kalman-filter based vertex fitter.
   *  It should be very efficient (CPU).
   *
   *  @todo proper implementation of IVertexFit:remove
   *
   *  <b>"Vertex Fitter as Kalman Filter"</b>
   *
   *  Let  \f$\vec{p}_k\f$ be a vector containing the parameters of the
   *  \f$k\f$-track. Let  \f$V_k\f$ be the covariance matrix of these parameters,
   *  and  \f$G_k\f$ is <b>the formal</b> inverse matrix: \f$G_k = V_k^{-1}\f$.
   *  For the trivial case of \f$dim(\vec{p}_k)=5\f$ both matrices
   *  \f$V_k\f$ and \f$G_k\f$ could be defined properly and are non-singular.
   *
   *  Here we'll try to extend the standard Kalman Filter formalism
   *  for <b>almost</b> linear model, by embedding the
   *  minimal \f$dim=5\f$-model into more dimensions.
   *  In particular we choose the \f$dim(\vec{p}_k)=7\f$:
   *  \f[ \vec{p} = p^T =
   *      \left( \vec{x} , \vec{q} \right) =
   *      \left( x , y , z , p_x , p_y , p_z , E \right)
   *  \f]
   *
   * The vertex fit determines the new parameters
   * \f$\vec{x}\f$- position of the common vertex and
   * \f$\vec{q}\f$- 4-momentum of the track,
   * constrained to originate in \f$\vec{x}\f$.
   *
   *
   * The covariance matrix of fit parameters could be split into sub-matrices:
   *
   *  - the vertex part \f$ 3\times3\f$:       \f$C = cov\left( \vec{x} \right)\f$
   *  - particle momentum part \f$4\times4\f$: \f$D_k = cov\left( \vec{q}_k \right)\f$
   *  - correlation between the vertex position and the particle
   *    momentum part \f$3\times4\f$: \f$E_k = cor\left( \vec{x},\vec{q}_k \right)\f$
   *
   *  - correlation between the momenta of different particles \f$4\times4\f$:
   *    \f$S_{k,l} = cor\left( \vec{q}_l,\vec{q}_k \right)\f$
   *
   *  Such choice of \f$\vec{p}\f$ and \f$\vec{q}\f$
   *  allows to obtain almost trivial projection matrices \f$A\f$ and
   *  \f$B\f$:
   *  \f[ \vec{p}_k \left( \vec{x}_k , \vec{q}_k \right) =
   *      c_k^0 +  A_k \vec{x} + B_k \vec{q},  \f]
   *  where
   *  \f$ A_k = \left[ \frac{\partial \vec{p} }{\partial \vec{x}} \right] \f$ and
   *  \f$ B_k  = \left[ \frac{\partial \vec{p} }{\partial \vec{q}} \right] \f$.
   *  Clearly with the the chosen  parameterization of \f$\vec{p}\f$ one gets:
   *  \f$A_k = A = \bigl( \begin{smallmatrix}1 \\ 0\end{smallmatrix}\bigr)\f$,
   *  \f$B_k = B = \bigl( \begin{smallmatrix}0 \\ 1\end{smallmatrix}\bigr)\f$.
   *
   *
   *  The Kalman filter works as:
   *
   *  - \f$ C^{-1}_k = C^{-1}_{k-1} + A^T_{k}G_{k}^{B}A_{k} \f$
   *  - \f$ E_k      = -F_{k}C_{k}  \f$
   *  - \f$ D_{k}    = W_{k} - E_{k}F^{T}_{k} \f$
   *
   *  where:
   *
   *  - \f$ W_{k}     = \left( B^T_{k}G_{k}B_{k}\right)^{-1}  \f$
   *  - \f$ G^{B}_{k} = G_{k} - G_{k}B_{k}W_{k}B^T_{k}G_{k} \f$
   *  - \f$ F_{k}     = W_{k}B^T_{k}G_{k}A_{k} \f$
   *
   *  New fit values are:
   *
   * - \f$ \vec{x}_{k} = C_{k}\left(C^{-1}_{k-1}\vec{x}{k-1}
   *           + A^T_{k}G^B_{k}\left(\vec{p}_{k}-c^0_{k}\right)\right) \f$
   * - \f$ \vec{q}_{k} = W_{k}B^T_{k}G_{k}\left(\vec{p}_{k}
   *           - c^0_k-A_k\vec{x}_k\right) \f$
   *
   *  New \f$\chi^2\f$ is :
   *   \f$ \chi^2_{k} = \chi^2_{k-1} + \chi_1 +\chi_2\f$, where
   *  - \f$ \chi_1 = \vec{\delta_1}^{T} G_k \vec{\delta_1} \f$
   *  - \f$ \chi_2 = \vec{\delta_2}^{T} C^{-1}_{k-1} \vec{\delta_2} \f$
   *  - \f$ \vec{\delta}_1 = \vec{p}-c^0_{k}-A_{k}\vec{x}-B_{k}\vec{q}_{k}\f$
   *  - \f$ \vec{\delta}_2 = \vec{x}_{k}-\vec{x}_{k-1} \f$
   *
   *  The Kalman smoothing step is fairly trivial:
   *
   *  - \f$ \vec{x}^{n}_{k} = \vec{x}_{n} \f$
   *  - \f$ \vec{q}^{n}_{k} = W_{k}B^T_{k}G_{k}\left[\vec{p}-A_{k}\vec{x}_{n}\right] \f$
   *
   *  After the iteration the covariance matrices are calculated as:
   *
   *  - \f$ C^{n}_{k} = C_{n} \f$
   *  - \f$ E^{n}_{k} = -F_{k}C_{n} \f$
   *  - \f$ D^{n}_{k} = W_k - E^{n}_{k}F^T_{k} \f$
   *  - \f$ cov(\vec{q}^{n}_{k},\vec{q}^{n}_{l}) = F_{k}C_{n}F^T_{l} \f$
   *
   *  Since for our parameterization
   *  \f$A_k = A = \bigl( \begin{smallmatrix}1 \\ 0\end{smallmatrix}\bigr)\f$,
   *  \f$B_k = B = \bigl( \begin{smallmatrix}0 \\ 1\end{smallmatrix}\bigr)\f$,
   *  all formulare become relaively simle:
   *
   *  Assuming that
   *  \f$ G_{k} = \bigl( \begin{smallmatrix}
   *       G_x  & G_{xp} \\ G^T_{xp} & G_{p} \end{smallmatrix}\bigr), \f$
   *  one has
   *
   *  - \f$ W_k = G_p^{-1} \f$
   *  - \f$ G^B_k = \bigl( \begin{smallmatrix}
   *   G_x - G_{xp}G_{p}^{-1}G^{T}_{xp} & 0 \\ 0 & 0 \end{smallmatrix} \bigr) \f$
   *  - \f$ F_k       = G_p^{-1} G^{T}_{xp} \f$
   *  - \f$ C_k^{-1}  = C_{k-1}^{-1} + \left( G_x - G_{xp}G_p^{-1}G_{xp}^T\right) \f$
   *  - \f$ D_k       = G_p^{-1}+ G_p^{-1}G^T_{xp}C_kG_{xp}G_p^{-1} \f$
   *  - \f$ \vec{x}_x = C_k\left( C^{-1}_{k-1}\vec{x}_{k-1}+
   *            \left( G_x - G_{xp}G^{-1}_{p}G^{T}_{xp} , 0 \right)
   *            \left( \vec{p}_k - c^0_k \right) \right) \f$
   *  - \f$ \chi^2_{k} = \chi^2_{k-1} + \delta_{\vec{x}}^T
   *           \left[ G_x - G_{xp}G_{p}^{-1}G^T_{xp} \right] \delta_{\vec{x}}
   *           + ... \f$
   *
   *
   *  In an obvious way:
   *
   *  - \f$ G_x - G_{xp}G_{p}^{-1}G_{xp} = V^{-1}_{x}     \f$
   *  - \f$ G_{p}^{-1} = V_{p} - V^T_{xp}V^{-1}_{x}V_{xp} \f$
   *  - \f$ G^{-1}G^T_{xp} = -V^T_{xp}V^{-1}_{x}          \f$
   *
   *  @author Vanya BELYAEV ibelyaev@physics.syr.edu
   *  @date 2006-05-24
   */

  class VertexFitter : public extends<ParticleClassificator, IVertexFit> {
  public:
    /// error codes
    enum {
      /// Invalid Particle
      InvalidParticle = 701, // Invalid Particle
      /// Invalid Data
      InvalidData = 702, //     Invalid Data
      /// Error in Matrix Inversion
      ErrorInMatrixInversion = 703, //  Error in Matrix Inversion
      /// No Convergency
      NoConvergency = 710, // No Convergency is detected
      /// Not Implemented Yet
      NotImplemented = 720 //        not yet implemented
    };

  protected:
    typedef LoKi::KalmanFilter::Entry   Entry;
    typedef LoKi::KalmanFilter::Entries Entries;
    typedef Entries::iterator           EIT;

    struct FitData {
      /// local auxillary storages
      Entries             m_entries;
      const LHCb::Vertex* m_vertex;
      // seed (evaluated internally)
      Gaudi::Vector3      m_seed;
      Gaudi::SymMatrix3x3 m_seedci;
      // internal helper matrices
      Gaudi::SymMatrix4x4 m_cmom;
      Gaudi::Matrix4x4    m_cmom1;
      Gaudi::Matrix4x3    m_mpcov;
      // for measured mass
      Gaudi::SymMatrix4x4 m_mm_c;
    };

  public:
    /** The vertex fitting method without creation of a Particle
     *
     *  @code
     *
     *  // locate  the tool
     *  const IVertexFit* fitter = tool<IVertexFit>( ... ) ;
     *
     *  // container of particles to be fitetr into common vertex
     *  LHCb::Particle::ConstVector daughters = ... ;
     *
     *  Vertex vertex ;
     *  StatusCode sc = fitter->fit( daughters , vertex )
     *  if( sc.isFailure() ) { Warning("Error in vertex fit", sc ) ; }
     *
     *  // check chi2 of common vertex fit
     *  double chi2 = vertex->chi2() ;
     *
     *  @endcode
     *
     *  @see IVertexFit
     *  @see Particle
     *  @see Vertex
     *
     *  @param daughters vector of daughter particles  (input)
     *  @param vertex result of vertex fit             (output)
     *  @return status code
     */
    StatusCode fit( LHCb::Vertex& vertex, LHCb::Particle::ConstVector const& daughters,
                    IGeometryInfo const& geometry ) const override;

    /** The vertex fitting method with the creation of LHCb::Particle
     *  ("classical")
     *
     *  @code
     *
     *  // locate  the tool
     *  const IVertexFit* fitter = tool<IVertexFit>( ... ) ;
     *
     *  // container of particles to be fitted into common vertex
     *  LHCb::Particle::ConstVector daughters = ... ;
     *
     *  Vertex   vertex   ;
     *  LHCb::Particle particle ;
     *  StatusCode sc = fitter->fit( daughters , particle , vertex )
     *  if ( sc.isFailure() ) { Warning ( "Error in vertex fit", sc ) ; }
     *
     *  // check chi2 of common vertex fit
     *  double chi2 = vertex -> chi2() ;
     *
     *  @endcode
     *
     *  @see IVertexFit
     *  @see Particle
     *  @see Vertex
     *
     *  @param daughters vector of daughter particles  (input)
     *  @param particle  result of vertex fit             (output)
     *  @param vertex    result of vertex fit             (output)
     *  @return status code
     */
    StatusCode fit( LHCb::Particle::ConstVector const& daughters, LHCb::Vertex& vertex, LHCb::Particle& particle,
                    IGeometryInfo const& geometry ) const override;

    /** add the particle to the vertex and refit the vertex
     *
     *  @code
     *
     *  // locate  the tool
     *  const IVertexFit* fitter = get<IVertexFit>( ... ) ;
     *
     *  LHCb::Vertex*   vertex  = ... ;
     *  LHCb::Particle* oneMore = ... ;
     *
     *  StatusCode sc = fitter -> add ( oneMore , vertex ) ;
     *  if( sc.isFailure() ) { Warning("Error in 'add' ", sc ) ; }
     *
     *  // check chi2 of common vertex fit
     *  double chi2 = vertex -> chi2 () ;
     *
     *  @endcode
     *
     *  @see IVertexFit
     *  @param particle particle to be added
     *  @param vertex   vertex to be modified
     *  @return status code
     */
    StatusCode add( LHCb::Particle const* particle, LHCb::Vertex& vertex,
                    IGeometryInfo const& geometry ) const override;

    /** remove the particle from the vertex and refit the vertex
     *
     *  @code
     *
     *  // locate  the tool
     *  const IVertexFit* fitter = get<IVertexFit>( ... ) ;
     *
     *  LHCb::Vertex*   vertex = ... ;
     *  LHCb::Particle* remove = ... ;
     *
     *  StatusCode sc = fitter -> remove ( remove , vertex ) ;
     *  if( sc.isFailure() ) { Warning("Error in 'remove' ", sc ) ; }
     *
     *  // check chi2 of common vertex fit
     *  double chi2 = vertex -> chi2 () ;
     *
     *  @endcode
     *
     *  @param particle particle to be removed
     *  @param vertex   vertex to be modified
     *  @return status code
     */
    StatusCode remove( LHCb::Particle const* particle, LHCb::Vertex& vertex,
                       IGeometryInfo const& geometry ) const override;

    /** The major method for "combining" the daughter particles
     *  into "mother" particle.
     *
     *  The container of input particles is "combined" into
     *  "mother" particle and its decay vertex.
     *
     *  The intermediate constrains (mass, vertex, mass-vertex,
     *  etc.. could be applied in the process of "combining",
     *  dependent on the used implementation)
     *
     *  @code
     *
     *  // locate the tool
     *  const IParticleCombiner* combiner = get<IParticleCombiner>( ... )
     *
     *  // loop over the first daughter particle
     *  for( ... ipi1 = ... ; ... ; ++ipi1 )
     *  {
     *     for( ... ipi2 = ... ; ... ; ++ipi2 )
     *     {
     *        const LHCb::Particle* pi1 = *ipi1 ;
     *        const LHCb::Particle* pi2 = *ipi2 ;
     *
     *        IParticleCombiner::LHCb::Particle::ConstVector daughters ;
     *        daughters.push_back ( pi1 ) ;
     *        daughters.push_back ( pi2 ) ;
     *
     *        LHCb::Particle K0S  ;
     *        Vertex   Vrtx ;
     *
     *        StatusCode sc = combiner -> combine ( daughters , K0S , Vrtx ) ;
     *        if( sc.isFailure() )
     *        { Warning ( "Error in K0S fit, skip the pair", sc ) ; }
     *
     *        ....
     *     }
     *  }
     *
     *  @endcode
     *
     *  @see IParticleCombiner
     *
     *  @param daughters the vector of daughter particles    (input)
     *  @param mother    the "mother" particle               (output)
     *  @param vertex    the decay vertex of mother particle (output)
     *  @return status code
     */
    StatusCode combine( LHCb::Particle::ConstVector const& daughters, LHCb::Particle& mother, LHCb::Vertex& vertex,
                        IGeometryInfo const& geometry ) const override {
      return fit( daughters, vertex, mother, geometry );
    }

    /** The basic method for "refit" of the particle
     *
     *  @code
     *
     *  // locate the tool
     *  const IParticleReFitter* refitter = tool<IParticleRefitter>( ... ) ;
     *
     *  // particle to be refit
     *  LHCb::Particle* p = ... ;
     *
     *  StatusCode sc = refitter -> reFit ( *p ) ;
     *  if ( sc.isFailure() ) { Warning ( "Error in reFit" , sc ) ; }
     *
     *  @endcode
     *
     *  @see IParticleReFitter
     *
     *  @see Particle
     *
     *  @param particle reference to the particle
     *  @return statsu code
     */
    StatusCode reFit( LHCb::Particle& particle, IGeometryInfo const& geometry ) const override {
      // first refit daughters (if needed)
      if ( m_recursive_refit ) {
        typedef SmartRefVector<LHCb::Particle> DAUGHTERS;
        const DAUGHTERS&                       daughters = particle.daughters();
        for ( DAUGHTERS::const_iterator ip = daughters.begin(); daughters.end() != ip; ++ip ) {
          //
          const LHCb::Particle* d = *ip;
          if ( nullptr == d ) { continue; }         // CONTINUE
          if ( d->isBasicParticle() ) { continue; } // CONTINUE
          //
          const LHCb::Vertex* v = d->endVertex();
          if ( nullptr == v ) { continue; } // CONTINUE
          //
          LoKi::KalmanFilter::ParticleType pType = particleType_( d );
          //
          if ( LoKi::KalmanFilter::UnspecifiedParticle == pType ) {
            continue;
          } else if ( LoKi::KalmanFilter::GammaLikeParticle == pType ) {
            continue;
          } else if ( LoKi::KalmanFilter::DiGammaLikeParticle == pType ) {
            continue;
          } else if ( LoKi::KalmanFilter::MergedPi0LikeParticle == pType ) {
            continue;
          } else if ( LoKi::KalmanFilter::JetLikeParticle == pType ) {
            continue;
          } else if ( LoKi::KalmanFilter::RhoPlusLikeParticle == pType && !m_use_rho_like_branch ) {
            continue;
          }
          //
          StatusCode sc = reFit( *const_cast<LHCb::Particle*>( d ), geometry );
          if ( sc.isFailure() ) {
            Error( "Can't re-fit recursively, skip", sc ).ignore();
            continue;
          }
          //
        }
      }
      //
      Chrono chrono( timing() ? chronoSvc() : nullptr, timing() ? name() + ":reFit" : std::string() );
      //
      // play a bit with extra-info
      if ( particle.hasInfo( LHCb::Particle::additionalInfo::Chi2OfParticleReFitter ) ) {
        particle.eraseInfo( LHCb::Particle::additionalInfo::Chi2OfParticleReFitter );
      }
      //
      LHCb::Vertex*                          vertex = particle.endVertex();
      const IVertexFit*                      vFit   = this;
      typedef SmartRefVector<LHCb::Particle> DAUGHTERS;
      const DAUGHTERS&                       daughters = particle.daughters();
      StatusCode sc = vFit->fit( daughters.begin(), daughters.end(), *vertex, particle, geometry );
      if ( sc.isFailure() ) { return Error( "reFit(): error from fit()", sc ); }
      //
      // in the case of success update the extra-info:
      if ( particle.hasInfo( LHCb::Particle::additionalInfo::Chi2OfVertexConstrainedFit ) ) {
        particle.addInfo( LHCb::Particle::additionalInfo::Chi2OfParticleReFitter,
                          particle.info( LHCb::Particle::additionalInfo::Chi2OfVertexConstrainedFit, -1000 ) );
      }
      //
      return StatusCode::SUCCESS;
    }

    /// the standard initialization of the tool
    StatusCode initialize() override; // the standard initialization of the tool
    /// the standard finalization of the tool
    StatusCode finalize() override; // the standard finalization of the tool

    /// standard constructor
    VertexFitter( const std::string& type, const std::string& name, const IInterface* parent );

  protected:
    // fit methods that also accept FitData struct, wrapped by the public methods.
    StatusCode _fit( LHCb::Vertex& vertex, LHCb::Particle::ConstVector const& daughters, IGeometryInfo const& geometry,
                     FitData& data ) const;
    StatusCode _fit( LHCb::Particle::ConstVector const& daughters, LHCb::Vertex& vertex, LHCb::Particle& particle,
                     IGeometryInfo const& geometry, FitData& data ) const;

    /// load the data into internal representation
    StatusCode _load( const LHCb::Particle* particle, FitData& data ) const;
    /// load the data from the daughter particles into the internal structures
    StatusCode _load( const LHCb::Particle::ConstVector& daughters, FitData& data ) const;
    /// add one particle at the end of the queue
    StatusCode _add( LHCb::Particle const* child, Gaudi::XYZPoint const& point, IGeometryInfo const& geometry,
                     FitData& data ) const;
    /// transport the data to a certain position
    StatusCode _transport( Entry& entry, Gaudi::XYZPoint const& point, IGeometryInfo const& geometry ) const;
    /// transport the data to a certain position
    StatusCode _transportRhoPlusLike( Entry& entry, Gaudi::XYZPoint const& point, IGeometryInfo const& geometry ) const;
    /// transport all data to a certain position
    StatusCode _transport( Gaudi::XYZPoint const& point, IGeometryInfo const& geometry, FitData& data ) const {
      for ( EIT entry = data.m_entries.begin(); data.m_entries.end() != entry; ++entry ) {
        StatusCode sc = _transport( *entry, point, geometry );
        if ( sc.isFailure() ) { Warning( "_transport(): the error from transport(), ignore", sc ).ignore(); }
      }
      return StatusCode::SUCCESS;
    }
    /** transport calorimeteric particles:
     *  - gamma
     *  - digmma
     *  - merged pi0
     */
    StatusCode _transportCalo( LHCb::Particle const* gamma, LHCb::Particle& transported, Gaudi::XYZPoint const& point,
                               Gaudi::SymMatrix3x3 const* pointCov2 = 0 ) const;
    /// transport all data to a certain position
    StatusCode _transport( const Gaudi::Vector3& p, IGeometryInfo const& geometry, FitData& data ) const {
      return _transport( Gaudi::XYZPoint( p[0], p[1], p[2] ), geometry, data );
    }
    /// make optimised fKalman filter iterations
    StatusCode _iterate_opt( size_t nMax, IGeometryInfo const& geometry, FitData& data ) const;
    /// special iterations for rho+like branch
    StatusCode _iterate_rho( size_t nMax, IGeometryInfo const& geometry, FitData& data ) const;
    /// stopping criteria for iterations
    bool stop_iter( const Gaudi::Vector3& dx, const Gaudi::SymMatrix3x3& ci, const size_t iter, FitData& data ) const;
    /// make few Kalman filter iterations
    StatusCode _iterate( size_t nMax, Gaudi::Vector3 const& x, IGeometryInfo const& geometry, FitData& data ) const;
    /// make a seed
    StatusCode _seed( LHCb::Vertex const* vertex, IGeometryInfo const& geometry, FitData& data ) const;
    /// is seed ok?
    bool seedOK( const Gaudi::XYZPoint& seed ) const {
      return seed.Z() < m_seedZmax && seed.Z() > m_seedZmin && seed.Perp2() < rho2max( seed.Z() );
    }
    //
    bool seedOK( const Gaudi::Vector3& seed ) const { return seedOK( Gaudi::XYZPoint( seed[0], seed[1], seed[2] ) ); }
    /// get maximal allowed value of rho^2 for given z
    double rho2max( const double z ) const {
      //
      const double rho =
          ( m_seedRhoZmax * ( z - m_seedZmin ) - m_seedRhoZmin * ( z - m_seedZmax ) ) / ( m_seedZmin - m_seedZmax );
      //
      return rho * rho;
    }

    /// measure CPU performance ?
    bool timing() const { return m_timing; } // measure CPU performance ?

    /// get the transporter
    IParticleTransporter* transporter() const {
      if ( 0 != m_transporter ) { return m_transporter; }
      m_transporter = tool<IParticleTransporter>( m_transporterName, this );
      return m_transporter;
    }
    // ========================================================================
    //
    bool forMassage( const LHCb::Track* t ) const {
      return 0 != t && ( LHCb::Track::Types::Long != t->type() && LHCb::Track::Types::Velo != t->type() );
    }
    //
    bool forMassage( const LHCb::ProtoParticle* p ) const { return 0 != p && forMassage( p->track() ); }
    // not electron!
    bool forMassage( const LHCb::Particle* p ) const {
      return 0 != p && 11 != p->particleID().abspid() && forMassage( p->proto() );
    }
    //
    bool forMassage( const Entry& entry ) const { return forMassage( entry.m_p0 ); }
    //
    bool forMassage( const Entries& entries ) const {
      for ( Entries::const_iterator ientry = entries.begin(); entries.end() != ientry; ++ientry ) {
        if ( forMassage( *ientry ) ) { return true; }
      }
      return false;
    }

  private:
    /// maximal number of iterations for the vertex fit
    unsigned short m_nIterMaxI; // maximal number of iteration for vertex fit
    /// maximal number of iterations for "add"
    unsigned short m_nIterMaxII; // maximal number of iteration for "add"
    /// maximal number of iterations for "remove"
    unsigned short m_nIterMaxIII; // maximal number of iteration for "remove"
    /// distance (stop-iteration criterion)
    double m_DistanceMax; // distance (stop-iteration criterion)
    /// chi2 distance (stop iteration criterion)
    double m_DistanceChi2; // chi2 distance (stop iteration criterion)
    //
    /// propagator/extrapolator/transporter name
    std::string m_transporterName;
    /// the transporter itself
    mutable IParticleTransporter* m_transporter;
    //
    /// fiducial volume for valid seed/vertex
    double m_seedZmin;
    double m_seedZmax;
    double m_seedRhoZmax;
    double m_seedRhoZmin;
    /// allow "rho+"-like particles
    bool m_use_rho_like_branch; // allow "rho+"-like particles
    /// use short-lived particles as seed
    bool m_use_shortlived_as_seed;
    /// transport tolerance
    double m_transport_tolerance; // the transport tolerance
    /// use optimized algorithm ?
    bool m_use_optimized;
    /// recursive refit ?
    bool m_recursive_refit;

    // some action for track massage, induced by Edwige's findings
    std::vector<double> m_massage;

    // counters
    mutable Gaudi::Accumulators::Counter<>                    m_nbSeedCase0Counter{ this, "Seed:case(0)" };
    mutable Gaudi::Accumulators::Counter<>                    m_nbSeedCase1Counter{ this, "Seed:case(1)" };
    mutable Gaudi::Accumulators::Counter<>                    m_nbSeedCase2Counter{ this, "Seed:case(2)" };
    mutable Gaudi::Accumulators::Counter<>                    m_nbSeedCase3Counter{ this, "Seed:case(3)" };
    mutable Gaudi::Accumulators::Counter<>                    m_nbSeedCase4Counter{ this, "Seed:case(4)" };
    mutable Gaudi::Accumulators::Counter<>                    m_nbSeedCase5Counter{ this, "Seed:case(5)" };
    mutable Gaudi::Accumulators::Counter<>                    m_nbSeedCase6Counter{ this, "Seed:case(6)" };
    mutable Gaudi::Accumulators::Counter<>                    m_nbSeedCase7Counter{ this, "Seed:case(7)" };
    mutable Gaudi::Accumulators::SummingCounter<unsigned int> m_nbIterGenCounter{ this, "#iterations/Gen" };
    mutable Gaudi::Accumulators::SummingCounter<unsigned int> m_nbIterRhoCounter{ this, "#iterations/rho" };
    mutable Gaudi::Accumulators::SummingCounter<unsigned int> m_nbSeedMassageCounter{ this, "Seed:massage" };
    mutable Gaudi::Accumulators::StatCounter<>                m_nIterations1{ this, "#iterations/1" };
    mutable Gaudi::Accumulators::StatCounter<>                m_nIterations2{ this, "#iterations/2" };
    mutable Gaudi::Accumulators::StatCounter<>                m_nIterationsOpt{ this, "#iterations/Opt" };
    mutable Gaudi::Accumulators::StatCounter<>                m_badMassError{ this, "bad mass error" };
    mutable Gaudi::Accumulators::StatCounter<>                m_badMass{ this, "bad mass " };
    mutable Gaudi::Accumulators::MsgCounter<MSG::WARNING> m_fit_seed_failure{ this, "fit(): failure from _seed()", 0 };
    mutable Gaudi::Accumulators::MsgCounter<MSG::WARNING> m_fit_failure_iterator{ this,
                                                                                  "fit(): failure from _iterate()", 0 };
    mutable Gaudi::Accumulators::MsgCounter<MSG::WARNING> m_fit_small_negative_chi2{
        this, "fit(): small negative chi2 detected, adjusted", 0 };
    mutable Gaudi::Accumulators::MsgCounter<MSG::WARNING> m_fit_large_negative_chi2{
        this, "fit(): large negative chi2 detected, ignore", 0 };
    mutable Gaudi::Accumulators::MsgCounter<MSG::WARNING> m_fit_outside_fiducial{
        this, "fit(): Vertex is outside of fiducial volume", 0 };
    mutable Gaudi::Accumulators::MsgCounter<MSG::WARNING> m_fit_failure{ this, "fit(): failure from fit", 0 };
    mutable Gaudi::Accumulators::MsgCounter<MSG::WARNING> m_fit_mmas_e_bit_np{
        this, "fit(): measured mass error is a bit    non-positive, adjust to 1 MeV", 0 };
    mutable Gaudi::Accumulators::MsgCounter<MSG::WARNING> m_fit_mmas_e_slight_np{
        this, "fit(): measured mass error is slightly non-positive, adjust to 1 MeV", 0 };
    mutable Gaudi::Accumulators::MsgCounter<MSG::WARNING> m_fit_mmas_e_np{
        this, "fit(): measured mass error is non-positive, adjust to 1 MeV", 0 };
    mutable Gaudi::Accumulators::MsgCounter<MSG::WARNING> m_fit_mmas_bit_np{
        this, "fit(): measured mass is a bit    non-positive , adjust to 1 MeV", 0 };
    mutable Gaudi::Accumulators::MsgCounter<MSG::WARNING> m_fit_mmas_slight_np{
        this, "fit(): measured mass is slightly non-positive , adjust to 1 MeV", 0 };
    mutable Gaudi::Accumulators::MsgCounter<MSG::WARNING> m_fit_mmas_np{
        this, "fit(): measured mass is non-positive , adjust to 1 MeV", 0 };
    mutable Gaudi::Accumulators::MsgCounter<MSG::ERROR> m_fit_load{ this, "fit(): failure from _load() ", 0 };
    mutable Gaudi::Accumulators::MsgCounter<MSG::ERROR> m_add_no_particle{ this, "add: Particle* point to NULL!" };
    mutable Gaudi::Accumulators::MsgCounter<MSG::ERROR> m_add_no_gamma{ this, "add: No way to add      Gamma-like" };
    mutable Gaudi::Accumulators::MsgCounter<MSG::ERROR> m_add_no_digamma{ this, "add: No way to add    DiGamma-like" };
    mutable Gaudi::Accumulators::MsgCounter<MSG::ERROR> m_add_no_mergedpi0{ this,
                                                                            "add: No way to add Merged-Pi0-like" };
    mutable Gaudi::Accumulators::MsgCounter<MSG::WARNING> m_add_fit{ this, "add: error from 'fit'" };
    mutable Gaudi::Accumulators::MsgCounter<MSG::WARNING> m_add_failure{ this, "add(): failure from _add", 0 };
    mutable Gaudi::Accumulators::MsgCounter<MSG::WARNING> m_add_failure_ignore{
        this, "_add(): the error from _add()      , ignore" };
    mutable Gaudi::Accumulators::MsgCounter<MSG::WARNING> m_add_transport_ignore{
        this, "_add(): the error from _transport(), ignore" };
    mutable Gaudi::Accumulators::MsgCounter<MSG::WARNING> m_add_step{ this, "add(): failure from _step", 0 };
    mutable Gaudi::Accumulators::MsgCounter<MSG::WARNING> m_add_smooth{ this, "add(): failure from _smooth", 0 };
    mutable Gaudi::Accumulators::MsgCounter<MSG::WARNING> m_add_iterate{ this, "add(): failure from _iterate", 0 };
    mutable Gaudi::Accumulators::MsgCounter<MSG::ERROR>   m_no_remove{ this, "remove(): not implemented (yet)!" };
    mutable Gaudi::Accumulators::MsgCounter<MSG::ERROR>   m_rho_wrong_structure{ this,
                                                                               "Rho+-transport: wrong structure[1]!" };
    mutable Gaudi::Accumulators::MsgCounter<MSG::WARNING> m_rho_transport{ this, "Error in rho+-transport", 0 };
    mutable Gaudi::Accumulators::MsgCounter<MSG::WARNING> m_rho_bad_inverse{ this,
                                                                             "Rho+-transport: can't inverse matrix!" };
    mutable Gaudi::Accumulators::MsgCounter<MSG::WARNING> m_rho_transport_calo{
        this, "Rho+-transport: can't transport `Calo'!" };
    mutable Gaudi::Accumulators::MsgCounter<MSG::WARNING> m_rho_transport_pi0{
        this, "Calo-transport: can't update merged pi0!" };
    mutable Gaudi::Accumulators::MsgCounter<MSG::WARNING> m_rho_kalman{ this, "Error from  rho+-like Kalman step" };
    mutable Gaudi::Accumulators::MsgCounter<MSG::WARNING> m_kalman{ this, "Error from Kalman-step, skip" };
    mutable Gaudi::Accumulators::MsgCounter<MSG::WARNING> m_rho_iterate_transport{
        this, "_iterate(): problem with transport " };
    mutable Gaudi::Accumulators::MsgCounter<MSG::WARNING> m_no_convergency{ this, "No convergency has been reached",
                                                                            0 };
    mutable Gaudi::Accumulators::MsgCounter<MSG::WARNING> m_no_proper_seed{ this, "No proper seed is found", 0 };
    mutable Gaudi::Accumulators::MsgCounter<MSG::WARNING> m_load_error{ this, "_load(): the error from _load:" };
    mutable Gaudi::Accumulators::MsgCounter<MSG::WARNING> m_load_no_valid_data{ this, "_load(): no valid data found" };
    mutable Gaudi::Accumulators::MsgCounter<MSG::WARNING> m_load_no_particle{ this, "_load(): invalid particle" };
    mutable Gaudi::Accumulators::MsgCounter<MSG::ERROR>   m_not_vertexed{ this, "Input set could not be vertexed" };
    mutable Gaudi::Accumulators::MsgCounter<MSG::WARNING> m_stop_iter_sn_chi2{
        this, "stop_iter: small negative chi2 detected, adjusted", 0 };
    mutable Gaudi::Accumulators::MsgCounter<MSG::WARNING> m_stop_iter_ln_chi2{
        this, "stop_iter: large negative chi2 detected, ignore", 0 };
    mutable Gaudi::Accumulators::MsgCounter<MSG::WARNING> m_stop_iter_smoother{
        this, "stop_iter(): problem with smoother", 0 };
    mutable Gaudi::Accumulators::MsgCounter<MSG::WARNING> m_stop_iter_cov{
        this, "stop_iter(): problems with covariances", 0 };
    mutable Gaudi::Accumulators::MsgCounter<MSG::WARNING> m_iterate_steps{ this, "_iterate_opt(): problem with steps" };
    mutable Gaudi::Accumulators::MsgCounter<MSG::WARNING> m_iterate_transport{
        this, "_iterate_opt(): problem with transport" };
    mutable Gaudi::Accumulators::MsgCounter<MSG::WARNING> m_iterate_no_convergence{
        this, "_iterate_opt: No convergency has been reached", 0 };
    mutable Gaudi::Accumulators::MsgCounter<MSG::WARNING> m_iterate_transport2{ this,
                                                                                "_iterate(): problem with transport " };
    mutable Gaudi::Accumulators::MsgCounter<MSG::WARNING> m_no_convergence{ this, "No convergency has been reached" };
    mutable Gaudi::Accumulators::MsgCounter<MSG::ERROR>   m_calo_wrong_proto{ this,
                                                                            "Calo-transport: wrong protoparticle[1]!" };
    mutable Gaudi::Accumulators::MsgCounter<MSG::ERROR>   m_calo_invalid_first_gamma{
        this, "Calo-transport: invalid first gamma" };
    mutable Gaudi::Accumulators::MsgCounter<MSG::ERROR> m_calo_invalid_second_gamma{
        this, "Calo-transport: invalid second gamma" };
    mutable Gaudi::Accumulators::MsgCounter<MSG::ERROR> m_calo_invalid_proto_1{
        this, "Calo-transport: invalid protoparticle-1" };
    mutable Gaudi::Accumulators::MsgCounter<MSG::ERROR> m_calo_invalid_proto_2{
        this, "Calo-transport: invalid protoparticle-2" };
    mutable Gaudi::Accumulators::MsgCounter<MSG::ERROR> m_calo_illegal{ this, "Calo-transport: illegal case" };
    mutable Gaudi::Accumulators::MsgCounter<MSG::ERROR> m_calo_invalid{ this, "Calo-transport: invalid calo momentum" };

    Gaudi::Property<bool> m_timing{ "MeasureCPUPerformance", false, "Measure CPU perormance" };
  };

} //                                                      end of namespace LoKi
