/*****************************************************************************\
* (c) Copyright 2000-2018 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/
// ============================================================================
#ifndef LOKI_DECAYS_H
#  define LOKI_DECAYS_H 1
// ============================================================================
// Include files
// ============================================================================
// PartProp
// ============================================================================
#  include "Kernel/Nodes.h"
#  include "Kernel/NodesPIDs.h"
// ============================================================================
// Event
// ============================================================================
#  include "Event/Particle.h"
// ============================================================================
// LoKi
// ============================================================================
#  include "LoKi/Trees.h"
// ============================================================================
/** @file
 *  Implementation file for LoKi specific decay/tree-functors
 *  @author Vanya BELYAEV Ivan.BELYAEV@nikhef.nl
 *  @date   2008-05-25
 */
// ============================================================================
namespace Decays {
  // ========================================================================
  namespace Trees {
    // ======================================================================
    /** @class Exclusive
     *  Simple sub tree which consists of the node ("mother") and
     *  subtrees ("children"). Essentially it represent also the
     *  "semi-exclusive" decay (as well as any decay with the fixed
     *  number of components for comparisons).
     *
     *  @author Vanya BELYAEV Ivan.Belyaev@nikhef.nl
     *  @date 2008-06-11
     */
    class GAUDI_API Exclusive : public Decays::iTree_<const LHCb::Particle*> {
    protected:
      // ======================================================================
      typedef const LHCb::Particle*           PARTICLE;
      typedef Decays::Trees::Types_<PARTICLE> Types;
      // ======================================================================
    public:
      // ======================================================================
      /// the actual type of the container of children trees
      typedef Types::TreeList TreeList;
      // ======================================================================
    public:
      // ======================================================================
      /** full constructor from the node (mother), subtrees & flag
       *  @param mother the mother node
       *  @param children the list of daughter substrees
       *  @param alg the matching algorithm
       */
      Exclusive( const Decays::iNode& mother, const TreeList& children, const Alg alg = Daughters );
      /** constructor from the node (mother) & flag
       *  @param mother the mother node
       *  @param alg the matching algorithm
       */
      Exclusive( const Decays::iNode& mother, const Alg alg = Daughters );
      /** constructor from the decay
       *  @param decay  the decay
       *  @param alg the matching algorithm
       */
      Exclusive( const Decays::Decay& decay, const Alg alg = Daughters );
      // ====================================================================
    public:
      // ====================================================================
      /// MANDATORY: clone method ("virtual constructor")
      Exclusive* clone() const override { return new Exclusive( *this ); }
      /// MANDATORY: the only one essential method:
      bool operator()( Decays::iTree_<PARTICLE>::argument p ) const override;
      /// MANDATORY: check the validity
      bool valid() const override;
      /// MANDATORY: validate the tree
      StatusCode validate( const LHCb::IParticlePropertySvc* svc ) const override;
      /// MANDATORY: reset the tree
      void reset() const override;
      /// collect the marked elements
      size_t collect( Decays::iTree_<PARTICLE>::Collection& output ) const override;
      /// MANDATORY: the specific printout
      std::ostream& fillStream( std::ostream& s ) const override;
      /// has marked elements in tree?
      bool marked() const override;
      // ====================================================================
    private:
      // ======================================================================
      /// implementation of operator()
      virtual bool p_match( Decays::iTree_<PARTICLE>::argument p ) const;
      // ======================================================================
    public:
      // ====================================================================
      /// check the validness & mother
      inline bool ok( const LHCb::Particle* p ) const { return 0 != p && mother( p->particleID() ); }
      // ====================================================================
      /// get the algorithm
      Alg alg() const { return m_alg; }
      /// set the algortihm
      void setAlgorith( const Alg value ) { m_alg = value; }
      // ====================================================================
    public:
      // ====================================================================
      /// set children
      void setChildren( const TreeList& children ) { m_children = children; }
      /// set children
      void setChildren( const std::vector<std::string>& children );
      /// set children
      void setChildren( const std::vector<Decays::Decay::Item>& children );
      /// set children
      void setChildren( const std::vector<LHCb::ParticleID>& children );
      /// set children
      void setChildren( const std::vector<const LHCb::ParticleProperty*>& children );
      // ====================================================================
    public:
      // ====================================================================
      /// add one more daughter to the decay
      void addDaughter( const Decays::iTree_<PARTICLE>& tree );
      /// add one more daughter to the decay
      void addDaughter( const Decays::iNode& node );
      /// add one more daughter to the decay
      void addDaughter( const Decays::Decay::Item& node );
      /// add one more daughter to the decay
      void addDaughter( const std::string& node );
      /// add one more daughter to the decay
      void addDaughter( const LHCb::ParticleID& node );
      /// add one more daughter to the decay
      void addDaughter( const LHCb::ParticleProperty* node );
      // ====================================================================
    public:
      // ====================================================================
      /// add one more node to the tree
      Exclusive& operator+=( const Decays::iTree_<PARTICLE>& node ) {
        addDaughter( node );
        return *this;
      }
      /// add one more node to the tree
      Exclusive& operator+=( const Decays::iNode& node ) {
        addDaughter( node );
        return *this;
      }
      /// add one more node to the tree
      Exclusive& operator+=( const Decays::Decay::Item& node ) {
        addDaughter( node );
        return *this;
      }
      /// add one more node to the tree
      Exclusive& operator+=( const std::string& node ) {
        addDaughter( node );
        return *this;
      }
      /// add one more node to the tree
      Exclusive& operator+=( const LHCb::ParticleID& node ) {
        addDaughter( node );
        return *this;
      }
      /// add one more node to the tree
      Exclusive& operator+=( const LHCb::ParticleProperty* node ) {
        addDaughter( node );
        return *this;
      }
      // ====================================================================
    protected:
      // ====================================================================
      inline const TreeList& children() const { return m_children; }
      // ====================================================================
      TreeList::const_iterator  childBegin() const { return m_children.begin(); }
      TreeList::const_iterator  childEnd() const { return m_children.end(); }
      TreeList::const_reference front() const { return m_children.front(); }
      TreeList::const_reference back() const { return m_children.back(); }
      // ======================================================================
      size_t nChildren() const { return m_children.size(); }
      // ======================================================================
      // reset the cache
      inline void i_reset() const {
        std::for_each( childBegin(), childEnd(), []( auto& child ) { child.reset(); } );
      }
      // ======================================================================
      const Decays::iNode& mother() const { return m_mother; }
      // ======================================================================
      inline bool mother( const LHCb::ParticleID& pid ) const { return m_mother.node( pid ); }
      // ======================================================================
    private:
      // ======================================================================
      /// The mother
      Decays::Node m_mother; //    the mother
      /// The children
      TreeList m_children; //  the children
      /// The algorithm
      Decays::Trees::Alg m_alg; // the algorithm
      // ======================================================================
    };
    // ========================================================================
    /** @class Inclusive
     *  Simple sub tree which consists of the node ("mother") and
     *  subtrees ("children"). Essentially it represent the
     *  inclusive decay.
     *  @author Vanya BELYAEV Ivan.Belyaev@nikhef.nl
     *  @date 2008-06-11
     */
    class GAUDI_API Inclusive : public Exclusive {
    public:
      // ======================================================================
      /** full constructor from the node (mother), subtrees & flag
       *  @param mother the mother node
       *  @param children the list of daughter substrees
       *  @param alg the matching algorithm
       */
      Inclusive( const Decays::iNode& mother, const TreeList& children, const Alg alg = Daughters );
      /** constructor from the node (mother) & flag
       *  @param mother the mother node
       *  @param alg the matching algorithm
       */
      Inclusive( const Decays::iNode& mother, const Alg alg = Daughters );
      /** constructor from the decay
       *  @param decay  the decay
       *  @param alg the matching algorithm
       */
      Inclusive( const Decays::Decay& decay, const Alg alg = Daughters );
      /// constructor from exclusive decay
      Inclusive( const Exclusive& right );
      // ====================================================================
    public:
      // ====================================================================
      /// MANDATORY: clone method ("virtual constructor")
      Inclusive* clone() const override { return new Inclusive( *this ); }
      /// MANDATORY: check the validity of the tree
      bool valid() const override;
      /// MANDATORY: the specific printout
      std::ostream& fillStream( std::ostream& s ) const override;
      // ====================================================================
    private:
      // ======================================================================
      /// implementation of operator()
      bool p_match( Decays::iTree_<PARTICLE>::argument p ) const override;
      // ======================================================================
    public:
      // ====================================================================
      /// add one more node to the tree
      Inclusive& operator+=( const Decays::iTree_<PARTICLE>& tree ) {
        addDaughter( tree );
        return *this;
      }
      /// add one more node to the tree
      Inclusive& operator+=( const Decays::iNode& node ) {
        addDaughter( node );
        return *this;
      }
      /// add one more node to the tree
      Inclusive& operator+=( const Decays::Decay::Item& node ) {
        addDaughter( node );
        return *this;
      }
      /// add one more node to the tree
      Inclusive& operator+=( const std::string& node ) {
        addDaughter( node );
        return *this;
      }
      /// add one more node to the tree
      Inclusive& operator+=( const LHCb::ParticleID& node ) {
        addDaughter( node );
        return *this;
      }
      /// add one more node to the tree
      Inclusive& operator+=( const LHCb::ParticleProperty* node ) {
        addDaughter( node );
        return *this;
      }
      // ======================================================================
    };
    // ========================================================================
    /** @class Optional
     *  Simple sub tree which consists of the node ("mother") and
     *  subtrees ("children") and optional nodes. Essentially it represent the
     *  decays with optional particles
     *  @author Vanya BELYAEV Ivan.Belyaev@nikhef.nl
     *  @date 2008-04-13
     */
    class GAUDI_API Optional : public Exclusive {
    public:
      // ======================================================================
      /** full constructor from the node (mother) and subtrees
       *  @param mother the mother node
       *  @param children the list of children
       *  @param optional the list of optional components
       *  @param alg the matching algorithm
       */
      Optional( const Decays::iNode& mother, const TreeList& children = TreeList(),
                const TreeList& optional = TreeList(), const Alg alg = Daughters );
      // ====================================================================
      /** constructor from decay descriptor, optional and flags
       *  @param decay the decay descriptor
       *  @param optional the list of optional components
       *  @param alg the matching algorithm
       */
      Optional( const Decays::Decay& decay, const TreeList& optional = TreeList(), const Alg alg = Daughters );
      /// constructor form the constructed tree
      Optional( const Exclusive& right, const TreeList& optional = TreeList() );
      // ====================================================================
    public:
      // ====================================================================
      /// MANDATORY: clone method ("virtual constructor")
      Optional* clone() const override { return new Optional( *this ); }
      /// MANDATORY: check the validity of the tree
      bool valid() const override;
      /// MANDATORY: the specific printout
      std::ostream& fillStream( std::ostream& s ) const override;
      /// MANDATORY: the proper validation of the tree
      StatusCode validate( const LHCb::IParticlePropertySvc* svc ) const override;
      /// has marked elements in tree?
      bool marked() const override;
      // ====================================================================
    private:
      // ======================================================================
      /// implementation of operator()
      bool p_match( Decays::iTree_<PARTICLE>::argument p ) const override;
      // ======================================================================
    public:
      // ====================================================================
      /// set optional
      void setOptional( const TreeList& optional ) { m_optional = optional; }
      /// set optional
      void setOptional( const std::vector<std::string>& optional );
      /// set children
      void setOptional( const std::vector<Decays::Decay::Item>& optional );
      /// set optional
      void setOptional( const std::vector<LHCb::ParticleID>& optional );
      /// set optional
      void setOptional( const std::vector<const LHCb::ParticleProperty*>& optional );
      // ====================================================================
    public:
      // ====================================================================
      /// add one more daughter to the decay
      void addOptional( const Decays::iTree_<PARTICLE>& tree );
      /// add one more daughter to the decay
      void addOptional( const Decays::iNode& node );
      /// add one more daughter to the decay
      void addOptional( const std::string& node );
      /// add one more daughter to the decay
      void addOptional( const LHCb::ParticleID& node );
      /// add one more daughter to the decay
      void addOptional( const Decays::Decay::Item& node );
      /// add one more daughter to the decay
      void addOptional( const LHCb::ParticleProperty* node );
      // ====================================================================
    public:
      // ====================================================================
      /// add one more node to the tree
      Optional& operator+=( const Decays::iTree_<PARTICLE>& node ) {
        addDaughter( node );
        return *this;
      }
      /// add one more node to the tree
      Optional& operator+=( const Decays::iNode& node ) {
        addDaughter( node );
        return *this;
      }
      /// add one more node to the tree
      Optional& operator+=( const std::string& node ) {
        addDaughter( node );
        return *this;
      }
      /// add one more node to the tree
      Optional& operator+=( const LHCb::ParticleID& node ) {
        addDaughter( node );
        return *this;
      }
      /// add one more node to the tree
      Optional& operator+=( const Decays::Decay::Item& node ) {
        addDaughter( node );
        return *this;
      }
      /// add one more node to the tree
      Optional& operator+=( const LHCb::ParticleProperty* node ) {
        addDaughter( node );
        return *this;
      }
      // ======================================================================
    protected:
      // ======================================================================
      TreeList::const_iterator optBegin() const { return m_optional.begin(); }
      TreeList::const_iterator optEnd() const { return m_optional.end(); }
      size_t                   optSize() const { return m_optional.size(); }
      size_t                   nOptional() const { return optSize(); }
      // ======================================================================
    private:
      // ======================================================================
      /// the optional particles in the tree
      TreeList m_optional; // the optional particles in the tree
      // ======================================================================
    };
    // ========================================================================
    // treat properly the stable trees
    // ========================================================================
    template <>
    bool Stable_<const LHCb::Particle*>::operator()( Decays::iTree_<const LHCb::Particle*>::argument p ) const;
    // ========================================================================
  } // namespace Trees
  // ==========================================================================
} //                                                    end of namespace Decays
// ============================================================================
// The END
// ============================================================================
#endif // LOKI_DECAYS_H
// ============================================================================
