/*****************************************************************************\
* (c) Copyright 2000-2018 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/
// ============================================================================
#ifndef LOKI_PARTICLES1_H
#  define LOKI_PARTICLES1_H 1
// ============================================================================
// Include files
// ============================================================================
// Event
// ============================================================================
#  include "Event/Particle.h"
// ============================================================================
// LoKiPhys
// ============================================================================
#  include "LoKi/PhysTypes.h"
#  include "LoKi/Vertices1.h"
// ============================================================================
/** @file
 *
 *  This file is a part of LoKi project -
 *    "C++ ToolKit  for Smart and Friendly Physics Analysis"
 *
 *  The package has been designed with the kind help from
 *  Galina PAKHLOVA and Sergey BARSUK.  Many bright ideas,
 *  contributions and advices from G.Raven, J.van Tilburg,
 *  A.Golutvin, P.Koppenburg have been used in the design.
 *
 *  @author Vanya BELYAEV ibelyaev@physics.syr.edu
 *  @date 2006-02-19
 */
// ============================================================================
namespace LoKi {
  namespace Particles {
    // ========================================================================
    /** @class VertexDistance
     *
     *  evaluator of the GEOMETRY
     *  distance between the particle
     *  "endVertex" and "the vertex".
     *
     *  The actual evaluation
     *  is delegated to the function
     *  LoKi::Vertices::VertexDistance
     *
     *  @see LoKi::Vertices::VertexDistance
     *  @see LHCb::Vertex
     *  @see LHCb::Particle
     *
     *  @see LoKi::Cuts::VD
     *
     *  @author Vanya Belyaev Ivan.Belyaev@itep.ru
     *  @date   2002-07-15
     */
    class GAUDI_API VertexDistance : public LoKi::BasicFunctors<const LHCb::Particle*>::Function {
    public:
      // ======================================================================
      /// constructor
      VertexDistance( const LHCb::VertexBase* vertex );
      /// constructor
      VertexDistance( const LoKi::Point3D& vertex );
      /// constructor
      VertexDistance( const LoKi::Vertices::VertexHolder& base );
      /// constructor
      VertexDistance( const LoKi::Vertices::VertexDistance& base );
      /// MANDATORY: clone method ("virtual constructor")
      VertexDistance* clone() const override;
      /// MANDATORY: the only one essential method
      result_type operator()( argument p ) const override;
      /// OPTIONAL: the specific printout
      std::ostream& fillStream( std::ostream& s ) const override;
      /// the actual computation
      result_type distance( argument p ) const;
      // ======================================================================
    public:
      // ======================================================================
      void setVertex( const LHCb::VertexBase* vertex ) const { m_fun.setVertex( vertex ); }
      void setVertex( const LoKi::Point3D& vertex ) const { m_fun.setVertex( vertex ); }
      // ======================================================================
    private:
      // ======================================================================
      /// default constructor is private
      VertexDistance();
      // ======================================================================
    private:
      // ======================================================================
      // the actual evaluator
      LoKi::Vertices::VertexDistance m_fun;
      // ======================================================================
    };
    // ========================================================================
    /** @class VertexSignedDistance
     *
     *  evaluator of the GEOMETRY
     *  distance between the particle
     *  "endVertex" and "the vertex".
     *  The distance is signed according
     *  to the sign of (Zv-Z0)
     *
     *  The actual evaluation
     *  is delegated to the function
     *  LoKi::Vertices::VertexSignedDistance
     *
     *  @see LoKi::Vertices::VertexSignedDistance
     *  @see LHCb::Vertex
     *  @see LHCb::Particle
     *
     *  @see LoKi::Cuts::VDSIGN
     *
     *  @author Vanya Belyaev Ivan.Belyaev@itep.ru
     *  @date   2002-07-15
     */
    class GAUDI_API VertexSignedDistance : public LoKi::BasicFunctors<const LHCb::Particle*>::Function {
    public:
      /// constructor
      VertexSignedDistance( const LHCb::VertexBase* vertex );
      /// constructor
      VertexSignedDistance( const LoKi::Point3D& vertex );
      /// constructor
      VertexSignedDistance( const LoKi::Vertices::VertexHolder& base );
      /// constructor
      VertexSignedDistance( const LoKi::Vertices::VertexDistance& base );
      /// constructor
      VertexSignedDistance( const LoKi::Vertices::VertexSignedDistance& base );
      /// MANDATORY: clone method ("virtual constructor")
      VertexSignedDistance* clone() const override;
      /// MANDATORY: the only one essential method
      result_type operator()( argument p ) const override { return signedDistance( p ); }
      /// OPTIONAL: the specific printout
      std::ostream& fillStream( std::ostream& s ) const override;
      /// the actual computation
      result_type signedDistance( argument p ) const;
      // ======================================================================
    public:
      // ======================================================================
      void setVertex( const LHCb::VertexBase* vertex ) const { m_fun.setVertex( vertex ); }
      void setVertex( const LoKi::Point3D& vertex ) const { m_fun.setVertex( vertex ); }
      // ======================================================================
    private:
      // ======================================================================
      /// default constructor is private
      VertexSignedDistance();
      // ======================================================================
    private:
      // ======================================================================
      // the actual evaluator
      LoKi::Vertices::VertexSignedDistance m_fun;
      // ======================================================================
    };
    // ========================================================================
    /** @class VertexDotDistance
     *
     *  evaluator of the
     *  distance between the particle
     *  "endVertex" and "the vertex" along
     *  the particle momentum
     *
     *  @see LHCb::Vertex
     *  @see LHCb::Particle
     *
     *  @see LoKi::Cuts::VDDOT
     *
     *  @author Vanya Belyaev Ivan.Belyaev@itep.ru
     *  @date   2002-07-15
     */
    class GAUDI_API VertexDotDistance : public LoKi::BasicFunctors<const LHCb::Particle*>::Function,
                                        public LoKi::Vertices::VertexHolder {
    public:
      // ======================================================================
      /// constructor
      VertexDotDistance( const LHCb::VertexBase* vertex );
      /// constructor
      VertexDotDistance( const LoKi::Point3D& vertex );
      /// constructor
      VertexDotDistance( const LoKi::Vertices::VertexHolder& base );
      /// MANDATORY: clone method ("virtual constructor")
      VertexDotDistance* clone() const override;
      /// MANDATORY: the only one essential method
      result_type operator()( argument p ) const override;
      /// OPTIONAL: the specific printout
      std::ostream& fillStream( std::ostream& s ) const override;
      /// the actual computation
      result_type distance( argument p ) const;
      // ======================================================================
    private:
      // ======================================================================
      /// default constructor is private
      VertexDotDistance(); // default constructor is private
      // ======================================================================
    };
    // ========================================================================
    /** @class VertexChi2Distance
     *
     *  evaluator of the chi2 of GEOMETRY
     *  distance between the particle
     *  "endVertex" and "the vertex"
     *
     *  The actual evaluation
     *  is delegated to the function
     *  LoKi::Vertices::VertexChi2Distance
     *
     *  @see LoKi::Vertices::VertexChi2Distance
     *  @see LHCb::Vertex
     *  @see LHCb::Particle
     *
     *  @see LoKi::Cuts::VDCHI2
     *
     *  @author Vanya Belyaev Ivan.Belyaev@itep.ru
     *  @date   2002-07-15
     */
    class GAUDI_API VertexChi2Distance : public LoKi::BasicFunctors<const LHCb::Particle*>::Function {
    public:
      // ======================================================================
      /// constructor
      VertexChi2Distance( const LHCb::VertexBase* vertex );
      /// constructor
      VertexChi2Distance( const LoKi::Point3D& vertex );
      /// constructor
      VertexChi2Distance( const LoKi::Vertices::VertexHolder& base );
      /// constructor
      VertexChi2Distance( const LoKi::Vertices::VertexChi2Distance& vertex );
      /// MANDATORY: clone method ("virtual constructor")
      VertexChi2Distance* clone() const override;
      /// MANDATORY: the only one essential method
      result_type operator()( argument p ) const override;
      /// OPTIONAL: the specific printout
      std::ostream& fillStream( std::ostream& s ) const override;
      /// the actual computation
      result_type chi2( argument p ) const;
      // ======================================================================
    public:
      // ======================================================================
      void setVertex( const LHCb::VertexBase* vertex ) const { m_fun.setVertex( vertex ); }
      void setVertex( const LoKi::Point3D& vertex ) const { m_fun.setVertex( vertex ); }
      // ======================================================================
    private:
      // ======================================================================
      /// default constructor is private
      VertexChi2Distance();
      // ======================================================================
    private:
      // ======================================================================
      // the actual evaluator
      LoKi::Vertices::VertexChi2Distance m_fun;
      // ======================================================================
    };
    // ========================================================================
    /** @class MinVertexDistance
     *
     *  class for evaluation of minimal GEOMETRY distance
     *  between the "endVertex" and other vertices
     *
     *  @see LoKi::Cuts::MINVD
     *  @see LoKi::Cuts::VDMIN
     *
     *  @author Vanya ELYAEV Ivan.Belyaev@itep.ru
     *  @date   2004-07-08
     */
    class GAUDI_API MinVertexDistance : public LoKi::BasicFunctors<const LHCb::Particle*>::Function {
    public:
      // ======================================================================
      /** constructor from container of vertices
       *  @param vs container of vertices
       */
      MinVertexDistance( const LHCb::VertexBase::ConstVector& vs );
      /** constructor from container of vertices
       *  @param vs container of vertices
       */
      MinVertexDistance( const LHCb::Vertex::ConstVector& vs );
      /** constructor from container of vertices
       *  @param vs container of primary vertices
       */
      MinVertexDistance( const LHCb::RecVertex::ConstVector& vs );
      /** constructor from container of vertices
       *  @param vs container of primary vertices
       */
      MinVertexDistance( const LHCb::RecVertex::Container* vs );
      /** constructor from container of vertices
       *  @param vs container of primary vertices
       */
      MinVertexDistance( const LoKi::PhysTypes::VRange& vs );
      // ======================================================================
      /** templated constructor from arbitrary sequence
       *  of vertices and helper object
       *  @param first begin iterator of arbitrary sequence of primary vertices
       *  @param last  end iterator of arbitrary sequence of primary vertices
       */
      template <class VERTEX>
      MinVertexDistance( VERTEX first, VERTEX last )
          : LoKi::BasicFunctors<const LHCb::Particle*>::Function(), m_fun( first, last ) {}
      /** templated constructor from arbitrary sequence
       *  of vertices and helper object
       *  @param first begin iterator of arbitrary sequence of primary vertices
       *  @param last  end iterator of arbitrary sequence of primary vertices
       */
      template <class VERTEX>
      MinVertexDistance( const LoKi::Keeper<VERTEX>& keeper )
          : LoKi::BasicFunctors<const LHCb::Particle*>::Function(), m_fun( keeper ) {}
      /** templated constructor from arbitrary sequence
       *  of vertices and helper object
       *  @param first begin iterator of arbitrary sequence of primary vertices
       *  @param last  end iterator of arbitrary sequence of primary vertices
       */
      template <class VERTEX>
      MinVertexDistance( const LoKi::UniqueKeeper<VERTEX>& keeper )
          : LoKi::BasicFunctors<const LHCb::Particle*>::Function(), m_fun( keeper ) {}
      // ======================================================================
      /// MANDATORY: clone method ("virtual constructor")
      MinVertexDistance* clone() const override { return new MinVertexDistance( *this ); }
      /// MANDATORY: the only one essential method
      result_type operator()( argument v ) const override;
      /// OPTIONAL: the specific printout
      std::ostream& fillStream( std::ostream& s ) const override;
      // ======================================================================
    public:
      // ======================================================================
      size_t size() const { return m_fun.size(); }
      bool   empty() const { return m_fun.empty(); }
      // ======================================================================
    private:
      // ======================================================================
      /// the functor itself
      LoKi::Vertices::MinVertexDistance m_fun;
      // ======================================================================
    };
    // ========================================================================
    /** @class MinVertexChi2Distance
     *  class for evaluation of minimal chi2 distance
     *  between the vertex and other vertices
     *
     *  @see LoKi::Cuts::MINVDCHI2
     *  @see LoKi::Cuts::VDMINCHI2
     *
     *  @author Vanya ELYAEV Ivan.Belyaev@itep.ru
     *  @date   2004-07-08
     */
    class GAUDI_API MinVertexChi2Distance : public LoKi::BasicFunctors<const LHCb::Particle*>::Function {
    public:
      // ======================================================================
      /** constructor from container of vertices
       *  @param vs container of vertices
       */
      MinVertexChi2Distance( const LHCb::VertexBase::ConstVector& vs );
      /** constructor from container of vertices
       *  @param vs container of vertices
       */
      MinVertexChi2Distance( const LHCb::Vertex::ConstVector& vs );
      /** constructor from container of vertices
       *  @param vs container of primary vertices
       */
      MinVertexChi2Distance( const LHCb::RecVertex::ConstVector& vs );
      /** constructor from container of vertices
       *  @param vs container of primary vertices
       */
      MinVertexChi2Distance( const LHCb::RecVertex::Container* vs );
      /** constructor from container of vertices
       *  @param vs container of primary vertices
       */
      MinVertexChi2Distance( const LoKi::PhysTypes::VRange& vs );
      /** templated constructor from arbitrary sequence
       *  of vertices and helper object
       *  @param first begin iterator of arbitrary sequence of primary vertices
       *  @param last  end iterator of arbitrary sequence of primary vertices
       */
      template <class VERTEX>
      MinVertexChi2Distance( VERTEX first, VERTEX last )
          : LoKi::BasicFunctors<const LHCb::Particle*>::Function(), m_fun( first, last ) {}
      /// MANDATORY: clone method ("virtual constructor")
      MinVertexChi2Distance* clone() const override { return new MinVertexChi2Distance( *this ); }
      /// MANDATORY: the only one essential method
      result_type operator()( argument v ) const override;
      /// OPTIONAL: the specific printout
      std::ostream& fillStream( std::ostream& s ) const override;
      // ======================================================================
    private:
      // ======================================================================
      /// the functor
      LoKi::Vertices::MinVertexChi2Distance m_fun;
      // ======================================================================
    };
    // ========================================================================
  } // namespace Particles
  // ==========================================================================
} //                                                      end of namespace LoKi
// ============================================================================
//                                                                      The END
// ============================================================================
#endif // LOKI_PARTICLES1_H
// ============================================================================
