/*****************************************************************************\
* (c) Copyright 2000-2018 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/
// ============================================================================
#ifndef LOKI_PARTICLES24_H
#  define LOKI_PARTICLES24_H 1
// ============================================================================
// Include files
// ============================================================================
// STD & STL
// ============================================================================
#  include <iosfwd>
// ============================================================================
// PartProp
// ============================================================================
#  include "Kernel/iNode.h"
// ============================================================================
// LoKi
// ============================================================================
#  include "LoKi/BasicFunctors.h"
#  include "LoKi/iTree.h"
#  include <Event/Particle.h>
// ============================================================================
namespace LoKi {
  // ==========================================================================
  namespace Particles {
    // ========================================================================
    /** @class DecNode
     *  simple predicate, which acts on the particleID fo the particle
     *  @see LHCb::ParticleID
     *  @see Decays::iNode
     *  @author Vanya BELYAEV Ivan.Belyaev@nikhef.nl
     *  @date 2008-12-17
     */
    class GAUDI_API DecNode : public LoKi::BasicFunctors<const LHCb::Particle*>::Predicate {
    public:
      // ======================================================================
      /// constructor from the actual node
      DecNode( const Decays::iNode& node, const bool autovalidate = true );
      /// constructor from node descriptor
      DecNode( const std::string& node, const bool autovalidate = true );
      /// MANDATORY: clone method ("virtual constructor")
      DecNode* clone() const override { return new DecNode( *this ); }
      /// MANDATORY: the only one essential method
      result_type operator()( argument p ) const override;
      /// OPTIONAL: the nice printout
      std::ostream& fillStream( std::ostream& s ) const override;
      // ======================================================================
    private:
      // ======================================================================
      /// the default constructor is disabled
      DecNode(); // the default constructor is disabled
      // ======================================================================
    public:
      // ======================================================================
      /// get the decay node
      const Decays::iNode& node() const { return m_node.node(); }
      /// valid node ?
      bool valid() const { return m_node.valid(); }
      // validate the node
      StatusCode validate( const LHCb::IParticlePropertySvc* svc ) const { return m_node.validate( svc ); }
      // ======================================================================
    private:
      // ======================================================================
      /// the decay node itself
      Decays::Node m_node; // the decay node itself
      /// autovalidate ?
      bool m_autovalidate; // autovalidate
      // ======================================================================
    };
    // ========================================================================
    /** @class DecTree
     *  simple predicate, which acts on the decay structuire/tree for the particle
     *  @see Decays::iTree
     *  @author Vanya BELYAEV Ivan.Belyaev@nikhef.nl
     *  @date 2008-12-17
     */
    class GAUDI_API DecTree : public LoKi::BasicFunctors<const LHCb::Particle*>::Predicate {
    public:
      // ======================================================================
      /// the actual type of tree (interface
      typedef Decays::iTree_<const LHCb::Particle*> iTree; // the tree
      // ======================================================================
    private:
      // ======================================================================
      /// the actual type of tree (assignable)
      typedef Decays::Tree_<const LHCb::Particle*> Tree; // the tree
      // ======================================================================
    public:
      // ======================================================================
      /// constructor from the actual node
      DecTree( const iTree& node, const bool autovalidate = true );
      /// constructor from decay descriptot
      DecTree( const std::string& tree, const bool autovalidate = true );
      /// MANDATORY: clone method ("virtual constructor")
      DecTree* clone() const override { return new DecTree( *this ); }
      /// MANDATORY: the only one essential method
      result_type operator()( argument p ) const override;
      /// OPTIONAL: the nice printout
      std::ostream& fillStream( std::ostream& s ) const override;
      // ======================================================================
    private:
      // ======================================================================
      /// the default constructor is disabled
      DecTree(); // the default constructor is disabled
      // ======================================================================
    public:
      // ======================================================================
      /// get the decay node
      const iTree& tree() const { return m_tree.tree(); }
      // ======================================================================
      /// valid tree ?
      bool valid() const { return m_tree.valid(); }
      // validate the teh tree
      StatusCode validate( const LHCb::IParticlePropertySvc* svc ) const { return m_tree.validate( svc ); }
      // reset the collection
      void reset() const { m_tree.reset(); }
      // ======================================================================
    private:
      // ======================================================================
      /// the decay tree itself
      Tree m_tree; // the decay tree itself
      /// autovalidate ?
      bool m_autovalidate; // autovalidate
      // ======================================================================
    };
    // ========================================================================
  } // namespace Particles
  // ==========================================================================
} //                                                      end of namespace LoKi
// ============================================================================
//                                                                      The END
// ============================================================================
#endif // LOKI_PARTICLES24_H
// ============================================================================
