/*****************************************************************************\
* (c) Copyright 2000-2018 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/

#include "Kernel/IPVReFitter.h" // Interface
#include "TrackInterfaces/ITrackExtrapolator.h"

#include "Core/FloatComparison.h"
#include "Event/Particle.h"
#include "Event/ProtoParticle.h"
#include "Event/RecVertex.h"
#include "Event/VertexBase.h"

#include "GaudiAlg/GaudiTool.h"

using namespace LHCb;
using namespace Gaudi::Units;

/**
 *  reimplementation of the PVReFitter for DC06
 *
 *  @author Yuehong Xie
 *  @date   2006-06-08
 */
class PVReFitter : public GaudiTool, virtual public IPVReFitter {

private:
  typedef ROOT::Math::SMatrix<double, 10, 10, ROOT::Math::MatRepSym<double, 10>> SymMatrix10x10;
  typedef ROOT::Math::SMatrix<double, 8, 8, ROOT::Math::MatRepSym<double, 8>>    SymMatrix8x8;

public:
  /// Standard constructor
  PVReFitter( const std::string& type, const std::string& name, const IInterface* parent );
  StatusCode initialize() override;

  /// refit PV
  StatusCode reFit( LHCb::VertexBase* PV, IGeometryInfo const& geometry ) const override;

  /// remove track used for a (B) LHCb::Particle and refit PV
  StatusCode remove( const LHCb::Particle* part, LHCb::VertexBase* PV, IGeometryInfo const& geometry ) const override;

private:
  StatusCode fitPV( LHCb::RecVertex* PV, LHCb::Track::Vector& tracks, IGeometryInfo const& geometry ) const;
  void       getFinalTracks( const LHCb::Particle* part, LHCb::Track::ConstVector& tracks ) const;
  StatusCode seedPV( LHCb::RecVertex*, LHCb::Track*, LHCb::Track*, IGeometryInfo const& geometry ) const;
  StatusCode addTr( LHCb::RecVertex*, LHCb::Track*, IGeometryInfo const& geometry ) const;

  int    m_maxIter;
  double m_maxDeltaChi2;
  double m_maxDeltaZ;

  ITrackExtrapolator* m_fullExtrapolator;
  std::string         m_fullExtrapolatorName;
  ITrackExtrapolator* m_veloExtrapolator;
  std::string         m_veloExtrapolatorName;
};

// Declaration of the Tool Factory
DECLARE_COMPONENT( PVReFitter )

//=============================================================================
// Standard constructor, initializes variables
//=============================================================================
PVReFitter::PVReFitter( const std::string& type, const std::string& name, const IInterface* parent )
    : GaudiTool( type, name, parent ) {
  declareInterface<IPVReFitter>( this );

  declareProperty( "maxIter", m_maxIter = 10 );
  declareProperty( "maxDeltaChi2", m_maxDeltaChi2 = 0.001 );
  declareProperty( "maxDeltaZ", m_maxDeltaZ = 1.0 * mm );
  declareProperty( "fullExtrapolatorName", m_fullExtrapolatorName = "TrackMasterExtrapolator" );
  declareProperty( "veloExtrapolatorName", m_veloExtrapolatorName = "TrackLinearExtrapolator" );
}

//=============================================================================
// Initialize
//=============================================================================
StatusCode PVReFitter::initialize() {
  StatusCode sc = GaudiTool::initialize();
  if ( !sc ) return sc;

  m_fullExtrapolator = tool<ITrackExtrapolator>( m_fullExtrapolatorName );
  m_veloExtrapolator = tool<ITrackExtrapolator>( m_veloExtrapolatorName );

  warning() << "PVReFitter is no longer maintained and thus depreciated." << endmsg;

  return sc;
}

//=============================================================================
// refit PV
//=============================================================================
StatusCode PVReFitter::reFit( LHCb::VertexBase* PV, IGeometryInfo const& geometry ) const {

  debug() << "Now reFit PV" << endmsg;

  StatusCode sc = StatusCode::SUCCESS;

  if ( !( PV->isPrimary() ) ) {
    debug() << "PVReFitter is used to reFit a  non-PV" << endmsg;
    return StatusCode::FAILURE;
  }

  RecVertex* primvtx = dynamic_cast<RecVertex*>( PV );
  if ( !primvtx ) return StatusCode::FAILURE;

  std::vector<LHCb::Track*> tracks;

  SmartRefVector<LHCb::Track> pvtracks = primvtx->tracks();
  for ( SmartRefVector<LHCb::Track>::iterator trIt = pvtracks.begin(); pvtracks.end() != trIt; trIt++ ) {
    LHCb::Track* track = *trIt;
    tracks.push_back( track );
  }

  sc = fitPV( primvtx, tracks, geometry );
  if ( !sc.isSuccess() ) {
    debug() << "fitPV fails" << endmsg;
    return StatusCode::FAILURE;
  }

  return sc;
}

//=============================================================================
// remove track used for a LHCb::Particle and refit PV
//=============================================================================
StatusCode PVReFitter::remove( const LHCb::Particle* part, LHCb::VertexBase* PV, IGeometryInfo const& geometry ) const {

  debug() << "Now remove and reFit " << endmsg;

  if ( !( PV->isPrimary() ) ) {
    debug() << "PVReFitter is used to remove a trackfrom a  non-PV" << endmsg;
    return StatusCode::FAILURE;
  }

  RecVertex* primvtx = dynamic_cast<RecVertex*>( PV );
  if ( !primvtx ) return StatusCode::FAILURE;

  std::vector<const LHCb::Track*> dautracks;
  getFinalTracks( part, dautracks );

  std::vector<LHCb::Track*>   tracks;
  SmartRefVector<LHCb::Track> pvtracks = primvtx->tracks();
  for ( SmartRefVector<LHCb::Track>::iterator trIt = pvtracks.begin(); pvtracks.end() != trIt; trIt++ ) {
    LHCb::Track* track = *trIt;
    bool         drop  = false;
    for ( std::vector<const LHCb::Track*>::iterator idtrack = dautracks.begin(); dautracks.end() != idtrack;
          idtrack++ ) {
      if ( track == *idtrack ) {
        drop = true;
        break;
      }
    }
    if ( !drop ) tracks.push_back( track );
  }

  StatusCode sc = fitPV( primvtx, tracks, geometry );
  if ( !sc.isSuccess() ) {
    debug() << "fitPV fails" << endmsg;
    return StatusCode::FAILURE;
  }

  return sc;
}

//=============================================================================
// fit PV from a vector of tracks
//=============================================================================
StatusCode PVReFitter::fitPV( LHCb::RecVertex* PV, LHCb::Track::Vector& tracks, IGeometryInfo const& geometry ) const {
  StatusCode sc = StatusCode::SUCCESS;

  debug() << "Now entering fitPV!" << endmsg;

  PV->clearTracks();

  if ( tracks.size() < 2 ) {
    debug() << "number of track left for the PV " << tracks.size() << endmsg;
    return StatusCode::FAILURE;
  }

  LHCb::Track* tr1 = tracks.back();
  tracks.pop_back();
  LHCb::Track* tr2 = tracks.back();
  tracks.pop_back();

  if ( !tr1 || !tr2 ) {
    debug() << "Null track pointer found ! " << endmsg;
    return StatusCode::FAILURE;
  }

  sc = seedPV( PV, tr1, tr2, geometry );
  if ( sc.isFailure() ) {
    debug() << "Fail to seedPV" << endmsg;
    return StatusCode::FAILURE;
  }

  for ( std::vector<LHCb::Track*>::iterator itrack = tracks.begin(); tracks.end() != itrack; itrack++ ) {
    LHCb::Track* tr = *itrack;
    sc              = addTr( PV, tr, geometry );
    if ( sc.isFailure() ) {
      debug() << "Fail to addTr" << endmsg;
      return StatusCode::FAILURE;
    }
  }

  return sc;
}

//=============================================================================
// get final tracks of a particle
//=============================================================================
void PVReFitter::getFinalTracks( const LHCb::Particle* part, LHCb::Track::ConstVector& tracks ) const {
  const LHCb::ProtoParticle* proto = part->proto();

  if ( proto ) {
    const LHCb::Track* thetrack = proto->track();
    if ( thetrack ) tracks.push_back( thetrack );
  } else {
    const SmartRefVector<LHCb::Particle>&          Prods = part->daughters();
    SmartRefVector<LHCb::Particle>::const_iterator iProd;
    for ( iProd = Prods.begin(); iProd != Prods.end(); ++iProd ) {
      const LHCb::Particle* daughter = *iProd;
      getFinalTracks( daughter, tracks );
    }
  }
}

//=============================================================================
// seeding a PV with two tracks
//=============================================================================

StatusCode PVReFitter::seedPV( LHCb::RecVertex* PV, LHCb::Track* tr1, LHCb::Track* tr2,
                               IGeometryInfo const& geometry ) const {
  StatusCode sc = StatusCode::SUCCESS;

  bool isVelo1 = tr1->checkType( Track::Types::Velo );
  bool isVelo2 = tr2->checkType( Track::Types::Velo );

  double chi2Fit = 999.;

  Gaudi::Vector3      V3;
  Gaudi::SymMatrix3x3 Cov3;

  LHCb::State statetr1 = tr1->firstState();
  LHCb::State statetr2 = tr2->firstState();

  double tx1 = statetr1.tx();
  double ty1 = statetr1.ty();
  double x1  = statetr1.x();
  double y1  = statetr1.y();
  double z1  = statetr1.z();

  double tx2 = statetr2.tx();
  double ty2 = statetr2.ty();
  double x2  = statetr2.x();
  double y2  = statetr2.y();
  double z2  = statetr2.z();

  double sumSquaredSlopes = tx1 * tx1 + ty1 * ty1 + tx2 * tx2 + ty2 * ty2;
  double sumCrossedProduct =
      tx1 * ( x1 - tx1 * z1 ) + ty1 * ( y1 - ty1 * z1 ) + tx2 * ( x2 - tx2 * z2 ) + ty2 * ( y2 - ty2 * z2 );

  double sumX = x1 - tx1 * z1 + x2 - tx2 * z2;
  double sumY = y1 - ty1 * z1 + y2 - ty2 * z2;

  double sumSlopeX = tx1 + tx2;
  double sumSlopeY = ty1 + ty2;

  double det = sumSquaredSlopes - ( ( sumSlopeX * sumSlopeX + sumSlopeY * sumSlopeY ) ) / 2.;

  double zEstimate = 0;
  if ( !LHCb::essentiallyZero( det ) ) {
    zEstimate = ( ( ( sumX * sumSlopeX + sumY * sumSlopeY ) / 2. ) - sumCrossedProduct ) / det;
  } else {
    err() << "Unable to make z estimate " << endmsg;
    if ( z1 < z2 )
      zEstimate = z1 - .001;
    else
      zEstimate = z2 - 0.001;
  }

  double zfit         = zEstimate;
  double zPreviousFit = -999999.;

  int iterTransport = 0;

  while ( fabs( zPreviousFit - zfit ) > m_maxDeltaZ && iterTransport < m_maxIter ) {
    zPreviousFit = zfit;
    iterTransport++;

    LHCb::State newstate1 = statetr1;
    if ( isVelo1 )
      sc = m_veloExtrapolator->propagate( newstate1, zPreviousFit, geometry );
    else
      sc = m_fullExtrapolator->propagate( newstate1, zPreviousFit, geometry );
    if ( sc.isFailure() ) {
      debug() << "failed to propagate tarck state in seedPV!" << endmsg;
      return sc;
    }

    LHCb::State newstate2 = statetr2;
    if ( isVelo2 )
      sc = m_veloExtrapolator->propagate( newstate2, zPreviousFit, geometry );
    else
      sc = m_fullExtrapolator->propagate( newstate2, zPreviousFit, geometry );
    if ( sc.isFailure() ) {
      debug() << "failed to propagate tarck state in seedPV!" << endmsg;
      return sc;
    }

    ROOT::Math::SVector<double, 10> X;
    X[0] = newstate1.x();
    X[1] = newstate1.y();
    X[2] = newstate1.momentum().x();
    X[3] = newstate1.momentum().y();
    X[4] = newstate1.momentum().z();
    X[5] = newstate2.x();
    X[6] = newstate2.y();
    X[7] = newstate2.momentum().x();
    X[8] = newstate2.momentum().y();
    X[9] = newstate2.momentum().z();

    Gaudi::SymMatrix6x6 covtr1 = newstate1.posMomCovariance();
    Gaudi::SymMatrix6x6 covtr2 = newstate2.posMomCovariance();
    Gaudi::SymMatrix5x5 newcovtr1;
    Gaudi::SymMatrix5x5 newcovtr2;

    for ( int l1 = 0; l1 <= 4; l1++ ) {
      int n1 = l1;
      if ( n1 >= 2 ) n1++;
      for ( int l2 = 0; l2 <= l1; l2++ ) {
        int n2 = l2;
        if ( n2 >= 2 ) n2++;
        newcovtr1( l1, l2 ) = covtr1( n1, n2 );
        newcovtr2( l1, l2 ) = covtr2( n1, n2 );
      }
    }

    SymMatrix10x10 Cx;
    for ( int l1 = 0; l1 < 5; l1++ ) {
      for ( int l2 = 0; l2 <= l1; l2++ ) {
        Cx( l1, l2 )         = newcovtr1( l1, l2 );
        Cx( l1 + 5, l2 + 5 ) = newcovtr2( l1, l2 );
      }
      for ( int l2 = 0; l2 < 5; l2++ ) Cx( l1 + 5, l2 ) = 0.0;
    }

    //(x1,y1,px1,py1,pz1,x2,y2,px2,py2,pz2) at the same z
    ROOT::Math::SVector<double, 10> vfit = X;
    SymMatrix10x10                  cfit = Cx;

    bool converged = false;
    int  iter      = 0;

    double chi2Previous = 9999.;

    while ( !converged && iter < m_maxIter ) {
      iter++;
      verbose() << ":-) Iteration   " << iter << endmsg;

      // f=(x2-x1)*(py2*pz1-py1*pz2)-(y2-y1)*(px2*pz1-px1*pz2)
      ROOT::Math::SVector<double, 1> f;
      f( 0 ) = ( vfit( 5 ) - vfit( 0 ) ) * ( vfit( 8 ) * vfit( 4 ) - vfit( 3 ) * vfit( 9 ) ) -
               ( vfit( 6 ) - vfit( 1 ) ) * ( vfit( 7 ) * vfit( 4 ) - vfit( 2 ) * vfit( 9 ) );

      verbose() << "constraint values   " << f( 0 ) << endmsg;

      // D is the derivative matrix
      ROOT::Math::SMatrix<double, 1, 10> D;
      D( 0, 0 ) = -( vfit( 8 ) * vfit( 4 ) - vfit( 3 ) * vfit( 9 ) );
      D( 0, 1 ) = ( vfit( 7 ) * vfit( 4 ) - vfit( 2 ) * vfit( 9 ) );
      D( 0, 2 ) = ( vfit( 6 ) - vfit( 1 ) ) * vfit( 9 );
      D( 0, 3 ) = -( vfit( 5 ) - vfit( 0 ) ) * vfit( 9 );
      D( 0, 4 ) = ( vfit( 5 ) - vfit( 0 ) ) * vfit( 8 ) - ( vfit( 6 ) - vfit( 1 ) ) * vfit( 7 );
      D( 0, 5 ) = ( vfit( 8 ) * vfit( 4 ) - vfit( 3 ) * vfit( 9 ) );
      D( 0, 6 ) = -( vfit( 7 ) * vfit( 4 ) - vfit( 2 ) * vfit( 9 ) );
      D( 0, 7 ) = -( vfit( 6 ) - vfit( 1 ) ) * vfit( 4 );
      D( 0, 8 ) = ( vfit( 5 ) - vfit( 0 ) ) * vfit( 4 );
      D( 0, 9 ) = -( vfit( 5 ) - vfit( 0 ) ) * vfit( 3 ) + ( vfit( 6 ) - vfit( 1 ) ) * vfit( 2 );

      ROOT::Math::SVector<double, 1> d = f - D * vfit;

      Gaudi::SymMatrix1x1 VD = Similarity( D, Cx );

      if ( !VD.Invert() ) {
        debug() << "could not invert matrix VD in seedPV! " << endmsg;
        return StatusCode::FAILURE;
      }

      ROOT::Math::SVector<double, 1> alpha = D * X + d;

      ROOT::Math::SVector<double, 1> lambda = VD * alpha;

      ROOT::Math::SMatrix<double, 10, 1> DT = Transpose( D );
      vfit                                  = X - Cx * DT * lambda;

      SymMatrix10x10 delataC1 = Similarity( DT, VD );

      SymMatrix10x10 delataC2 = Similarity( Cx, delataC1 );

      cfit = Cx - delataC2;

      chi2Fit = Dot( alpha, lambda );
      // chi2Fit+= 2*Dot(lambda,f);

      if ( fabs( chi2Fit - chi2Previous ) < m_maxDeltaChi2 ) {
        converged = true;
      } else {
        chi2Previous = chi2Fit;
      }
    } // end chi2 minimization iteration

    if ( !converged ) return StatusCode::FAILURE;

    V3[0] = vfit( 0 ) -
            vfit( 2 ) * ( vfit( 5 ) - vfit( 0 ) ) * vfit( 9 ) / ( vfit( 7 ) * vfit( 4 ) - vfit( 2 ) * vfit( 9 ) );
    V3[1] = vfit( 1 ) -
            vfit( 3 ) * ( vfit( 6 ) - vfit( 1 ) ) * vfit( 9 ) / ( vfit( 8 ) * vfit( 4 ) - vfit( 3 ) * vfit( 9 ) );
    V3[2] = zPreviousFit -
            ( vfit( 5 ) - vfit( 0 ) ) * vfit( 4 ) * vfit( 9 ) / ( vfit( 7 ) * vfit( 4 ) - vfit( 2 ) * vfit( 9 ) );

    ROOT::Math::SMatrix<double, 3, 10> JA;
    JA( 0, 0 ) = 1. + vfit( 2 ) * vfit( 9 ) / ( vfit( 7 ) * vfit( 4 ) - vfit( 2 ) * vfit( 9 ) );
    JA( 0, 2 ) = -( vfit( 5 ) - vfit( 0 ) ) * vfit( 9 ) / ( vfit( 7 ) * vfit( 4 ) - vfit( 2 ) * vfit( 9 ) ) -
                 vfit( 2 ) * ( vfit( 5 ) - vfit( 0 ) ) * vfit( 9 ) / ( vfit( 7 ) * vfit( 4 ) - vfit( 2 ) * vfit( 9 ) ) /
                     ( vfit( 7 ) * vfit( 4 ) - vfit( 2 ) * vfit( 9 ) ) * vfit( 9 );
    JA( 0, 4 ) = vfit( 2 ) * ( vfit( 5 ) - vfit( 0 ) ) * vfit( 9 ) / ( vfit( 7 ) * vfit( 4 ) - vfit( 2 ) * vfit( 9 ) ) /
                 ( vfit( 7 ) * vfit( 4 ) - vfit( 2 ) * vfit( 9 ) ) * vfit( 7 );
    JA( 0, 5 ) = -vfit( 2 ) * vfit( 9 ) / ( vfit( 7 ) * vfit( 4 ) - vfit( 2 ) * vfit( 9 ) );
    JA( 0, 7 ) = vfit( 2 ) * ( vfit( 5 ) - vfit( 0 ) ) * vfit( 9 ) / ( vfit( 7 ) * vfit( 4 ) - vfit( 2 ) * vfit( 9 ) ) /
                 ( vfit( 7 ) * vfit( 4 ) - vfit( 2 ) * vfit( 9 ) ) * vfit( 4 );
    JA( 0, 9 ) = -vfit( 2 ) * ( vfit( 5 ) - vfit( 0 ) ) / ( vfit( 7 ) * vfit( 4 ) - vfit( 2 ) * vfit( 9 ) ) -
                 vfit( 2 ) * ( vfit( 5 ) - vfit( 0 ) ) * vfit( 9 ) / ( vfit( 7 ) * vfit( 4 ) - vfit( 2 ) * vfit( 9 ) ) /
                     ( vfit( 7 ) * vfit( 4 ) - vfit( 2 ) * vfit( 9 ) ) * vfit( 2 );

    JA( 1, 1 ) = 1. + vfit( 3 ) * vfit( 9 ) / ( vfit( 8 ) * vfit( 4 ) - vfit( 3 ) * vfit( 9 ) );
    JA( 1, 3 ) = -( vfit( 6 ) - vfit( 1 ) ) * vfit( 9 ) / ( vfit( 8 ) * vfit( 4 ) - vfit( 3 ) * vfit( 9 ) ) -
                 vfit( 3 ) * ( vfit( 6 ) - vfit( 1 ) ) * vfit( 9 ) / ( vfit( 8 ) * vfit( 4 ) - vfit( 3 ) * vfit( 9 ) ) /
                     ( vfit( 8 ) * vfit( 4 ) - vfit( 3 ) * vfit( 9 ) ) * vfit( 9 );
    JA( 1, 4 ) = vfit( 3 ) * ( vfit( 6 ) - vfit( 1 ) ) * vfit( 9 ) / ( vfit( 8 ) * vfit( 4 ) - vfit( 3 ) * vfit( 9 ) ) /
                 ( vfit( 8 ) * vfit( 4 ) - vfit( 3 ) * vfit( 9 ) ) * vfit( 8 );
    JA( 1, 6 ) = -vfit( 3 ) * vfit( 9 ) / ( vfit( 8 ) * vfit( 4 ) - vfit( 3 ) * vfit( 9 ) );
    JA( 1, 8 ) = vfit( 3 ) * ( vfit( 6 ) - vfit( 1 ) ) * vfit( 9 ) / ( vfit( 8 ) * vfit( 4 ) - vfit( 3 ) * vfit( 9 ) ) /
                 ( vfit( 8 ) * vfit( 4 ) - vfit( 3 ) * vfit( 9 ) ) * vfit( 4 );
    JA( 1, 9 ) = -vfit( 3 ) * ( vfit( 6 ) - vfit( 1 ) ) / ( vfit( 8 ) * vfit( 4 ) - vfit( 3 ) * vfit( 9 ) ) -
                 vfit( 3 ) * ( vfit( 6 ) - vfit( 1 ) ) * vfit( 9 ) / ( vfit( 8 ) * vfit( 4 ) - vfit( 3 ) * vfit( 9 ) ) /
                     ( vfit( 8 ) * vfit( 4 ) - vfit( 3 ) * vfit( 9 ) ) * vfit( 3 );

    JA( 2, 0 ) = vfit( 4 ) * vfit( 9 ) / ( vfit( 7 ) * vfit( 4 ) - vfit( 2 ) * vfit( 9 ) );
    JA( 2, 2 ) = -( vfit( 5 ) - vfit( 0 ) ) * vfit( 4 ) * vfit( 9 ) /
                 ( vfit( 7 ) * vfit( 4 ) - vfit( 2 ) * vfit( 9 ) ) / ( vfit( 7 ) * vfit( 4 ) - vfit( 2 ) * vfit( 9 ) ) *
                 vfit( 9 );
    JA( 2, 4 ) = -( vfit( 5 ) - vfit( 0 ) ) * vfit( 9 ) / ( vfit( 7 ) * vfit( 4 ) - vfit( 2 ) * vfit( 9 ) ) +
                 ( vfit( 5 ) - vfit( 0 ) ) * vfit( 4 ) * vfit( 9 ) / ( vfit( 7 ) * vfit( 4 ) - vfit( 2 ) * vfit( 9 ) ) /
                     ( vfit( 7 ) * vfit( 4 ) - vfit( 2 ) * vfit( 9 ) ) * vfit( 7 );
    JA( 2, 5 ) = -vfit( 4 ) * vfit( 9 ) / ( vfit( 7 ) * vfit( 4 ) - vfit( 2 ) * vfit( 9 ) );
    JA( 2, 7 ) = ( vfit( 5 ) - vfit( 0 ) ) * vfit( 4 ) * vfit( 9 ) / ( vfit( 7 ) * vfit( 4 ) - vfit( 2 ) * vfit( 9 ) ) /
                 ( vfit( 7 ) * vfit( 4 ) - vfit( 2 ) * vfit( 9 ) ) * vfit( 4 );
    JA( 2, 9 ) = -( vfit( 5 ) - vfit( 0 ) ) * vfit( 4 ) / ( vfit( 7 ) * vfit( 4 ) - vfit( 2 ) * vfit( 9 ) ) -
                 ( vfit( 5 ) - vfit( 0 ) ) * vfit( 4 ) * vfit( 9 ) / ( vfit( 7 ) * vfit( 4 ) - vfit( 2 ) * vfit( 9 ) ) /
                     ( vfit( 7 ) * vfit( 4 ) - vfit( 2 ) * vfit( 9 ) ) * vfit( 2 );

    Cov3 = Similarity( JA, cfit );

    zfit = V3[2];
  }

  Gaudi::XYZPoint position( V3[0], V3[1], V3[2] );
  PV->setPosition( position );
  PV->setCovMatrix( Cov3 );

  double chi2new = chi2Fit;
  PV->setChi2( chi2new );

  int nDoFnew = 1;
  PV->setNDoF( nDoFnew );

  PV->addToTracks( tr1 );
  PV->addToTracks( tr2 );

  return sc;
}

//=============================================================================
// add a track to a PV
//=============================================================================

StatusCode PVReFitter::addTr( LHCb::RecVertex* PV, LHCb::Track* tr, IGeometryInfo const& geometry ) const {
  StatusCode sc = StatusCode::SUCCESS;

  bool isVelo = tr->checkType( Track::Types::Velo );

  Gaudi::SymMatrix3x3 covPV = PV->covMatrix();
  Gaudi::XYZPoint     PosPV = PV->position();

  double z2 = PosPV.z();

  LHCb::State statetr  = tr->firstState();
  LHCb::State newstate = statetr;

  if ( isVelo )
    sc = m_veloExtrapolator->propagate( newstate, z2, geometry );
  else
    sc = m_fullExtrapolator->propagate( newstate, z2, geometry );
  if ( sc.isFailure() ) {
    debug() << "failed to propagate tarck state in addTr!" << endmsg;
    return sc;
  }

  ROOT::Math::SVector<double, 8> X;
  X[0] = PosPV.x();
  X[1] = PosPV.y();
  X[2] = PosPV.z();
  X[3] = newstate.x();
  X[4] = newstate.y();
  X[5] = newstate.momentum().x();
  X[6] = newstate.momentum().y();
  X[7] = newstate.momentum().z();

  Gaudi::SymMatrix6x6 covtr = newstate.posMomCovariance();

  SymMatrix8x8 Cx;
  for ( int l1 = 0; l1 < 3; l1++ ) {
    for ( int l2 = 0; l2 <= l1; l2++ ) { Cx( l1, l2 ) = covPV( l1, l2 ); }
  }
  for ( int l1 = 3; l1 < 8; l1++ ) {
    int n1 = l1 - 3;
    if ( n1 >= 2 ) n1++;
    for ( int l2 = 3; l2 <= l1; l2++ ) {
      int n2 = l2 - 3;
      if ( n2 >= 2 ) n2++;
      Cx( l1, l2 ) = covtr( n1, n2 );
    }
  }
  for ( int l1 = 3; l1 < 8; l1++ ) {
    for ( int l2 = 0; l2 < 3; l2++ ) { Cx( l1, l2 ) = 0.0; }
  }

  // xpv,ypv,zpv, and xtr,ytr,pxtr,pytr,pztr at z2
  ROOT::Math::SVector<double, 8> vfit = X;
  SymMatrix8x8                   cfit = Cx;

  double chi2PreviousFit = 9999.;
  double chi2Fit         = 999.;

  bool converged = false;
  int  iter      = 0;

  while ( !converged && iter < m_maxIter ) {
    iter++;
    verbose() << ":-) Iteration   " << iter << endmsg;

    // f(0)=(xtr-xpv)*pztr-(z2-zpv)*pxtr
    // f(1)=(ytr-ypv)*pztr-(z2-zpv)*pytr

    ROOT::Math::SVector<double, 2> f;
    f( 0 ) = ( vfit[3] - vfit[0] ) * vfit[7] - ( z2 - vfit[2] ) * vfit[5];
    f( 1 ) = ( vfit[4] - vfit[1] ) * vfit[7] - ( z2 - vfit[2] ) * vfit[6];

    // D is the derivative matrix
    ROOT::Math::SMatrix<double, 2, 8> D;
    D( 0, 0 ) = -vfit[7];
    D( 0, 1 ) = 0.0;
    D( 0, 2 ) = vfit[5];
    D( 0, 3 ) = vfit[7];
    D( 0, 4 ) = 0.0;
    D( 0, 5 ) = -( z2 - vfit[2] );
    D( 0, 6 ) = 0.0;
    D( 0, 7 ) = ( vfit[3] - vfit[0] );

    D( 1, 0 ) = 0.0;
    D( 1, 1 ) = -vfit[7];
    D( 1, 2 ) = vfit[6];
    D( 1, 3 ) = 0.0;
    D( 1, 4 ) = vfit[7];
    D( 1, 5 ) = 0.0;
    D( 1, 6 ) = -( z2 - vfit[2] );
    D( 1, 7 ) = ( vfit[4] - vfit[1] );

    ROOT::Math::SVector<double, 2> d = f - D * vfit;

    Gaudi::SymMatrix2x2 VD = Similarity( D, Cx );
    if ( !VD.Invert() ) {
      debug() << "could not invert matrix VD in addTr! " << endmsg;
      return StatusCode::FAILURE;
    }

    ROOT::Math::SVector<double, 2> alpha = D * X + d;

    ROOT::Math::SVector<double, 2> lambda = VD * alpha;

    ROOT::Math::SMatrix<double, 8, 2> DT = Transpose( D );

    vfit = X - Cx * DT * lambda;

    SymMatrix8x8 delataC1 = Similarity( DT, VD );

    SymMatrix8x8 delataC2 = Similarity( Cx, delataC1 );

    cfit = Cx - delataC2;

    chi2Fit = Dot( alpha, lambda );
    // chi2Fit+= 2*Dot(lambda,f);

    if ( fabs( chi2Fit - chi2PreviousFit ) < m_maxDeltaChi2 ) {
      converged = true;
    } else {
      chi2PreviousFit = chi2Fit;
    }
  } // end chi2 minimization iteration

  if ( !converged ) return StatusCode::FAILURE;

  Gaudi::XYZPoint     position( vfit[0], vfit[1], vfit[2] );
  Gaudi::SymMatrix3x3 cov = cfit.Sub<Gaudi::SymMatrix3x3>( 0, 0 );

  PV->setPosition( position );
  PV->setCovMatrix( cov );

  double chi2new = chi2Fit + PV->chi2();
  PV->setChi2( chi2new );

  int nDoFnew = 2 + PV->nDoF();
  PV->setNDoF( nDoFnew );

  PV->addToTracks( tr );

  return sc;
}
