#!/usr/bin/env python3 -B

###############################################################################
# (c) Copyright 2000-2020 CERN for the benefit of the LHCb Collaboration      #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################

import glob
import math
import os

import GaussFit
import matplotlib.pyplot as plt
import ROOT
from parse import parse
from scipy.signal import savgol_filter

ROOT.gROOT.SetBatch(True)
ROOT.gStyle.SetPadTickX(1)
ROOT.gStyle.SetPadTickY(1)
ROOT.gStyle.SetOptFit(101)
ROOT.gStyle.SetOptStat(1110)
ROOT.gErrorIgnoreLevel = ROOT.kWarning

plt.style.use("seaborn-whitegrid")

dataSet = "FunctionalCKRes-OnlinePhotonSel"
# dataSet = "ParameterisedCKRes-OnlinePhotonSel"

fname = (
    "data/" + dataSet + "/RichFuture-Feature-x86_64-centos7-gcc9-opt-Expert-Histos.root"
)
print("Processing", fname)
root_file = ROOT.TFile.Open(fname)
if not root_file or root_file.IsZombie():
    quit()

# need to deduce this from root file somehow...
nCKBins = 200

# Scale factors by track type for CK res extracted from RMS
# scaleF_Fit = {"Long": 0.90, "Down": 0.85, "Up": 0.85, "Seed": 0.90}
scaleF_RMS = {
    "Rich1Gas": {"Long": 0.85, "Down": 0.85, "Up": 0.70, "Seed": 0.70},
    "Rich2Gas": {"Long": 0.85, "Down": 0.85, "Up": 0.70, "Seed": 0.70},
}
scaleF_Fit = {
    "Rich1Gas": {"Long": 0.85, "Down": 0.85, "Up": 1.10},
    "Rich2Gas": {"Long": 0.90, "Down": 0.85, "Seed": 1.10},
}


def fitHist(tk, rad, pid, c, h, pdfName, binMean):
    # Set title
    h.SetTitle(
        rad
        + " "
        + tk
        + " "
        + pid
        + " Rec-Exp CKTheta | <Bin> = "
        + str(round(binMean, 7))
    )

    # run the fit
    res = GaussFit.fitHist(h)

    # Save in PDF
    if c:
        c.SaveAs(pdfName)

    # return scaled resolution
    return res.GetParameter(2) * scaleF_Fit[rad][tk]


# Values to use for C++ class
cppUse = "Fitted"
# cppUse = "RMS"

outDir = "Performance/CKT Resolution Parameterisation/" + dataSet
if not os.path.exists(outDir):
    os.makedirs(outDir)
print("Output Dir", outDir)

# Write out resolutions to C++
f_cpp = open(outDir + "/ResolutionParameters.h", mode="w")

f_cpp.write("""
/*****************************************************************************\\
* (c) Copyright 2000-2020 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\\*****************************************************************************/
""")

f_cpp.write("\n// NB - This file is autogenerated. Do not change by hand.")
f_cpp.write("\n//    - Using results from " + cppUse + " CK Theta data.\n\n")
f_cpp.write("#pragma once\n\n")
f_cpp.write("#include <map>\n")
f_cpp.write("#include <string>\n")
f_cpp.write('#include "Kernel/RichParticleIDType.h"\n')
f_cpp.write('#include "Kernel/RichRadiatorType.h"\n\n')

f_cpp.write("namespace {\n")
f_cpp.write("   ///< Type for resolution parameters for a track type\n")
f_cpp.write(
    "   using Params = Rich::RadiatorArray<Rich::ParticleArray<std::map<double,double>>>;\n"
)
f_cpp.write("   ///< Access resolution parameters for given track type\n")
f_cpp.write(
    "   Params resolutionParameters( const std::string& tkType, const std::string& mode ) {\n"
)
f_cpp.write("   std::map<std::string,Params> resParams;\n")

# Track types by RICH radiator
tkTypes = {"Rich1Gas": ["Long", "Down", "Up"], "Rich2Gas": ["Long", "Down", "Seed"]}

# loop over radiators
for rad in ["Rich1Gas", "Rich2Gas"]:
    # Loop over track types
    for tk in tkTypes[rad]:
        # Plot dir
        dir = outDir + "/" + rad + "/" + tk
        if not os.path.exists(dir):
            os.makedirs(dir)
        for pdf in glob.glob(dir + "/*.pdf"):
            os.remove(pdf)

        # Loop over particles
        for pid in ["electron", "muon", "pion", "kaon", "proton"]:
            # Load the expected CK resolution profile
            h_exp_ck = root_file.Get(
                "RICH/RiCKTkMCRes" + tk + "/" + rad + "/" + pid + "/ckresVcktheta"
            )
            h_exp_p = root_file.Get(
                "RICH/RiCKTkMCRes" + tk + "/" + rad + "/" + pid + "/ckresVptot"
            )
            h_exp_lp = root_file.Get(
                "RICH/RiCKTkMCRes" + tk + "/" + rad + "/" + pid + "/ckresVlogptot"
            )

            # Min entries to form an interpolation point
            # muons are rare, so need lower limit
            minEntForInterp = 50000 if pid != "muon" else 10000
            # Min RMS value to use
            minRMSForInterp = 0.000001

            # Values
            ckmeans = []
            ckrmss_ck = []
            ckexps_ck = []
            ckfits_ck = []
            curr_entries = 0
            curr_ckmean = 0
            curr_ckrms = 0
            curr_ckexp = 0
            last_ckmean = 0
            curr_h = None

            # min distance in units of CKT between data points
            ckmean_diff_tol = 0.0004

            # ROOT directory object
            dName = (
                "RICH/RiMCCKResParam" + tk + "/" + rad + "/" + pid + "/CKBins/ckResTrue"
            )

            # Plot name
            name = rad + "-" + tk + "-" + pid

            # ROOT canvas
            # c = ROOT.TCanvas(name+"CKT", name+"CKT", 1400, 1000)
            c = None

            # Open PDF for fits
            pdfName = dir + "/CKT-Fits-" + name + ".pdf"
            if c:
                c.SaveAs(pdfName + "[")

            # loop over bins and extract values
            for iBin in range(0, nCKBins):
                # CK res V ck histo
                h = root_file.Get(dName + "/bin" + str(iBin + 1))
                if h and h.GetEntries() > 0:
                    # Extract expected CK res from profile for this bin
                    exp_ck = h_exp_ck.GetBinContent(iBin)
                    if exp_ck > 0 and h.GetRMS() > minRMSForInterp:
                        # Number of entries in this bin
                        ents = h.GetEntries()

                        # Parse the title to get CK bin min and max values
                        r = parse("{}-{}-{}-{}-{}<{}<{}", h.GetTitle())
                        ckmin = float(r[4])
                        ckmax = float(r[6])

                        # mean CK theta for bin
                        ckmean = 0.5 * (ckmin + ckmax)

                        # Extract RMS for this CK bin
                        ckrms_ck = scaleF_RMS[rad][tk] * h.GetRMS()

                        # Update cached values with this bin
                        curr_ckmean = (
                            (curr_entries * curr_ckmean) + (ents * ckmean)
                        ) / (ents + curr_entries)
                        curr_ckrms = (
                            (curr_entries * curr_ckrms) + (ents * ckrms_ck)
                        ) / (ents + curr_entries)
                        curr_ckexp = ((curr_entries * curr_ckexp) + (ents * exp_ck)) / (
                            ents + curr_entries
                        )
                        curr_entries = ents + curr_entries
                        if not curr_h:
                            curr_h = h.Clone()
                        else:
                            curr_h.Add(h)

                        # If enough entries save to list and reset
                        if (
                            curr_entries > minEntForInterp
                            and abs(last_ckmean - curr_ckmean) > ckmean_diff_tol
                        ):
                            ckfits_ck += [
                                fitHist(tk, rad, pid, c, curr_h, pdfName, curr_ckmean)
                            ]
                            ckmeans += [curr_ckmean]
                            ckrmss_ck += [curr_ckrms]
                            ckexps_ck += [curr_ckexp]
                            last_ckmean = curr_ckmean
                            curr_entries = 0
                            curr_ckmean = 0
                            curr_ckrms = 0
                            curr_ckexp = 0
                            curr_h = None

            # Add last values saved if OK
            if curr_entries > minEntForInterp / 100:
                ckfits_ck += [fitHist(tk, rad, pid, c, curr_h, pdfName, curr_ckmean)]
                ckmeans += [curr_ckmean]
                ckrmss_ck += [curr_ckrms]
                ckexps_ck += [curr_ckexp]

            # Finalise PDF
            if c:
                c.SaveAs(pdfName + "]")

            # Values
            pmeans = []
            ckrmss_p = []
            ckexps_p = []
            ckfits_p = []
            curr_entries = 0
            curr_pmean = 0
            curr_ckrms = 0
            curr_ckexp = 0
            last_pmean = 0
            curr_h = None

            # min distance in units of Ptot between data points
            pmean_diff_tol = 1000

            # ROOT directory name
            dName = (
                "RICH/RiMCCKResParam" + tk + "/" + rad + "/" + pid + "/PBins/ckResTrue"
            )

            # make canvas
            c = ROOT.TCanvas(name + "Ptot", name + "Ptot", 1400, 1000)

            # Open PDF for fits
            pdfName = dir + "/Ptot-fits-" + name + ".pdf"
            if c:
                c.SaveAs(pdfName + "[")

            # loop over bins and extract values
            for iBin in range(0, nCKBins):
                # CK res V p histo path
                h = root_file.Get(dName + "/bin" + str(iBin + 1))
                if h and h.GetEntries() > 0:
                    # Extract expected CK res from profile for this bin
                    exp_ck = h_exp_p.GetBinContent(iBin)
                    if exp_ck > 0 and h.GetRMS() > minRMSForInterp:
                        # Number of entries in this bin
                        ents = h.GetEntries()

                        # Parse the title to get CK bin min and max values (in MeV/c)
                        r = parse("{}-{}-{}-{}-{}<{}<{}", h.GetTitle())
                        pmin = float(r[4]) * 1000
                        pmax = float(r[6]) * 1000

                        # mean P
                        pmean = 0.5 * (pmin + pmax)

                        # Extract RMS for this CK bin
                        ckrms_p = scaleF_RMS[rad][tk] * h.GetRMS()

                        # Update cached values with this bin
                        curr_pmean = ((curr_entries * curr_pmean) + (ents * pmean)) / (
                            ents + curr_entries
                        )
                        curr_ckrms = (
                            (curr_entries * curr_ckrms) + (ents * ckrms_p)
                        ) / (ents + curr_entries)
                        curr_ckexp = ((curr_entries * curr_ckexp) + (ents * exp_ck)) / (
                            ents + curr_entries
                        )
                        curr_entries = ents + curr_entries
                        if not curr_h:
                            curr_h = h.Clone()
                        else:
                            curr_h.Add(h)

                        # If enough entries save to list and reset
                        if (
                            curr_entries > minEntForInterp
                            and abs(last_pmean - curr_pmean) > pmean_diff_tol
                        ):
                            ckfits_p += [
                                fitHist(tk, rad, pid, c, curr_h, pdfName, curr_pmean)
                            ]
                            pmeans += [curr_pmean]
                            ckrmss_p += [curr_ckrms]
                            ckexps_p += [curr_ckexp]
                            last_pmean = curr_pmean
                            curr_entries = 0
                            curr_pmean = 0
                            curr_ckrms = 0
                            curr_ckexp = 0
                            curr_h = None

            # Add last values saved if OK
            if curr_entries > minEntForInterp / 100:
                ckfits_p += [fitHist(tk, rad, pid, c, curr_h, pdfName, curr_pmean)]
                pmeans += [curr_pmean]
                ckrmss_p += [curr_ckrms]
                ckexps_p += [curr_ckexp]

            # Finalise PDF
            if c:
                c.SaveAs(pdfName + "]")

            # Minimum data plots for plots
            minPoints = 4
            # Smoothing ploy order
            ployOrder = 9

            print(
                " ->",
                rad,
                tk,
                pid,
                "# P values",
                len(pmeans),
                "# CKT values",
                len(ckmeans),
            )

            # If enough points plot results and make output text files
            if len(ckmeans) > minPoints:
                # Smooth the RMS values
                # smoothptns = len(ckmeans)
                smoothptns = 45
                if smoothptns > len(ckmeans):
                    smoothptns = len(ckmeans)
                if smoothptns % 2 == 0:
                    smoothptns -= 1
                if ployOrder >= smoothptns:
                    ployOrder = smoothptns - 1
                ckrmss_ck_smooth = savgol_filter(ckrmss_ck, smoothptns, ployOrder)
                ckfits_ck_smooth = savgol_filter(ckfits_ck, smoothptns, ployOrder)

                # Make some plots versus CK theta
                (raw_rms,) = plt.plot(
                    ckmeans,
                    ckrmss_ck,
                    "-ok",
                    color="blue",
                    markersize=2,
                    linewidth=1,
                    label="Raw R.M.S.",
                )
                (smoothed_rms,) = plt.plot(
                    ckmeans,
                    ckrmss_ck_smooth,
                    "-ok",
                    color="red",
                    markersize=2,
                    linewidth=1,
                    label="Smoothed R.M.S.",
                )
                (raw_fit,) = plt.plot(
                    ckmeans,
                    ckfits_ck,
                    "-ok",
                    color="orange",
                    markersize=2,
                    linewidth=1,
                    label="Raw Fitted",
                )
                (smoothed_fit,) = plt.plot(
                    ckmeans,
                    ckfits_ck_smooth,
                    "-ok",
                    color="purple",
                    markersize=2,
                    linewidth=1,
                    label="Smoothed Fitted",
                )
                (expected,) = plt.plot(
                    ckmeans,
                    ckexps_ck,
                    "-ok",
                    color="green",
                    markersize=2,
                    linewidth=1,
                    label="Computed CK Res.",
                )
                # Add some labels
                plt.title(rad + " " + tk + " " + pid + " | CK Theta Resolution")
                plt.xlabel("CK theta / rad")
                plt.ylabel("CK theta resolution / rad")
                plt.legend(
                    handles=[raw_rms, smoothed_rms, raw_fit, smoothed_fit, expected],
                    prop={"size": 7},
                )
                # Save to PDF / PNG
                plt.savefig(dir + "/" + name + "-VCKTheta.pdf")
                plt.close()

                # Write out the resolutions for this radiator and hypo
                f_cpp.write(
                    '   resParams["'
                    + tk
                    + 'CKTheta"][Rich::'
                    + rad
                    + "][Rich::"
                    + pid.capitalize()
                    + "] = {\n"
                )
                for i in range(0, len(ckmeans)):
                    val = (
                        ckfits_ck_smooth[i]
                        if cppUse == "Fitted"
                        else ckrmss_ck_smooth[i]
                    )
                    f_cpp.write(
                        "    { "
                        + str("{:.7f}".format(ckmeans[i]))
                        + ", "
                        + str("{:.7f}".format(round(val, 7)))
                        + " }"
                    )
                    if i < len(ckmeans) - 1:
                        f_cpp.write(",")
                    f_cpp.write("\n")
                f_cpp.write("   };\n")

            # Smoothing ploy order
            ployOrder = 8

            if len(pmeans) > minPoints:
                # Smooth the RMS values
                # smoothptns = len(pmeans)
                smoothptns = 25
                if smoothptns > len(pmeans):
                    smoothptns = len(pmeans)
                if smoothptns % 2 == 0:
                    smoothptns -= 1
                if ployOrder >= smoothptns:
                    ployOrder = smoothptns - 1
                ckrmss_p_smooth = savgol_filter(ckrmss_p, smoothptns, ployOrder)
                ckfits_p_smooth = savgol_filter(ckfits_p, smoothptns, ployOrder)

                # Make some plots versus momentum
                (raw_rms,) = plt.plot(
                    pmeans,
                    ckrmss_p,
                    "-ok",
                    color="blue",
                    markersize=2,
                    linewidth=1,
                    label="Raw R.M.S.",
                )
                (smoothed_rms,) = plt.plot(
                    pmeans,
                    ckrmss_p_smooth,
                    "-ok",
                    color="red",
                    markersize=2,
                    linewidth=1,
                    label="Smoothed R.M.S.",
                )
                (raw_fit,) = plt.plot(
                    pmeans,
                    ckfits_p,
                    "-ok",
                    color="orange",
                    markersize=2,
                    linewidth=1,
                    label="Raw Fitted",
                )
                (smoothed_fit,) = plt.plot(
                    pmeans,
                    ckfits_p_smooth,
                    "-ok",
                    color="purple",
                    markersize=2,
                    linewidth=1,
                    label="Smoothed Fitted",
                )
                (expected,) = plt.plot(
                    pmeans,
                    ckexps_p,
                    "-ok",
                    color="green",
                    markersize=2,
                    linewidth=1,
                    label="Computed CK Res.",
                )
                # Add some labels
                plt.title(rad + " " + tk + " " + pid + " | CK Theta Resolution")
                plt.xlabel("Momentum / MeV/c")
                plt.ylabel("CK theta resolution / rad")
                plt.legend(
                    handles=[raw_rms, smoothed_rms, raw_fit, smoothed_fit, expected],
                    prop={"size": 7},
                )
                # Save to PDF
                plt.savefig(dir + "/" + name + "-VPtot.pdf")
                plt.close()

                # Write out the resolutions for this radiator and hypo
                f_cpp.write(
                    '   resParams["'
                    + tk
                    + 'Momentum"][Rich::'
                    + rad
                    + "][Rich::"
                    + pid.capitalize()
                    + "] = {\n"
                )
                for i in range(0, len(pmeans)):
                    val = (
                        ckfits_p_smooth[i] if cppUse == "Fitted" else ckrmss_p_smooth[i]
                    )
                    f_cpp.write(
                        "    { "
                        + str("{:.2f}".format(pmeans[i]))
                        + ", "
                        + str("{:.7f}".format(round(val, 7)))
                        + " }"
                    )
                    if i < len(pmeans) - 1:
                        f_cpp.write(",")
                    f_cpp.write("\n")
                f_cpp.write("   };\n")

        for mode in ("Momentum", "CKTheta"):
            # Hard code the deuteron type to use the same parameters as proton, the next heaviest
            f_cpp.write(
                '   resParams["'
                + tk
                + mode
                + '"][Rich::'
                + rad
                + '][Rich::Deuteron] = resParams["'
                + tk
                + mode
                + '"][Rich::'
                + rad
                + "][Rich::Proton];\n"
            )
            # Hardcode BelowThreshold type to just use pion values
            f_cpp.write(
                '   resParams["'
                + tk
                + mode
                + '"][Rich::'
                + rad
                + '][Rich::BelowThreshold] = resParams["'
                + tk
                + mode
                + '"][Rich::'
                + rad
                + "][Rich::Pion];\n"
            )

# Finish the c++ method and close
f_cpp.write("  return resParams[tkType+mode];\n")
f_cpp.write("}}\n")
f_cpp.close()
