/*****************************************************************************\
* (c) Copyright 2000-2018 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/

#include "DetDesc/DetectorElement.h"
#include "Event/FitNode.h"
#include "Event/PrFitNode.h"
#include "Event/PrHits.h"
#include "Event/PrKalmanFitResult.h"
#include "Event/State.h"
#include "Event/Track.h"
#include "Gaudi/Accumulators/Histogram.h"
#include "GaudiKernel/Algorithm.h"
#include "GaudiKernel/PhysicalConstants.h"
#include "Kernel/LHCbID.h"
#include "Kernel/UTNames.h"
#include "LHCbAlgs/Consumer.h"
#include "TrackKernel/TrackFunctors.h"
#include "TrackMonitorTupleBase.h"
#include "UTDAQ/UTDAQHelper.h"
#include "UTDAQ/UTInfo.h"
#include <algorithm>
#include <bitset>
#include <cstdint>
#include <iomanip>
#include <string>
#include <vector>

/** @class UTGlobalEffMon UTGlobalEffMon.h
 * Tomasz Skwarnicki
 * Wojciech Krupa (wokrupa@cern.ch)
 */

namespace LHCb::UT {

  class UTGlobalEffMon
      : public Algorithm::Consumer<void( Track::Range const&, DetectorElement const& ),
                                   DetDesc::usesBaseAndConditions<TrackMonitorTupleBase, DetectorElement>> {

  public:
    UTGlobalEffMon( const std::string& name, ISvcLocator* pSvcLocator );
    void                    operator()( Track::Range const& tracks, DetectorElement const& ) const override;
    Gaudi::Property<double> m_low_multip_limit{ this, "LowMultiplicityLimity", 10000 };

  private:
    template <typename TFitResult>
    void monitor( const TFitResult*, LHCb::Track const*, const DetectorElement&, unsigned int ) const;

    mutable Gaudi::Accumulators::Histogram<1> m_h_ntracks{
        this, "h_ntracks", "Number of long track; Number of long tracks;Events", { 1000, 0, m_low_multip_limit } };
    mutable Gaudi::Accumulators::Histogram<1> m_h_nutlay{
        this,
        "h_nutlay",
        "Number of UT layers per long track; Number of UT layers;Number of tracks",
        { 5, -0.5, 4.5 } };
    mutable Gaudi::Accumulators::Histogram<1> m_h_nutlay_high_mult{
        this,
        "h_nutlay_high_mult",
        "Number of UT layers per long track high mult ; Number of UT layers;Number of tracks high mult",
        { 5, -0.5, 4.5 } };
    //    full-width sensor 2D bins
    mutable Gaudi::Accumulators::Histogram<2> m_h_all3{ this,
                                                        "h_all3",
                                                        "All tracks         Y vs X UTbX-midplane;X [mm];Y [mm]",
                                                        { { 18, -860.4, +860.4 }, { 14, -669.2, +669.2 } } };
    mutable Gaudi::Accumulators::Histogram<2> m_h_pass3{ this,
                                                         "h_pass3",
                                                         "UT-matched tracks  Y vs X UTbX-midplane;X [mm];Y [mm]",
                                                         { { 18, -860.4, +860.4 }, { 14, -669.2, +669.2 } } };

    mutable Gaudi::Accumulators::Histogram<2> m_h_pass3_low_mult{
        this,
        "h_pass3_low_mult",
        "UT-matched tracks  Y vs X UTbX-midplane low mult;X [mm];Y [mm]",
        { { 18, -860.4, +860.4 }, { 14, -669.2, +669.2 } } };

    mutable Gaudi::Accumulators::Histogram<1> m_h_residual_inner{
        this,
        "h_residual_inner",
        "UT track residual - inner detector;residual [mm];number of UT clusters",
        { 1000, -0.5, 0.5 } };
    mutable Gaudi::Accumulators::Histogram<1> m_h_residual_outer{
        this,
        "h_residual_outer",
        "UT track residual - outer detector;residual [mm];number of UT clusters",
        { 1000, -0.5, 0.5 } };

    mutable Gaudi::Accumulators::Histogram<1> m_h_residualpull_inner{
        this,
        "h_residualpull_inner",
        "UT track residual pull - inner detector;pull;number of UT clusters",
        { 100, -5, 5 } };
    mutable Gaudi::Accumulators::Histogram<1> m_h_residualpull_outer{
        this,
        "h_residualpull_outer",
        "UT track residual pull - outer detector;pull;number of UT clsuters",
        { 100, -5, 5 } };
  };

  DECLARE_COMPONENT_WITH_ID( UTGlobalEffMon, "UTGlobalEffMon" )

  UTGlobalEffMon::UTGlobalEffMon( const std::string& name, ISvcLocator* pSvcLocator )
      : Consumer(
            name, pSvcLocator,
            { { "TracksInContainer", TrackLocation::Default }, { "StandardGeometryTop", standard_geometry_top } } ) {}

  template <typename TFitResult>
  void UTGlobalEffMon::monitor( const TFitResult* fitResult, LHCb::Track const* track, const DetectorElement& lhcb,
                                unsigned int nTracks ) const {
    auto&        geometry        = *lhcb.geometry();
    const double zutlay3         = 2652.5;
    const double UT_sensor_width = 95.6; // do not change this because you think this is not correct!
    StateVector  aState;
    extrapolator()->propagate( *track, zutlay3, aState, geometry ).ignore();
    if ( aState.x() < -9990 ) { return; } // Skip unphysical tracks

    // tracks must be within UT layer 3 acceptance (exclude also generous region around the beam pipe )
    double tr2utlax3 = aState.x();
    double tr2utlay3 = aState.y();
    if ( fabs( tr2utlax3 ) > 9 * UT_sensor_width ) return;
    if ( fabs( tr2utlay3 ) > 7 * UT_sensor_width ) return;
    if ( sqrt( pow( tr2utlay3, 2 ) + pow( tr2utlax3, 2 ) ) < 45 ) return;

    // PT cut
    try {

      auto cState = track->closestToBeamState();
      if ( cState.pt() < 1000.0 ) return;

    } catch ( GaudiException& e ) { return; }

    // all tracks passing the cuts (for efficiency map denominator)
    ++m_h_all3[{ tr2utlax3, tr2utlay3 }];

    const double kDummy = 1000.0;

    std::array<LHCb::Detector::UT::ChannelID, 4> _tr2ut;
    std::array<double, 4>                        _tr2utr   = { kDummy, kDummy, kDummy, kDummy };
    std::array<double, 4>                        _tr2utchi = { 0.0, 0.0, 0.0, 0.0 };

    for ( const auto& node : nodes( *fitResult ) ) {
      if ( !( node.hasMeasurement() && node.isHitOnTrack() && node.isUT() ) ) continue;
      if ( node.isOutlier() ) continue;
      LHCbID lhcbID = id( node );

      Detector::UT::ChannelID chan = lhcbID.utID();

      double       res    = node.residual();
      double       chi    = res / node.errResidual();
      double       ares   = fabs( res );
      unsigned int _layer = chan.layer();

      // Loop over layers
      if ( ares < fabs( _tr2utr[_layer] ) ) {
        _tr2utr[_layer]   = res;
        _tr2utchi[_layer] = chi;
        _tr2ut[_layer]    = chan;
      }
    } // nodes loop

    unsigned int nlay = 0;
    for ( int layer = 0; layer < 4; ++layer ) {
      if ( _tr2ut[layer] != LHCb::Detector::UT::ChannelID{} ) {
        ++nlay;
        // Extract values from the current layer
        UInt_t _mo   = ( _tr2ut[layer] >> 10 ) & 7;
        UInt_t _st   = ( _tr2ut[layer] >> 14 ) & 15;
        bool   inner = ( _st == 0 ) && ( _mo == 3 || _mo == 4 );
        if ( inner ) {
          ++m_h_residual_inner[_tr2utr[layer]];
          ++m_h_residualpull_inner[_tr2utchi[layer]];
        } else {
          ++m_h_residual_outer[_tr2utr[layer]];
          ++m_h_residualpull_outer[_tr2utchi[layer]];
        }
      }
    }

    // tracks matched to UT hits (for nominator of efficiency map)
    if ( nlay > 0 ) ++m_h_pass3[{ tr2utlax3, tr2utlay3 }];

    // change it for protons 2025
    ++m_h_nutlay_high_mult[nlay];

    if ( nTracks > 4 && nTracks <= 100 ) { // you may want to move these limits to be configurable parameters

      // UT layer counting for efficiency and ghost rate estimates
      ++m_h_nutlay[nlay];
      if ( nlay > 0 ) ++m_h_pass3_low_mult[{ tr2utlax3, tr2utlay3 }];
    }
  }

  void UTGlobalEffMon::operator()( Track::Range const& tracks, const DetectorElement& lhcb ) const {
    unsigned int nTracks = tracks.size();

    if ( nTracks > m_low_multip_limit ) { nTracks = m_low_multip_limit; } // saturation
    ++m_h_ntracks[nTracks];

    for ( const Track* track : tracks ) {
      if ( auto fitResult = dynamic_cast<const TrackFitResult*>( track->fitResult() ); fitResult ) {
        monitor( fitResult, track, lhcb, nTracks );
      } else if ( auto fitResult = dynamic_cast<const PrKalmanFitResult*>( track->fitResult() ); fitResult ) {
        monitor( fitResult, track, lhcb, nTracks );
      }
    }
  }
} // namespace LHCb::UT
