/***************************************************************************** \
* (c) Copyright 2000-2018 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the testandard_deviation of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/

#include "Kernel/IUTReadoutTool.h"
#include "Kernel/UTDAQBoard.h"
#include "Kernel/UTDAQDefinitions.h"
#include "Kernel/UTDAQID.h"
#include "LHCbAlgs/Consumer.h"
#include "UTDAQ/UTCoordinatesMap.h"
#include "UTDAQ/UTDAQHelper.h"
#include "UTDAQ/UTInfo.h"
#include "UTDet/DeUTDetector.h"
#include <Event/RawBank.h>
#include <Event/RawEvent.h>
#include <Kernel/IUTReadoutTool.h>
#include <optional>
#include <string>
#include <unordered_map>
#include <vector>

/** @class UTBSMonitor UTBSMonitor.cpp
 * --------------------------------------------------
 *  Counters for Data Online Monitoring : UTZS+UTErrors (Extra, ZS)
 *  @author Wojciech Krupa (wokrupa@cern.ch)
 *  @date   07/06/2024 / 2025/02
 * --------------------------------------------------
 **/

namespace GA = Gaudi::Accumulators;

namespace LHCb::UT {

  struct GeomCache {
    std::unordered_set<std::string> m_map;
    GeomCache( const LHCb::Detector::UT::DeUT& det ) {
      // Loop over all valid indices from DeUTObject
      for ( unsigned int side = 0; side < LHCb::Detector::UT::detail::DeUTObject::NBSIDE; ++side ) {
        for ( unsigned int layer = 0; layer < LHCb::Detector::UT::detail::DeUTObject::NBHALFLAYER; ++layer ) {
          for ( unsigned int stave = 0; stave < LHCb::Detector::UT::detail::DeUTObject::NBSTAVE; ++stave ) {
            for ( unsigned int face = 0; face < LHCb::Detector::UT::detail::DeUTObject::NBFACE; ++face ) {
              for ( unsigned int module = 0; module < LHCb::Detector::UT::detail::DeUTObject::NBMODULE; ++module ) {
                for ( unsigned int sec = 0; sec < LHCb::Detector::UT::detail::DeUTObject::NBSUBSECTOR; ++sec ) {
                  auto sector = det.getSector( static_cast<LHCb::Detector::UT::DeUT::Side_t>( side ),
                                               static_cast<LHCb::Detector::UT::DeUT::Layer_t>( layer ),
                                               static_cast<LHCb::Detector::UT::DeUT::Stave_t>( stave ),
                                               static_cast<LHCb::Detector::UT::DeUT::Face_t>( face ),
                                               static_cast<LHCb::Detector::UT::DeUT::Module_t>( module ),
                                               static_cast<LHCb::Detector::UT::DeUT::Sector_t>( sec ) );
                  // Store the hardware name
                  m_map.insert( sector.getHardwareName() );
                }
              }
            }
          }
        }
      }
    }
  };

  // Definition of axis
  const GA::Axis<double> AxisADC        = GA::Axis<double>( 64, -32.5, 31.5, "nADC" ); //-32, 31
  const GA::Axis<double> AxisADC_signed = GA::Axis<double>( 32, -0.5, 31.5, "nADC" );  // 0-31
  const GA::Axis<double> AxisDIGIT      = GA::Axis<double>( 101, -0.5, 10000.5, "UTDigits (size of digit container)" );
  const GA::Axis<double> AxisUAChannels = GA::Axis<double>( 268288, 0, 268288, "GlobalChannel_ID" );
  const GA::Axis<double> AxisUCChannels = GA::Axis<double>( 268288, 268288, 536576, "GlobalChannel_ID" );
  const GA::Axis<double> AxisUA_aUChips = GA::Axis<double>( 496, -0.5, 495.5, "GlobalASIC_ID" );
  const GA::Axis<double> AxisUA_aXChips = GA::Axis<double>( 496, 495.5, 991.5, "GlobalASIC_ID" );
  const GA::Axis<double> AxisUA_bVChips = GA::Axis<double>( 552, 991.5, 1543.5, "GlobalASIC_ID" );
  const GA::Axis<double> AxisUA_bXChips = GA::Axis<double>( 552, 1543.5, 2095.5, "GlobalASIC_ID" );
  const GA::Axis<double> AxisUC_aUChips = GA::Axis<double>( 496, 2095.5, 2591.5, "GlobalASIC_ID" );
  const GA::Axis<double> AxisUC_aXChips = GA::Axis<double>( 496, 2591.5, 3087.5, "GlobalASIC_ID" );
  const GA::Axis<double> AxisUC_bVChips = GA::Axis<double>( 552, 3087.5, 3639.5, "GlobalASIC_ID" );
  const GA::Axis<double> AxisUC_bXChips = GA::Axis<double>( 552, 3639.5, 4191.5, "GlobalASIC_ID" );

  class UTBSMonitor : public Algorithm::Consumer<void( UTDigits const&, UTDigits const&, DeUTDetector const& ),
                                                 Algorithm::Traits::usesConditions<DeUTDetector>> {
  public:
    UTBSMonitor( const std::string& name, ISvcLocator* svcloc )
        : Consumer{ name,
                    svcloc,
                    { { "InputData", UTDigitLocation::UTDigits },
                      { "InputErrorData", UTDigitLocation::UTDigits },
                      { "UTLocation", DeUTDetLocation::location() } } } {}
    ToolHandle<IUTReadoutTool> readoutTool{ this, "ReadoutTool", "UTReadoutTool" };

    std::once_flag m_alreadyInitialized;

    StatusCode initialize() override {
      return Consumer::initialize().andThen( [&] {
        // Add condition callback to initialize the histograms on the first event.
        addConditionDerivation( { DeUTDetLocation::location() }, "GeomCache", [this]( const DeUTDetector& det ) {
          const GeomCache geometryinfo( det );
          std::call_once( this->m_alreadyInitialized, [this, &geometryinfo] {
            for ( const std::string& module : geometryinfo.m_map ) {
              for ( unsigned int i = 0; i < 4; i++ ) {
                Utility::map_emplace( m_1d_ADCCounter, module + ".Chip" + std::to_string( i ), this,
                                      module + ".Chip" + std::to_string( i ), { 32, -0.5, 31.5 } );
              }
            }
            info() << "Histograms booked." << endmsg;
          } );
          return geometryinfo;
        } );
      } );
    }

    void operator()( const UTDigits& digitsCont, const UTDigits& digitserrCont,
                     DeUTDetector const& det ) const override {

      // General counters
      const unsigned int nDigits = digitsCont.size() + digitserrCont.size();

      if ( msgLevel( MSG::DEBUG ) ) debug() << "----------  Found nDigits: " << nDigits << endmsg;
      ++m_nDigits[nDigits];

      // -------------------------------------------------------
      // Loop on digits
      // side==1 is Aside (left), side==0 is Cside (right)
      // -------------------------------------------------------
      // Hitmaps plots and ADC

      // Number of hit's counters
      std::vector<unsigned int> channel_counter( 536577, 0 );
      std::vector<unsigned int> asic_counter( 4192, 0 );

      // Helpers for maxADC monitors
      unsigned int prev_channel = 0;
      unsigned int prev_adc     = 0;
      UTDigit*     aDigit_max;

      // UT bank
      for ( const auto& d : digitsCont ) {

        // Where are we?
        Detector::UT::ChannelID channelID = d->channelID();

        std::string  module_name   = det.findSector( channelID ).getHardwareName();
        unsigned int globalASIC_ID = det.findSector( channelID ).getLocalASICID()[d->strip() / 128];

        // Hit counters
        ++channel_counter[globalASIC_ID * 128 + d->strip() % 128];
        ++asic_counter[globalASIC_ID];

        fillHeatPlots( d, det );

        // First event
        if ( prev_channel == 0 ) {
          prev_channel = d->strip();
          prev_adc     = d->depositedCharge();
          aDigit_max   = d;
          continue;
        }
        // If cluster size >1
        if ( d->strip() - prev_channel == 1 )
          if ( d->depositedCharge() > prev_adc ) aDigit_max = d; // if ADC max in claster
        // End of cluster
        if ( d->strip() - prev_channel != 1 ) {
          fillHeatPlots_MAX( aDigit_max, det );
          aDigit_max = d;
        }

        prev_channel = d->strip();
        prev_adc     = d->depositedCharge();
      }

      // UTError bank
      for ( const auto& d : digitserrCont ) {

        Detector::UT::ChannelID channelID     = d->channelID();
        std::string             module_name   = det.findSector( channelID ).getHardwareName();
        unsigned int            globalASIC_ID = det.findSector( channelID ).getLocalASICID()[d->strip() / 128];

        // Hit counters
        ++channel_counter[globalASIC_ID * 128 + d->strip() % 128];
        ++asic_counter[globalASIC_ID];

        fillHeatPlots( d, det );

        // First event
        if ( prev_channel == 0 ) {
          prev_channel = d->strip();
          prev_adc     = d->depositedCharge();
          aDigit_max   = d;
          continue;
        }
        // If cluster size >1
        if ( d->strip() - prev_channel == 1 )
          if ( d->depositedCharge() > prev_adc ) aDigit_max = d; // if ADC max in claster
        // End of cluster
        if ( d->strip() - prev_channel != 1 ) {
          fillHeatPlots_MAX( aDigit_max, det );
          aDigit_max = d;
        }

        prev_channel = d->strip();
        prev_adc     = d->depositedCharge();
      }

      // Channel hit counters (plots)
      for ( size_t i = 0; i < channel_counter.size(); ++i ) {
        if ( i <= 268288 )
          hitRateChannels_UA[i] += channel_counter[i];
        else
          hitRateChannels_UC[i] += channel_counter[i];
      }

      // Chip hit counters (plots)
      for ( size_t i = 0; i < asic_counter.size(); ++i ) {
        if ( i <= 495 ) {
          hitRateChipAverage_UA_aU[i] += asic_counter[i];
          hitRateChipAverage_UA_aU_radial[UTMap.getRadialID( i )] += asic_counter[i];
        } else if ( i <= 991 ) {
          hitRateChipAverage_UA_aX[i] += asic_counter[i];
          hitRateChipAverage_UA_aX_radial[UTMap.getRadialID( i )] += asic_counter[i];
        } else if ( i <= 1543 ) {
          hitRateChipAverage_UA_bV[i] += asic_counter[i];
          hitRateChipAverage_UA_bV_radial[UTMap.getRadialID( i )] += asic_counter[i];
        } else if ( i <= 2095 ) {
          hitRateChipAverage_UA_bX[i] += asic_counter[i];
          hitRateChipAverage_UA_bX_radial[UTMap.getRadialID( i )] += asic_counter[i];
        } else if ( i <= 2591 ) {
          hitRateChipAverage_UC_aU[i] += asic_counter[i];
          hitRateChipAverage_UC_aU_radial[UTMap.getRadialID( i )] += asic_counter[i];
        } else if ( i <= 3087 ) {
          hitRateChipAverage_UC_aX[i] += asic_counter[i];
          hitRateChipAverage_UC_aX_radial[UTMap.getRadialID( i )] += asic_counter[i];
        } else if ( i <= 3639 ) {
          hitRateChipAverage_UC_bV[i] += asic_counter[i];
          hitRateChipAverage_UC_bV_radial[UTMap.getRadialID( i )] += asic_counter[i];
        } else if ( i <= 4191 ) {
          hitRateChipAverage_UC_bX[i] += asic_counter[i];
          hitRateChipAverage_UC_bX_radial[UTMap.getRadialID( i )] += asic_counter[i];
        }
      }
    }

  private:
    UTCoordinatesMap UTMap;

    // General plots
    mutable GA::Histogram<1> m_nDigits{ this, "nDigits", "Total nDigits", AxisDIGIT };
    mutable GA::Histogram<1> m_nADC{ this, "nADC", "Total nADC", AxisADC_signed };

    mutable GA::ProfileHistogram<1> hitRateChannels_UA{ this, "hitRateChannels_UA", "hitRateChannels_UA",
                                                        AxisUAChannels };
    mutable GA::ProfileHistogram<1> hitRateChannels_UC{ this, "hitRateChannels_UC", "hitRateChannels_UC",
                                                        AxisUCChannels };

    mutable GA::ProfileHistogram<1> hitRateChipAverage_UA_aU{ this, "Average_hitRate_UAaU", "Average_hitRate_UAaU",
                                                              AxisUA_aUChips };
    mutable GA::ProfileHistogram<1> hitRateChipAverage_UA_aX{ this, "Average_hitRate_UAaX", "Average_hitRate_UAaX",
                                                              AxisUA_aXChips };
    mutable GA::ProfileHistogram<1> hitRateChipAverage_UA_bV{ this, "Average_hitRate_UAbV", "Average_hitRate_UAbV",
                                                              AxisUA_bVChips };
    mutable GA::ProfileHistogram<1> hitRateChipAverage_UA_bX{ this, "Average_hitRate_UAbX", "Average_AhitRate_UAbX",
                                                              AxisUA_bXChips };
    mutable GA::ProfileHistogram<1> hitRateChipAverage_UC_aU{ this, "Average_hitRate_UCaU", "Average_hitRate_UCaU",
                                                              AxisUC_aUChips };
    mutable GA::ProfileHistogram<1> hitRateChipAverage_UC_aX{ this, "Average_hitRate_UCaX", "Average_hitRate_UCaX",
                                                              AxisUC_aXChips };
    mutable GA::ProfileHistogram<1> hitRateChipAverage_UC_bV{ this, "Average_hitRate_UCbV", "Average_hitRate_UCbV",
                                                              AxisUC_bVChips };
    mutable GA::ProfileHistogram<1> hitRateChipAverage_UC_bX{ this, "Average_hitRate_UCbX", "Average_hitRate_UCbX",
                                                              AxisUC_bXChips };

    // Hitrate counters - chip radial
    mutable GA::ProfileHistogram<1> hitRateChipAverage_UA_aU_radial{ this, "Average_hitRate_UAaU_radial",
                                                                     "Average_hitRate_UAaU_radial", AxisUA_aUChips };
    mutable GA::ProfileHistogram<1> hitRateChipAverage_UA_aX_radial{ this, "Average_hitRate_UAaX_radial",
                                                                     "Average_hitRate_UAaX_radial", AxisUA_aXChips };
    mutable GA::ProfileHistogram<1> hitRateChipAverage_UA_bV_radial{ this, "Average_hitRate_UAbV_radial",
                                                                     "Average_hitRate_UAbV_radial", AxisUA_bVChips };
    mutable GA::ProfileHistogram<1> hitRateChipAverage_UA_bX_radial{ this, "Average_hitRate_UAbX_radial",
                                                                     "Average_AhitRate_UAbX_radial", AxisUA_bXChips };
    mutable GA::ProfileHistogram<1> hitRateChipAverage_UC_aU_radial{ this, "Average_hitRate_UCaU_radial",
                                                                     "Average_hitRate_UCaU_radial", AxisUC_aUChips };
    mutable GA::ProfileHistogram<1> hitRateChipAverage_UC_aX_radial{ this, "Average_hitRate_UCaX_radial",
                                                                     "Average_hitRate_UCaX_radial", AxisUC_aXChips };
    mutable GA::ProfileHistogram<1> hitRateChipAverage_UC_bV_radial{ this, "Average_hitRate_UCbV_radial",
                                                                     "Average_hitRate_UCbV_radial", AxisUC_bVChips };
    mutable GA::ProfileHistogram<1> hitRateChipAverage_UC_bX_radial{ this, "Average_hitRate_UCbX_radial",
                                                                     "Average_hitRate_UCbX_radial", AxisUC_bXChips };

    // Special ADC counters - average ADC
    mutable GA::ProfileHistogram<1> ADCChipAverage_UA_aU{ this, "Average_ADC_UAaU", "Average_ADC_UAaU",
                                                          AxisUA_aUChips };
    mutable GA::ProfileHistogram<1> ADCChipAverage_UA_aX{ this, "Average_ADC_UAaX", "Average_ADC_UAaX",
                                                          AxisUA_aXChips };
    mutable GA::ProfileHistogram<1> ADCChipAverage_UA_bV{ this, "Average_ADC_UAbV", "Average_ADC_UAbV",
                                                          AxisUA_bVChips };
    mutable GA::ProfileHistogram<1> ADCChipAverage_UA_bX{ this, "Average_ADC_UAbX", "Average_ADC_UAbX",
                                                          AxisUA_bXChips };
    mutable GA::ProfileHistogram<1> ADCChipAverage_UC_aU{ this, "Average_ADC_UCaU", "Average_ADC_UCaU",
                                                          AxisUC_aUChips };
    mutable GA::ProfileHistogram<1> ADCChipAverage_UC_aX{ this, "Average_ADC_UCaX", "Average_ADC_UCaX",
                                                          AxisUC_aXChips };
    mutable GA::ProfileHistogram<1> ADCChipAverage_UC_bV{ this, "Average_ADC_UCbV", "Average_ADC_UCbV",
                                                          AxisUC_bVChips };
    mutable GA::ProfileHistogram<1> ADCChipAverage_UC_bX{ this, "Average_ADC_UCbX", "Average_ADC_UCbX",
                                                          AxisUC_bXChips };

    // The container for pedestal per channel
    mutable std::map<std::string, GA::Histogram<1>> m_1d_ADCCounter;

    void fillHeatPlots( const UTDigit* aDigit, DeUTDetector const& det ) const {

      Detector::UT::ChannelID channelID     = aDigit->channelID();
      std::string             module_name   = det.findSector( channelID ).getHardwareName();
      unsigned int            globalASIC_ID = det.findSector( channelID ).getLocalASICID()[aDigit->strip() / 128];

      if ( globalASIC_ID <= 495 )
        ADCChipAverage_UA_aU[globalASIC_ID] += aDigit->depositedCharge();
      else if ( globalASIC_ID <= 991 )
        ADCChipAverage_UA_aX[globalASIC_ID] += aDigit->depositedCharge();
      else if ( globalASIC_ID <= 1543 )
        ADCChipAverage_UA_bV[globalASIC_ID] += aDigit->depositedCharge();
      else if ( globalASIC_ID <= 2095 )
        ADCChipAverage_UA_bX[globalASIC_ID] += aDigit->depositedCharge();
      else if ( globalASIC_ID <= 2591 )
        ADCChipAverage_UC_aU[globalASIC_ID] += aDigit->depositedCharge();
      else if ( globalASIC_ID <= 3087 )
        ADCChipAverage_UC_aX[globalASIC_ID] += aDigit->depositedCharge();
      else if ( globalASIC_ID <= 3639 )
        ADCChipAverage_UC_bV[globalASIC_ID] += aDigit->depositedCharge();
      else if ( globalASIC_ID <= 4191 )
        ADCChipAverage_UC_bX[globalASIC_ID] += aDigit->depositedCharge();

      // ADC counter
      ++m_nADC[aDigit->depositedCharge()];
    }

    void fillHeatPlots_MAX( const UTDigit* aDigit, DeUTDetector const& det ) const {
      // Let's find where we are
      Detector::UT::ChannelID channelID = aDigit->channelID();
#ifdef USE_DD4HEP
      std::string module_name = det.findSector( channelID ).getHardwareName();
#else
      std::string module_name = det.findSector( channelID )->getHardwareName();
#endif
      ++m_1d_ADCCounter.at( module_name + ".Chip" +
                            std::to_string( aDigit->strip() / 128 ) )[aDigit->depositedCharge()];
    }
  }; // End of class UTBSMonitor
  DECLARE_COMPONENT( UTBSMonitor )
} // namespace LHCb::UT
