/***************************************************************************** \
* (c) Copyright 2000-2018 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/

#include "Kernel/IUTReadoutTool.h"
#include "Kernel/UTDAQBoard.h"
#include "Kernel/UTDAQDefinitions.h"
#include "Kernel/UTDAQID.h"
#include "LHCbAlgs/Consumer.h"
#include "UTDAQ/UTCoordinatesMap.h"
#include "UTDAQ/UTDAQHelper.h"
#include "UTDAQ/UTInfo.h"
#include "UTDet/DeUTDetector.h"
#include <Event/ODIN.h>
#include <Event/RawBank.h>
#include <Event/RawEvent.h>
#include <Kernel/IUTReadoutTool.h>
#include <optional>
#include <string>

/** @class UTDigitsMonitor UTDigitsMonitor.cpp
 * --------------------------------------------------
 *  Counters for Data Online Monitoring : UTZS + UTError (ZS)
 *  @author Wojciech Krupa (wokrupa@cern.ch)
 *  @date   2024/03/20 / 2025/02
 * --------------------------------------------------
 **/

namespace GA = Gaudi::Accumulators;

namespace LHCb::UT {

  // Definition of axis
  const GA::Axis<double> AxisADC        = GA::Axis<double>( 64, -32.5, 31.5, "nADC" );       //-32, 31
  const GA::Axis<double> AxisADC_signed = GA::Axis<double>( 32, -0.5, 31.5, "nADC" );        // 0-31
  const GA::Axis<double> AxisBCID       = GA::Axis<double>( 3565, -0.5, 3564.5, "BunchID" ); //
  const GA::Axis<double> AxisDIGIT;

  // Definition of histogram names
  const std::array<std::string, 4> s_HistNames = { "UTA_Tell_40_vs_ASIC_Flow0", "UTA_Tell_40_vs_ASIC_Flow1",
                                                   "UTC_Tell_40_vs_ASIC_Flow0", "UTC_Tell_40_vs_ASIC_Flow1" };

  const std::array<std::string, 6> s_HistNames_BXType = { "nADC", "nADC_A", "nADC_B", "nADC_C", "nADC_D" };

  const std::array<std::string, 6> s_HistNames_BXType_labels = {
      "Total nADC", "Total nADC - type A sensors", "Total nADC - type B sensors", "Total nADC - type C sensors",
      "Total nADC - type D sensors" };

  const std::array<std::string, 15> s_HistNames_EventType = { "NoBias",
                                                              "Lumi",
                                                              "Isolated",
                                                              "Leading",
                                                              "Trailing",
                                                              "Empty before isolated",
                                                              "Empty after isolated",
                                                              "Empty before leading",
                                                              "Empty after trailing",
                                                              "Empty after 19 clock from trailling",
                                                              "NoBias bb",
                                                              "Lumi bb",
                                                              "Isolated bb",
                                                              "Leading bb",
                                                              "Trailing bb" };

  const std::array<std::string, 15> s_HistNames_EventType_labels = { "NoBias",
                                                                     "Lumi",
                                                                     "Isolated - ADC",
                                                                     "Leading - ADC",
                                                                     "Trailing - ADC",
                                                                     "Empty before isolated - ADC",
                                                                     "Empty after isolated - ADC",
                                                                     "Empty before leading - ADC",
                                                                     "Empty after trailing -  ADC",
                                                                     "Empty after 19 clock from trailling - ADC",
                                                                     "NoBias bb - ADC",
                                                                     "Lumi bb - ADC",
                                                                     "Isolated bb - ADC",
                                                                     "Leading bb - ADC",
                                                                     "Trailing bb - ADC" };

  /** ------------------------------------------------------------------------------------------------------------------------ */

  class UTDigitsMonitor
      : public Algorithm::Consumer<void( UTDigits const&, UTDigits const&, DeUTDetector const&, ODIN const& ),
                                   Algorithm::Traits::usesConditions<DeUTDetector>> {
  public:
    UTDigitsMonitor( const std::string& name, ISvcLocator* svcloc )
        : Consumer{ name,
                    svcloc,
                    { { "InputData", UTDigitLocation::UTDigits },
                      { "InputErrorData", UTDigitLocation::UTDigits },
                      { "UTLocation", DeUTDetLocation::location() },
                      { "ODINLocation", ODINLocation::Default } } } {}
    ToolHandle<IUTReadoutTool> readoutTool{ this, "ReadoutTool", "UTReadoutTool" };

    Gaudi::Property<unsigned int> m_events{ this, "EventsAxis", 10000 };

    void operator()( const UTDigits& digitsCont, const UTDigits& errdigitsCont, DeUTDetector const& det,
                     ODIN const& odin ) const override {

      // General counters
      const unsigned int nDigits = digitsCont.size() + errdigitsCont.size();
      const unsigned int nBuncId = odin.bunchId();

      if ( msgLevel( MSG::DEBUG ) ) debug() << "----------  Found nDigits: " << nDigits << endmsg;
      ++m_nDigits[nDigits];
      ++m_nBuncId[nBuncId];
      ++m_nDigitsnBuncId[{ nBuncId, nDigits }];

      switch ( odin.bunchCrossingType() ) {
      case ODIN::BXTypes::BeamCrossing:
        ++m_nDigits_bb[nDigits];
        break;
      case ODIN::BXTypes::Beam1:
        ++m_nDigits_be[nDigits];
        break;
      case ODIN::BXTypes::Beam2:
        ++m_nDigits_eb[nDigits];
        break;
      case ODIN::BXTypes::NoBeam:
        ++m_nDigits_ee[nDigits];
        break;
      }

      // -------------------------------------------------------
      // Loop on digits
      // side==1 is Aside (left), side==0 is Cside (right)
      // -------------------------------------------------------
      // Hitmaps plots and ADC

      // For (A/C) side histograms
      unsigned int ua_size = 0;
      unsigned int uc_size = 0;

      // For max ADC plots
      unsigned int   prev_channel = 0;
      unsigned int   prev_adc     = 0;
      const UTDigit* aDigit_max   = nullptr;

      // UT bank
      for ( const auto& d : digitsCont ) {

        if ( d->face() == 0 )
          uc_size++;
        else if ( d->face() == 1 )
          ua_size++;

        fillHeatMap( d, det );
        fillADC( d, odin, det );

        // Cluster max ADC counters
        // First event
        if ( prev_channel == 0 ) {
          prev_channel = d->strip();
          prev_adc     = d->depositedCharge();
          aDigit_max   = d;
        }
        // If cluster size >1
        if ( d->strip() - prev_channel == 1 )
          if ( d->depositedCharge() > prev_adc ) aDigit_max = d; // if ADC max in claster
        // End of cluster let's save maxADC
        if ( d->strip() - prev_channel != 1 ) {
          fillADC_clusterMAX( aDigit_max, det );
          aDigit_max = d;
        }

        prev_channel = d->strip();
        prev_adc     = d->depositedCharge();
      }

      // UT error bank
      for ( const auto& d : errdigitsCont ) {

        if ( d->face() == 0 )
          uc_size++;
        else if ( d->face() == 1 )
          ua_size++;

        fillHeatMap( d, det );
        fillADC( d, odin, det );

        // Cluster max ADC counters
        // First event
        if ( prev_channel == 0 ) {
          prev_channel = d->strip();
          prev_adc     = d->depositedCharge();
          aDigit_max   = d;
        }
        // If cluster size >1
        if ( d->strip() - prev_channel == 1 )
          if ( d->depositedCharge() > prev_adc ) aDigit_max = d; // if ADC max in claster
        // End of cluster let's save maxADC
        if ( d->strip() - prev_channel != 1 ) {
          fillADC_clusterMAX( aDigit_max, det );
          aDigit_max = d;
        }

        prev_channel = d->strip();
        prev_adc     = d->depositedCharge();
      }

      // Per side (A/C) histograms
      ++m_nDigits_UA[ua_size];
      ++m_nDigits_UC[uc_size];
    }

  private:
    UTCoordinatesMap UTMap;

    // General plots
    mutable GA::Histogram<1> m_nBuncId{ this, "nBuncId", "Bunch IDs", AxisBCID };
    mutable GA::Histogram<2> m_nDigitsnBuncId{
        this,
        "nDigitsnBuncId",
        "nBuncId vs BunchIDs",
        AxisBCID,
        { m_events / 100, -0.5, m_events + 0.5, "UTDigits (size of digit container)" } };

    // Digits counters
    mutable GA::Histogram<1> m_nDigits{
        this,
        "nDigits",
        "Total nDigits",
        { m_events / 100, -0.5, m_events + 0.5, "UTDigits (size of digit container)" } };
    mutable GA::Histogram<1> m_nDigits_UA{
        this,
        "nDigits_UA",
        "Total nDigits UA",
        { m_events / 100, -0.5, m_events + 0.5, "UTDigits (size of digit container)" } };
    mutable GA::Histogram<1> m_nDigits_UC{
        this,
        "nDigits_UC",
        "Total nDigits UC",
        { m_events / 100, -0.5, m_events + 0.5, "UTDigits (size of digit container)" } };

    // Split for event type
    mutable GA::Histogram<1> m_nDigits_bb{
        this,
        "nDigits_bb",
        "Total nDigits bb",
        { m_events / 100, -0.5, m_events + 0.5, "UTDigits (size of digit container)" } };
    mutable GA::Histogram<1> m_nDigits_be{
        this,
        "nDigits_be",
        "Total nDigits be",
        { m_events / 100, -0.5, m_events + 0.5, "UTDigits (size of digit container)" } };
    mutable GA::Histogram<1> m_nDigits_eb{
        this,
        "nDigits_eb",
        "Total nDigits eb",
        { m_events / 100, -0.5, m_events + 0.5, "UTDigits (size of digit container)" } };
    mutable GA::Histogram<1> m_nDigits_ee{
        this,
        "nDigits_ee",
        "Total nDigits ee",
        { m_events / 100, -0.5, m_events + 0.5, "UTDigits (size of digit container)" } };

    // ADC counters
    mutable GA::Histogram<1> m_nADC{ this, "nADC", "Total nADC", AxisADC_signed };
    mutable GA::Histogram<1> m_nADC_A{ this, "nADC_A", "Total nADC - type A sensors", AxisADC_signed };
    mutable GA::Histogram<1> m_nADC_B{ this, "nADC_B", "Total nADC - type B sensors", AxisADC_signed };
    mutable GA::Histogram<1> m_nADC_C{ this, "nADC_C", "Total nADC - type C sensors", AxisADC_signed };
    mutable GA::Histogram<1> m_nADC_D{ this, "nADC_D", "Total nADC - type D sensors", AxisADC_signed };

    // Cluster max counters
    mutable GA::Histogram<1> m_nADC_clusterMAX{ this, "nADC_clusterMAX", "Total nADC", AxisADC_signed };
    mutable GA::Histogram<1> m_nADC_A_clusterMAX{ this, "nADC_A_clusterMAX", "Total nADC - type A sensors",
                                                  AxisADC_signed };
    mutable GA::Histogram<1> m_nADC_B_clusterMAX{ this, "nADC_B_clusterMAX", "Total nADC - type B sensors",
                                                  AxisADC_signed };
    mutable GA::Histogram<1> m_nADC_C_clusterMAX{ this, "nADC_C_clusterMAX", "Total nADC - type C sensors",
                                                  AxisADC_signed };
    mutable GA::Histogram<1> m_nADC_D_clusterMAX{ this, "nADC_D_clusterMAX", "Total nADC - type D sensors",
                                                  AxisADC_signed };

    // Split for A/C side (comissioning)
    mutable GA::HistogramArray<GA::Histogram<1>, 5> m_1d_ua{
        this,
        []( int i ) { return fmt::format( "{}", "UA_" + s_HistNames_BXType[i] ); },
        []( int i ) { return fmt::format( "{}", "UA " + s_HistNames_BXType_labels[i] ); },
        { 32, -0.5, 31.5, "nADC" } };

    mutable GA::HistogramArray<GA::Histogram<1>, 5> m_1d_uc{
        this,
        []( int i ) { return fmt::format( "{}", "UC_" + s_HistNames_BXType[i] ); },
        []( int i ) { return fmt::format( "{}", "UC " + s_HistNames_BXType_labels[i] ); },
        { 32, -0.5, 31.5, "nADC" } };

    mutable GA::HistogramArray<GA::Histogram<1>, 5> m_1d_bb{
        this,
        []( int i ) { return fmt::format( "{}_bb", s_HistNames_BXType[i] ); },
        []( int i ) { return fmt::format( "{}_bb", s_HistNames_BXType_labels[i] ); },
        { 32, -0.5, 31.5, "nADC" } };

    mutable GA::HistogramArray<GA::Histogram<1>, 5> m_1d_be{
        this,
        []( int i ) { return fmt::format( "{}_be", s_HistNames_BXType[i] ); },
        []( int i ) { return fmt::format( "{}_be", s_HistNames_BXType_labels[i] ); },
        { 32, -0.5, 31.5, "nADC" } };

    mutable GA::HistogramArray<GA::Histogram<1>, 5> m_1d_eb{
        this,
        []( int i ) { return fmt::format( "{}_eb", s_HistNames_BXType[i] ); },
        []( int i ) { return fmt::format( "{}_eb", s_HistNames_BXType_labels[i] ); },
        { 32, -0.5, 31.5, "nADC" } };

    mutable GA::HistogramArray<GA::Histogram<1>, 5> m_1d_ee{
        this,
        []( int i ) { return fmt::format( "{}_ee", s_HistNames_BXType[i] ); },
        []( int i ) { return fmt::format( "{}_ee", s_HistNames_BXType_labels[i] ); },
        { 32, -0.5, 31.5, "nADC" } };

    mutable GA::HistogramArray<GA::Histogram<1>, 15> m_1d_type{
        this,
        []( int i ) { return fmt::format( "{}", s_HistNames_EventType[i] ); },
        []( int i ) { return fmt::format( "{}", s_HistNames_EventType_labels[i] ); },
        { 32, -0.5, 31.5, "nADC" } };

    mutable GA::Histogram<2> m_nADCvsBCID{ this, "nADCvsBuncId", "nADC vs BuncId", AxisBCID, AxisADC_signed };

    mutable GA::HistogramArray<GA::Histogram<2>, 4> m_2d_hitmap_sectors{
        this,
        []( int i ) { return fmt::format( "UT_HitMap_Sectors_Layer{}", i ); },
        []( int i ) { return fmt::format( "Sectors of Layer{} (ASICs)", i ); },
        { 36, -9, 9, "Staves (ASICs)" },
        { 28, -7, 7, "Modules" } };

    // UT Hitmaps per asic physical position
    mutable GA::HistogramArray<GA::Histogram<2>, 4> m_2d_hitmap_chips{
        this,
        []( int i ) { return fmt::format( "UT_HitMap_Chips_Layer{}", i ); },
        []( int i ) { return fmt::format( "ASICs of Layer{}", i ); },
        { 144, -9, 9, "Staves (ASICs)" },
        { 28, -7, 7, "Modules" } };

    // TELL40 plots
    mutable GA::HistogramArray<GA::Histogram<2>, s_HistNames.size()> m_UT_Tell_40_vs_asic{
        this,
        []( int i ) { return fmt::format( "{}", s_HistNames[i] ); },
        []( int i ) { return fmt::format( "{}", s_HistNames[i] ); },
        { 54, 0, 54 },
        { 24, 0, 24 } };

    // Fill ADC plots
    void fillADC( const UTDigit* aDigit, ODIN const& odin, DeUTDetector const& det ) const {

      ++m_nADCvsBCID[{ odin.bunchId(), aDigit->depositedCharge() }];

      // Let's find where we are
      Detector::UT::ChannelID channelID = aDigit->channelID();
      std::string             type      = det.findSector( channelID ).getType();

      if ( odin.eventTypeBit( ODIN::EventTypes::NoBias ) ) ++m_1d_type[0][aDigit->depositedCharge()];
      if ( odin.eventTypeBit( ODIN::EventTypes::Lumi ) ) ++m_1d_type[1][aDigit->depositedCharge()];
      if ( odin.eventTypeBit( ODIN::EventTypes::et_bit_08 ) ) ++m_1d_type[2][aDigit->depositedCharge()];
      if ( odin.eventTypeBit( ODIN::EventTypes::et_bit_09 ) ) ++m_1d_type[3][aDigit->depositedCharge()];
      if ( odin.eventTypeBit( ODIN::EventTypes::et_bit_10 ) ) ++m_1d_type[4][aDigit->depositedCharge()];
      if ( odin.eventTypeBit( ODIN::EventTypes::et_bit_11 ) ) ++m_1d_type[5][aDigit->depositedCharge()];
      if ( odin.eventTypeBit( ODIN::EventTypes::et_bit_12 ) ) ++m_1d_type[6][aDigit->depositedCharge()];
      if ( odin.eventTypeBit( ODIN::EventTypes::et_bit_13 ) ) ++m_1d_type[7][aDigit->depositedCharge()];
      if ( odin.eventTypeBit( ODIN::EventTypes::et_bit_14 ) ) ++m_1d_type[8][aDigit->depositedCharge()];
      if ( odin.eventTypeBit( ODIN::EventTypes::et_bit_15 ) ) ++m_1d_type[9][aDigit->depositedCharge()];

      if ( odin.bunchCrossingType() == ODIN::BXTypes::BeamCrossing ) {
        if ( odin.eventTypeBit( ODIN::EventTypes::NoBias ) ) ++m_1d_type[10][aDigit->depositedCharge()];
        if ( odin.eventTypeBit( ODIN::EventTypes::Lumi ) ) ++m_1d_type[11][aDigit->depositedCharge()];
        if ( odin.eventTypeBit( ODIN::EventTypes::et_bit_08 ) ) ++m_1d_type[12][aDigit->depositedCharge()];
        if ( odin.eventTypeBit( ODIN::EventTypes::et_bit_09 ) ) ++m_1d_type[13][aDigit->depositedCharge()];
        if ( odin.eventTypeBit( ODIN::EventTypes::et_bit_10 ) ) ++m_1d_type[14][aDigit->depositedCharge()];
      }

      ++m_nADC[aDigit->depositedCharge()];
      if ( aDigit->face() == 0 ) {
        ++m_1d_uc[0][aDigit->depositedCharge()];

        if ( type == "A" ) {
          ++m_1d_uc[1][aDigit->depositedCharge()];
        } else if ( type == "B" ) {
          ++m_1d_uc[2][aDigit->depositedCharge()];
        } else if ( type == "C" ) {
          ++m_1d_uc[3][aDigit->depositedCharge()];
        } else if ( type == "D" ) {
          ++m_1d_uc[4][aDigit->depositedCharge()];
        }

      } else if ( aDigit->face() == 1 ) {
        ++m_1d_ua[0][aDigit->depositedCharge()];
        if ( type == "A" ) {
          ++m_1d_ua[1][aDigit->depositedCharge()];
        } else if ( type == "B" ) {
          ++m_1d_ua[2][aDigit->depositedCharge()];
        } else if ( type == "C" ) {
          ++m_1d_ua[3][aDigit->depositedCharge()];
        } else if ( type == "D" ) {
          ++m_1d_ua[4][aDigit->depositedCharge()];
        }
      }

      if ( odin.bunchCrossingType() == ODIN::BXTypes::BeamCrossing )
        ++m_1d_bb[0][aDigit->depositedCharge()];
      else if ( odin.bunchCrossingType() == ODIN::BXTypes::Beam1 )
        ++m_1d_be[0][aDigit->depositedCharge()];
      else if ( odin.bunchCrossingType() == ODIN::BXTypes::Beam2 )
        ++m_1d_eb[0][aDigit->depositedCharge()];
      else if ( odin.bunchCrossingType() == ODIN::BXTypes::NoBeam )
        ++m_1d_ee[0][aDigit->depositedCharge()];

      if ( type == "A" ) {
        ++m_nADC_A[aDigit->depositedCharge()];
        if ( odin.bunchCrossingType() == ODIN::BXTypes::BeamCrossing )
          ++m_1d_bb[1][aDigit->depositedCharge()];
        else if ( odin.bunchCrossingType() == ODIN::BXTypes::Beam1 )
          ++m_1d_be[1][aDigit->depositedCharge()];
        else if ( odin.bunchCrossingType() == ODIN::BXTypes::Beam2 )
          ++m_1d_eb[1][aDigit->depositedCharge()];
        else if ( odin.bunchCrossingType() == ODIN::BXTypes::NoBeam )
          ++m_1d_ee[1][aDigit->depositedCharge()];
      } else if ( type == "B" ) {
        ++m_nADC_B[aDigit->depositedCharge()];
        if ( odin.bunchCrossingType() == ODIN::BXTypes::BeamCrossing )
          ++m_1d_bb[2][aDigit->depositedCharge()];
        else if ( odin.bunchCrossingType() == ODIN::BXTypes::Beam1 )
          ++m_1d_be[2][aDigit->depositedCharge()];
        else if ( odin.bunchCrossingType() == ODIN::BXTypes::Beam2 )
          ++m_1d_eb[2][aDigit->depositedCharge()];
        else if ( odin.bunchCrossingType() == ODIN::BXTypes::NoBeam )
          ++m_1d_ee[2][aDigit->depositedCharge()];
      } else if ( type == "C" ) {
        ++m_nADC_C[aDigit->depositedCharge()];
        if ( odin.bunchCrossingType() == ODIN::BXTypes::BeamCrossing )
          ++m_1d_bb[3][aDigit->depositedCharge()];
        else if ( odin.bunchCrossingType() == ODIN::BXTypes::Beam1 )
          ++m_1d_be[3][aDigit->depositedCharge()];
        else if ( odin.bunchCrossingType() == ODIN::BXTypes::Beam2 )
          ++m_1d_eb[3][aDigit->depositedCharge()];
        else if ( odin.bunchCrossingType() == ODIN::BXTypes::NoBeam )
          ++m_1d_ee[3][aDigit->depositedCharge()];
      } else if ( type == "D" ) {
        ++m_nADC_D[aDigit->depositedCharge()];
        if ( odin.bunchCrossingType() == ODIN::BXTypes::BeamCrossing )
          ++m_1d_bb[4][aDigit->depositedCharge()];
        else if ( odin.bunchCrossingType() == ODIN::BXTypes::Beam1 )
          ++m_1d_be[4][aDigit->depositedCharge()];
        else if ( odin.bunchCrossingType() == ODIN::BXTypes::Beam2 )
          ++m_1d_eb[4][aDigit->depositedCharge()];
        else if ( odin.bunchCrossingType() == ODIN::BXTypes::NoBeam )
          ++m_1d_ee[4][aDigit->depositedCharge()];
      }

      UTDAQID      daqID     = readoutTool->channelIDToDAQID( channelID );
      auto         board_ID  = daqID.board();
      unsigned int sourceID  = UTMap.getSourceIDfromBoardNumber( board_ID );
      int          SourceID_ = sourceID / 1000;

      Flavour     flavour = UTMap.stringToFlavour( det.findSector( channelID ).getFlavour() );
      std::string name    = det.findSector( channelID ).getHardwareName();

      unsigned int asic_times_lane = int( aDigit->strip() / 128 ) + 4 * extract<masks::lane>( aDigit->getdaqID() );

      // Sadly we (I) can't do it better. But at least its ok!
      if ( SourceID_ == 10 ) {
        if ( flavour == Flavour::F2x4 ) {
          ++m_UT_Tell_40_vs_asic[0][{ UTMap.getTELL40Bin( sourceID ), UTMap.get2_4_Flow0_Bin( asic_times_lane ) }];
        } else if ( flavour == Flavour::F2x5 ) {
          ++m_UT_Tell_40_vs_asic[0][{ UTMap.getTELL40Bin( sourceID ), UTMap.get2_5_Flow0_Bin( asic_times_lane ) }];
        } else if ( flavour == Flavour::F4x3 || flavour == Flavour::F2x3 || flavour == Flavour::FJx3 ) {
          ++m_UT_Tell_40_vs_asic[0][{ UTMap.getTELL40Bin( sourceID ), asic_times_lane }];
        }
      } else if ( SourceID_ == 11 ) {
        if ( flavour == Flavour::F2x4 ) {
          if ( asic_times_lane == 16 || asic_times_lane == 17 )
            ++m_UT_Tell_40_vs_asic[0][{ UTMap.getTELL40Bin( sourceID ), UTMap.get2_4_Flow0_Bin( asic_times_lane ) }];
          else
            ++m_UT_Tell_40_vs_asic[1][{ UTMap.getTELL40Bin( sourceID ), UTMap.get2_4_Flow1_Bin( asic_times_lane ) }];
        } else if ( flavour == Flavour::F2x5 ) {
          ++m_UT_Tell_40_vs_asic[1][{ UTMap.getTELL40Bin( sourceID ), UTMap.get2_5_Flow1_Bin( asic_times_lane ) }];
        } else if ( flavour == Flavour::F4x3 || flavour == Flavour::F2x3 || flavour == Flavour::FJx3 )
          ++m_UT_Tell_40_vs_asic[1][{ UTMap.getTELL40Bin( sourceID ), asic_times_lane }];
      }
      if ( SourceID_ == 12 ) {
        if ( flavour == Flavour::F2x4 ) {
          ++m_UT_Tell_40_vs_asic[2][{ UTMap.getTELL40Bin( sourceID ), UTMap.get2_4_Flow0_Bin( asic_times_lane ) }];
        } else if ( flavour == Flavour::F2x5 ) {
          ++m_UT_Tell_40_vs_asic[2][{ UTMap.getTELL40Bin( sourceID ), UTMap.get2_5_Flow0_Bin( asic_times_lane ) }];
        } else if ( flavour == Flavour::F4x3 || flavour == Flavour::F2x3 || flavour == Flavour::FJx3 ) {
          ++m_UT_Tell_40_vs_asic[2][{ UTMap.getTELL40Bin( sourceID ), asic_times_lane }];
        }
      }

      else if ( SourceID_ == 13 ) {
        if ( flavour == Flavour::F2x4 ) {
          if ( asic_times_lane == 16 || asic_times_lane == 17 )
            ++m_UT_Tell_40_vs_asic[2][{ UTMap.getTELL40Bin( sourceID ), UTMap.get2_4_Flow0_Bin( asic_times_lane ) }];
          else
            ++m_UT_Tell_40_vs_asic[3][{ UTMap.getTELL40Bin( sourceID ), UTMap.get2_4_Flow1_Bin( asic_times_lane ) }];
        } else if ( flavour == Flavour::F2x5 ) {
          ++m_UT_Tell_40_vs_asic[3][{ UTMap.getTELL40Bin( sourceID ), UTMap.get2_5_Flow1_Bin( asic_times_lane ) }];
        } else if ( flavour == Flavour::F4x3 || flavour == Flavour::F2x3 || flavour == Flavour::FJx3 )
          ++m_UT_Tell_40_vs_asic[3][{ UTMap.getTELL40Bin( sourceID ), asic_times_lane }];
      }
    }

    // Fill cluster max ADC counters
    void fillADC_clusterMAX( const UTDigit* aDigit, DeUTDetector const& det ) const {

      // Let's find where we are
      Detector::UT::ChannelID channelID = aDigit->channelID();
      std::string             type      = det.findSector( channelID ).getType();

      ++m_nADC_clusterMAX[aDigit->depositedCharge()];

      if ( type == "A" ) {
        ++m_nADC_A_clusterMAX[aDigit->depositedCharge()];
      } else if ( type == "B" ) {
        ++m_nADC_B_clusterMAX[aDigit->depositedCharge()];
      } else if ( type == "C" ) {
        ++m_nADC_C_clusterMAX[aDigit->depositedCharge()];
      } else if ( type == "D" ) {
        ++m_nADC_D_clusterMAX[aDigit->depositedCharge()];
      }
    }

    void fillHeatMap( const UTDigit* aDigit, DeUTDetector const& det ) const {

      // Let's find where we are
      Detector::UT::ChannelID channelID   = aDigit->channelID();
      std::array<double, 4>   Coordinates = det.findSector( channelID ).getMonitoringCoordinates();

      float x       = Coordinates[0];
      float y       = Coordinates[1];
      auto  chanID  = aDigit->channelID();
      auto  aSector = det.findSector( chanID );

// This may be useful but allows to get only position of hybrind. It can be used to get ASIC position.
#ifdef USE_DD4HEP
      auto trajStrip0   = aSector.createTraj( 0, 0 );
      auto trajStrip511 = aSector.createTraj( 511, 0 );
#else
      auto trajStrip0   = aSector->trajectory( chanID, 0 );
      auto trajStrip511 = aSector->trajectory( Detector::UT::ChannelID{ chanID + 511 }, 0 );
#endif
      double stripx0    = -1 * ( trajStrip0.endPoint().x() + trajStrip0.beginPoint().x() ) * 0.5;
      double stripx511  = -1 * ( trajStrip511.endPoint().x() + trajStrip511.beginPoint().x() ) * 0.5;
      double stripwidth = stripx511 - stripx0;
      int    chip;

      if ( stripwidth > 0 )
        chip = int( aDigit->strip() / 128 );
      else
        chip = 3 - int( aDigit->strip() / 128 );

      double positionasic_one = positionasic[chip][0];
      double positionasic_two = positionasic[chip][1];
      double positionasic_thr = positionasic[chip][2];

      // Sectors Maps //
      if ( ( aDigit->stave() > 1 ) || ( fabs( y ) > 2 ) ) {
        ++m_2d_hitmap_sectors[aDigit->layer()][{ x + 0.25, y + 0.25 }];
        ++m_2d_hitmap_sectors[aDigit->layer()][{ x + 0.25, y - 0.25 }];
        ++m_2d_hitmap_sectors[aDigit->layer()][{ x - 0.25, y + 0.25 }];
        ++m_2d_hitmap_sectors[aDigit->layer()][{ x - 0.25, y - 0.25 }];
      }
      if ( ( ( aDigit->stave() == 1 ) && ( fabs( y ) < 2 ) ) || ( ( aDigit->stave() == 0 ) && ( fabs( y ) == 1.5 ) ) ) {
        ++m_2d_hitmap_sectors[aDigit->layer()][{ x, y + 0.25 }];
        ++m_2d_hitmap_sectors[aDigit->layer()][{ x, y - 0.25 }];
      }
      if ( ( aDigit->stave() == 0 ) && ( fabs( y ) < 1 ) ) { ++m_2d_hitmap_sectors[aDigit->layer()][{ x, y }]; }

      // Chips Maps //
      if ( ( aDigit->stave() > 1 ) || ( fabs( y ) > 2 ) ) {
        ++m_2d_hitmap_chips[aDigit->layer()][{ x + positionasic_one, y + 0.25 }];
        ++m_2d_hitmap_chips[aDigit->layer()][{ x + positionasic_one, y - 0.25 }];
        ++m_2d_hitmap_chips[aDigit->layer()][{ x + positionasic_two, y + 0.25 }];
        ++m_2d_hitmap_chips[aDigit->layer()][{ x + positionasic_two, y - 0.25 }];
      }
      if ( ( ( aDigit->stave() == 1 ) && ( fabs( y ) < 2 ) ) || ( ( aDigit->stave() == 0 ) && ( fabs( y ) == 1.5 ) ) ) {
        ++m_2d_hitmap_chips[aDigit->layer()][{ x + positionasic_thr, y + 0.25 }];
        ++m_2d_hitmap_chips[aDigit->layer()][{ x + positionasic_thr, y - 0.25 }];
      }
      if ( ( aDigit->stave() == 0 ) && ( fabs( y ) < 1 ) ) {
        ++m_2d_hitmap_chips[aDigit->layer()][{ x + positionasic_thr, y }];
      }
    }
  }; // End of class UTDigitsMonitor
  DECLARE_COMPONENT( UTDigitsMonitor )
} // namespace LHCb::UT
