/*****************************************************************************\
* (c) Copyright 2000-2018 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/
#include "Event/UTErrorASIC.h"
#include "Kernel/IUTReadoutTool.h"
#include "LHCbAlgs/Consumer.h"
#include "UTDAQ/UTCoordinatesMap.h"
#include "UTDAQ/UTDAQHelper.h"
#include "UTDet/DeUTDetector.h"
#include <Kernel/UTDAQID.h>

/** @class UTErrorMonitor UTErrorMonitor.h
 *  --------------------------------------------------
 *  Counters for Data Online Monitoring : UTError and UTSpecial banks
 *  @author Wojciech Krupa (wokrupa@cern.ch) based on code by:
 *  @author C. Hadjivasiliou
 *  @date   2024/03/20 / 2025/02
 *  --------------------------------------------------
 */

// Definition of axis
namespace GA = Gaudi::Accumulators;

namespace LHCb::UT {

  const int UT      = 18;
  const int UTError = 5;

  // Definition of histogram names
  const std::array<std::string, 4> s_HistBanksNames = { "UTA_BankType_vs_TELL40_Flow0", "UTA_BankType_vs_TELL40_Flow1",
                                                        "UTC_BankType_vs_TELL40_Flow0",
                                                        "UTC_BankType_vs_TELL40_Flow1" };

  const std::array<std::string, 4> s_HistBanksNames_error = {
      "UTA_BankType_vs_TELL40_Error_Flow0", "UTA_BankType_vs_TELL40_Error_Flow1", "UTC_BankType_vs_TELL40_Error_Flow0",
      "UTC_BankType_vs_TELL40_Error_Flow1" };

  const std::array<std::string, 4> s_HistBanksNames_ut = {
      "UTA_UT_BankType_vs_TELL40_Flow0", "UTA_UT_BankType_vs_TELL40_Flow1", "UTC_UT_BankType_vs_TELL40_Flow0",
      "UTC_UT_BankType_vs_TELL40_Flow1" };
  // UTError Types
  const std::array<std::string, 12> s_SpecialSubtypeMap = {
      "UTA_BusyEvent_vs_TELL40_Flow0",   "UTA_BusyEvent_vs_TELL40_Flow1",   "UTC_BusyEvent_vs_TELL40_Flow0",
      "UTC_BusyEvent_vs_TELL40_Flow1",   "UTA_BufferFull_vs_TELL40_Flow0",  "UTA_BufferFull_vs_TELL40_Flow1",
      "UTC_BufferFull_vs_TELL40_Flow0",  "UTC_BufferFull_vs_TELL40_Flow1",  "UTA_BufferFullN_vs_TELL40_Flow0",
      "UTA_BufferFullN_vs_TELL40_Flow1", "UTC_BufferFullN_vs_TELL40_Flow0", "UTC_BufferFullN_vs_TELL40_Flow1" };

  using namespace LHCb;

  class UTErrorMonitor : public Algorithm::Consumer<void( UTErrorASICs const&, DeUTDetector const& ),
                                                    Algorithm::Traits::usesConditions<DeUTDetector>> {

  public:
    /// constructer
    UTErrorMonitor( const std::string& name, ISvcLocator* svcloc )
        : Consumer{
              name,
              svcloc,
              { { "InputData", UTErrorASICLocation::UTErrorASICs }, { "UTLocation", DeUTDetLocation::location() } } } {}

    ToolHandle<IUTReadoutTool> readoutTool{ this, "ReadoutTool", "UTReadoutTool" };

    // UTCoordinations map
    UTCoordinatesMap UTMap;

    // TELL40 vs banktype plots
    mutable GA::HistogramArray<GA::Histogram<2>, s_HistBanksNames.size()> m_UT_Tell_40_vs_banktype{
        this,
        []( int i ) { return fmt::format( "{}", s_HistBanksNames[i] ); },
        []( int i ) { return fmt::format( "{}", s_HistBanksNames[i] ); },
        { 54, 0, 54 },
        { 19, 0, 19 } };

    // TELL40 vs banktype plots
    mutable GA::HistogramArray<GA::Histogram<2>, s_HistBanksNames_error.size()> m_UT_Tell_40_vs_banktype_error{
        this,
        []( int i ) { return fmt::format( "{}", s_HistBanksNames_error[i] ); },
        []( int i ) { return fmt::format( "{}", s_HistBanksNames_error[i] ); },
        { 54, 0, 54 },
        { 3, 0, 3 } };

    // TELL40 vs #IT+UTError events plots
    mutable GA::HistogramArray<GA::Histogram<1>, s_HistBanksNames_ut.size()> m_UT_Tell_40_vs_UT_banktype{
        this,
        []( int i ) { return fmt::format( "{}", s_HistBanksNames_ut[i] ); },
        []( int i ) { return fmt::format( "{}", s_HistBanksNames_ut[i] ); },
        { 54, 0, 54 } };

    // TELL40 vs packetID (UTError) plots
    mutable GA::HistogramArray<GA::Histogram<2>, s_SpecialSubtypeMap.size()> m_2d_packetype{
        this,
        []( int i ) { return fmt::format( "{}", s_SpecialSubtypeMap[i] ); },
        []( int i ) { return fmt::format( "{}", s_SpecialSubtypeMap[i] ); },
        { 54, 0, 54 },
        { 24, 0, 24 } };

    // Efficiency plots
    mutable GA::HistogramArray<GA::Histogram<2>, 4> m_2d_hitmap_tell40_eff{
        this,
        []( int i ) { return fmt::format( "UT_HitMap_TELL40_Eff_Layer{}", i ); },
        []( int i ) { return fmt::format( "TELL40 of Layer{} (ASICs)", i ); },
        { 36, -9, 9, "Staves (ASICs)" },
        { 28, -7, 7, "Modules" } };

    // 2D plot for geometrical pos of UTErrorKnown
    mutable GA::HistogramArray<GA::Histogram<2>, 4> m_asics_UTError_Known{
        this,
        []( int i ) { return fmt::format( "UTError_Known_Layer{}", i ); },
        []( int i ) { return fmt::format( "UTError_Known_Layer{}", i ); },
        { 144, -9, 9, "Staves (ASICs)" },
        { 28, -7, 7, "Modules" } };

    // Error flavour counters
    mutable GA::Histogram<1> m_nErrorsTypeCounter{ this, "ErrorsTypeCounter", "ErrorsTypeCounter", { 19, 0, 19 } };
    // Number of ASICs in error per event - subtype 19
    mutable GA::Histogram<1> m_ASICs_nErrorsType19{
        this, "Number_ASICs_Error_19", "Number of ASICs in Error - subtype 19", { 1001, -0.5, 1000.5 } };
    // Number of ASICs in error per event - subtype 20
    mutable GA::Histogram<1> m_ASICs_nErrorsType20{
        this, "Number_ASICs_Error_20", "Number of ASICs in Error - subtype 20", { 1001, -0.5, 1000.5 } };
    // Number of ASICs in error per event - subtype 21
    mutable GA::Histogram<1> m_ASICs_nErrorsType21{
        this, "Number_ASICs_Error_21", "Number of ASICs in Error - subtype 21", { 1001, -0.5, 1000.5 } };

    mutable GA::Histogram<1> m_errorCounter{ this, "nErrors", "nErrors", { 10000, -0.5, 9999 } };

    /// execute
    void operator()( const UTErrorASICs& errorsCont, DeUTDetector const& det ) const override {
      // number of errors
      unsigned int asic_inError_19 = 0;
      unsigned int asic_inError_20 = 0;
      unsigned int asic_inError_21 = 0;

      ++m_errorCounter[errorsCont.size()];

      // histos per bank flavour
      // histos per error
      for ( const auto& e : errorsCont ) {

        // Handling UTError enocded in strip 1
        if ( e->strip() % 128 == 1 && e->packetId() == 19 ) asic_inError_19++;
        if ( e->strip() % 128 == 1 && e->packetId() == 20 ) asic_inError_20++;
        if ( e->strip() % 128 == 1 && e->packetId() == 21 ) asic_inError_21++;

        fillBankHistograms( e, det );
        fillHeatMap_Error( e, det );
      }

      // Bank flavour counter
      if ( asic_inError_19 > 1000 ) asic_inError_19 = 1000; // saturation
      if ( asic_inError_20 > 1000 ) asic_inError_20 = 1000; // saturation
      if ( asic_inError_21 > 1000 ) asic_inError_21 = 1000; // saturation

      ++m_ASICs_nErrorsType19[asic_inError_19];
      ++m_ASICs_nErrorsType20[asic_inError_20];
      ++m_ASICs_nErrorsType21[asic_inError_21];
    }

  private:
    void fillBankHistograms( const UTErrorASIC* aError, DeUTDetector const& det ) const {
      // Channel and source ID
      auto         chanID          = aError->channelID();
      UTDAQID      daqID           = readoutTool->channelIDToDAQID( chanID );
      auto         board_ID        = daqID.board();
      unsigned int sourceID        = UTMap.getSourceIDfromBoardNumber( board_ID );
      int          bankType        = aError->bankType();
      auto         LaneID          = extract<masks::lane>( aError->getdaqID() );
      int          SourceID_       = sourceID / 1000;
      unsigned int asic_times_lane = aError->strip() / 128 + 4 * LaneID;
      Flavour      flavour         = UTMap.stringToFlavour( det.findSector( chanID ).getFlavour() );

      // Bank flavour counter
      ++m_nErrorsTypeCounter[bankType];

      // Sadly we (I) can't do it better. But at least its works!
      // Normal ErrorType
      if ( ( aError->strip() == 0 && LaneID == 0 ) ||
           ( flavour == Flavour::F2x4 && aError->strip() == 256 && LaneID == 0 ) ) {
        // TELL40 vs Bank flavour
        if ( sourceID / 1000 == 10 ) { // 10346-10269
          ++m_UT_Tell_40_vs_banktype[0][{ UTMap.getTELL40Bin( sourceID ), bankType }];
          if ( bankType == UT || bankType == UTError )
            ++m_UT_Tell_40_vs_UT_banktype[0][{ UTMap.getTELL40Bin( sourceID ) }];
        } else if ( sourceID / 1000 == 11 ) { // 11370-11293
          ++m_UT_Tell_40_vs_banktype[1][{ UTMap.getTELL40Bin( sourceID ), bankType }];
          if ( bankType == UT || bankType == UTError )
            ++m_UT_Tell_40_vs_UT_banktype[1][{ UTMap.getTELL40Bin( sourceID ) }];
        } else if ( sourceID / 1000 == 12 ) { // 12366-12291
          ++m_UT_Tell_40_vs_banktype[2][{ UTMap.getTELL40Bin( sourceID ), bankType }];
          if ( bankType == UT || bankType == UTError )
            ++m_UT_Tell_40_vs_UT_banktype[2][{ UTMap.getTELL40Bin( sourceID ) }];
        } else if ( sourceID / 1000 == 13 ) { // 13371-13315
          ++m_UT_Tell_40_vs_banktype[3][{ UTMap.getTELL40Bin( sourceID ), bankType }];
          if ( bankType == UT || bankType == UTError )
            ++m_UT_Tell_40_vs_UT_banktype[3][{ UTMap.getTELL40Bin( sourceID ) }];
        }

        if ( bankType == UT || bankType == UTError ) {
          for ( const auto& channelID : readoutTool->findBySourceID( LHCb::UTDAQ::boardIDfromSourceID( sourceID ) ) ) {
            if ( channelID % 512 ) continue; // skip subsectors from ReadoutConf map
            std::array<double, 4> Coordinates = det.findSector( channelID ).getMonitoringCoordinates();

            float x = Coordinates[0];
            float y = Coordinates[1];

            if ( ( channelID.stave() > 1 ) || ( fabs( y ) > 2 ) ) {
              ++m_2d_hitmap_tell40_eff[channelID.layer()][{ x + 0.25, y + 0.25 }];
              ++m_2d_hitmap_tell40_eff[channelID.layer()][{ x + 0.25, y - 0.25 }];
              ++m_2d_hitmap_tell40_eff[channelID.layer()][{ x - 0.25, y + 0.25 }];
              ++m_2d_hitmap_tell40_eff[channelID.layer()][{ x - 0.25, y - 0.25 }];
            }
            if ( ( ( channelID.stave() == 1 ) && ( fabs( y ) < 2 ) ) ||
                 ( ( channelID.stave() == 0 ) && ( fabs( y ) == 1.5 ) ) ) {
              ++m_2d_hitmap_tell40_eff[channelID.layer()][{ x, y + 0.25 }];
              ++m_2d_hitmap_tell40_eff[channelID.layer()][{ x, y - 0.25 }];
            }
            if ( ( channelID.stave() == 0 ) && ( fabs( y ) < 1 ) ) {
              ++m_2d_hitmap_tell40_eff[channelID.layer()][{ x, y }];
            }
          }
        }
      }
      // This comes from the logic of UT-Error decoder
      if ( aError->strip() % 128 == 1 &&
           ( aError->packetId() == 19 || aError->packetId() == 20 || aError->packetId() == 21 ) ) {
        // TELL40 vs Bank flavour
        if ( sourceID / 1000 == 10 ) // 10346-10269
          ++m_UT_Tell_40_vs_banktype_error[0][{ UTMap.getTELL40Bin( sourceID ), aError->packetId() - 19 }];
        else if ( sourceID / 1000 == 11 ) // 11370-11293
          ++m_UT_Tell_40_vs_banktype_error[1][{ UTMap.getTELL40Bin( sourceID ), aError->packetId() - 19 }];
        else if ( sourceID / 1000 == 12 ) // 12366-12291
          ++m_UT_Tell_40_vs_banktype_error[2][{ UTMap.getTELL40Bin( sourceID ), aError->packetId() - 19 }];
        else if ( sourceID / 1000 == 13 ) // 13371-13315
          ++m_UT_Tell_40_vs_banktype_error[3][{ UTMap.getTELL40Bin( sourceID ), aError->packetId() - 19 }];

        // Suberror plots
        if ( SourceID_ == 10 ) {
          if ( flavour == Flavour::F2x4 ) {
            ++m_2d_packetype[( aError->packetId() - 19 ) * 4]
                            [{ UTMap.getTELL40Bin( sourceID ), UTMap.get2_4_Flow0_Bin( asic_times_lane ) }];
          } else if ( flavour == Flavour::F2x5 ) {
            ++m_2d_packetype[( aError->packetId() - 19 ) * 4]
                            [{ UTMap.getTELL40Bin( sourceID ), UTMap.get2_5_Flow0_Bin( asic_times_lane ) }];
          } else if ( flavour == Flavour::F4x3 || flavour == Flavour::F2x3 || flavour == Flavour::FJx3 ) {
            ++m_2d_packetype[( aError->packetId() - 19 ) * 4][{ UTMap.getTELL40Bin( sourceID ), asic_times_lane }];
          }
        } else if ( SourceID_ == 11 ) {
          if ( flavour == Flavour::F2x4 ) {
            if ( asic_times_lane == 16 || asic_times_lane == 17 )
              ++m_2d_packetype[( aError->packetId() - 19 ) * 4]
                              [{ UTMap.getTELL40Bin( sourceID ), UTMap.get2_4_Flow0_Bin( asic_times_lane ) }];
            else
              ++m_2d_packetype[( aError->packetId() - 19 ) * 4 + 1]
                              [{ UTMap.getTELL40Bin( sourceID ), UTMap.get2_4_Flow1_Bin( asic_times_lane ) }];
          } else if ( flavour == Flavour::F2x5 )
            ++m_2d_packetype[( aError->packetId() - 19 ) * 4 + 1]
                            [{ UTMap.getTELL40Bin( sourceID ), UTMap.get2_5_Flow1_Bin( asic_times_lane ) }];
          else if ( flavour == Flavour::F4x3 || flavour == Flavour::F2x3 || flavour == Flavour::FJx3 )
            ++m_2d_packetype[( aError->packetId() - 19 ) * 4 + 1][{ UTMap.getTELL40Bin( sourceID ), asic_times_lane }];
        } else if ( SourceID_ == 12 ) {
          if ( flavour == Flavour::F2x4 ) {
            ++m_2d_packetype[( aError->packetId() - 19 ) * 4 + 2]
                            [{ UTMap.getTELL40Bin( sourceID ), UTMap.get2_4_Flow0_Bin( asic_times_lane ) }];
          } else if ( flavour == Flavour::F2x5 ) {
            ++m_2d_packetype[( aError->packetId() - 19 ) * 4 + +2]
                            [{ UTMap.getTELL40Bin( sourceID ), UTMap.get2_5_Flow0_Bin( asic_times_lane ) }];
          } else if ( flavour == Flavour::F4x3 || flavour == Flavour::F2x3 || flavour == Flavour::FJx3 ) {
            ++m_2d_packetype[( aError->packetId() - 19 ) * 4 + 2][{ UTMap.getTELL40Bin( sourceID ), asic_times_lane }];
          }
        } else if ( SourceID_ == 13 ) {
          if ( flavour == Flavour::F2x4 ) {
            if ( asic_times_lane == 16 || asic_times_lane == 17 )
              ++m_2d_packetype[( aError->packetId() - 19 ) * 4 + 2]
                              [{ UTMap.getTELL40Bin( sourceID ), UTMap.get2_4_Flow0_Bin( asic_times_lane ) }];
            else
              ++m_2d_packetype[( aError->packetId() - 19 ) * 4 + 3]
                              [{ UTMap.getTELL40Bin( sourceID ), UTMap.get2_4_Flow1_Bin( asic_times_lane ) }];
          } else if ( flavour == Flavour::F2x5 )
            ++m_2d_packetype[( aError->packetId() - 19 ) * 4 + 3]
                            [{ UTMap.getTELL40Bin( sourceID ), UTMap.get2_5_Flow1_Bin( asic_times_lane ) }];
          else if ( flavour == Flavour::F4x3 || flavour == Flavour::F2x3 || flavour == Flavour::FJx3 )
            ++m_2d_packetype[( aError->packetId() - 19 ) * 4 + 3][{ UTMap.getTELL40Bin( sourceID ), asic_times_lane }];
        }
      }
    }

    // Increment 2D histograms
    void incrementHistogram( GA::Histogram<2>& histogram, double x, double y, double positionasic_one,
                             double positionasic_two ) const {
      ++histogram[{ x + positionasic_one, y + 0.25 }];
      ++histogram[{ x + positionasic_two, y + 0.25 }];
      ++histogram[{ x + positionasic_one, y - 0.25 }];
      ++histogram[{ x + positionasic_two, y - 0.25 }];
    }

    // Fill error counters (2D)
    void fillHeatMap_Error( const UTErrorASIC* aError, DeUTDetector const& det ) const {

      // Let's find out where we are in the UT, it should be reoptimised in the future
      // To be checked if it works properly!

      if ( aError->strip() % 128 == 1 ) {

        Detector::UT::ChannelID channelID = aError->channelID();

        std::string           module_name = det.findSector( channelID ).getHardwareName();
        std::array<double, 4> Coordinates = det.findSector( channelID ).getMonitoringCoordinates();

        float x            = Coordinates[0];
        float y            = Coordinates[1];
        auto  chanID       = aError->channelID();
        auto  aSector      = det.findSector( chanID );
        auto  trajStrip0   = aSector.createTraj( 0, 0 );
        auto  trajStrip511 = aSector.createTraj( 511, 0 );

        double stripx0    = -1 * ( trajStrip0.endPoint().x() + trajStrip0.beginPoint().x() ) * 0.5;
        double stripx511  = -1 * ( trajStrip511.endPoint().x() + trajStrip511.beginPoint().x() ) * 0.5;
        double stripwidth = stripx511 - stripx0;
        int    chip;

        if ( stripwidth > 0 )
          chip = int( aError->strip() / 128 );
        else
          chip = 3 - int( aError->strip() / 128 );

        double positionasic_one = positionasic[chip][0];
        double positionasic_two = positionasic[chip][1];
        double positionasic_thr = positionasic[chip][2];

        // Chip is in its correct geometrical order
        // Sector map //
        if ( ( aError->stave() > 1 ) || ( fabs( y ) > 2 ) ) {
          assert( aError->layer() <= 3 );
          incrementHistogram( m_asics_UTError_Known[aError->layer()], x, y, positionasic_one, positionasic_two );
        }

        if ( ( ( aError->stave() == 1 ) && ( fabs( y ) < 2 ) ) ||
             ( ( aError->stave() == 0 ) && ( fabs( y ) == 1.5 ) ) ) {
          ++m_asics_UTError_Known[aError->layer() == 0][{ x + positionasic_thr, y + 0.25 }];
          ++m_asics_UTError_Known[aError->layer() == 0][{ x + positionasic_thr, y - 0.25 }];
        }
        if ( ( aError->stave() == 0 ) && ( fabs( y ) < 1 ) ) {
          ++m_asics_UTError_Known[aError->layer()][{ x + positionasic_thr, y }];
        }
      }
    }
  };
  DECLARE_COMPONENT( UTErrorMonitor )
} // namespace LHCb::UT
