/***************************************************************************** \
* (c) Copyright 2000-2018 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the teRMS_ADC of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/

#include "Kernel/IUTReadoutTool.h"
#include "Kernel/UTDAQBoard.h"
#include "Kernel/UTDAQDefinitions.h"
#include "Kernel/UTDAQID.h"
#include "LHCbAlgs/Consumer.h"
#include "UTDAQ/UTCoordinatesMap.h"
#include "UTDAQ/UTDAQHelper.h"
#include "UTDAQ/UTInfo.h"
#include "UTDet/DeUTDetector.h"
#include <Event/ODIN.h>
#include <Event/RawBank.h>
#include <Event/RawEvent.h>
#include <Gaudi/Accumulators/Histogram.h>
#include <Gaudi/Accumulators/HistogramArray.h>
#include <Kernel/IUTReadoutTool.h>
#include <algorithm>
#include <optional>
#include <string>
#include <unordered_set>

/** @class UTNZSMonitor UTNZSMonitor.cpp
 * --------------------------------------------------
 *  Counters for Data Online Monitoring : UTNZS banks
 *    @author Wojciech Krupa (wokrupa@cern.ch):
 *    @date   2024/02/10 / 2025/02
 * --------------------------------------------------
 */
namespace GA = Gaudi::Accumulators;

namespace LHCb::UT {

  struct GeomCache {
    std::unordered_set<std::string> m_map; // Outer abs(x) edge of station
    GeomCache( const LHCb::Detector::UT::DeUT& det ) {
      // Loop over all valid indices from DeUTObject
      for ( unsigned int side = 0; side < LHCb::Detector::UT::detail::DeUTObject::NBSIDE; ++side ) {
        for ( unsigned int layer = 0; layer < LHCb::Detector::UT::detail::DeUTObject::NBHALFLAYER; ++layer ) {
          for ( unsigned int stave = 0; stave < LHCb::Detector::UT::detail::DeUTObject::NBSTAVE; ++stave ) {
            for ( unsigned int face = 0; face < LHCb::Detector::UT::detail::DeUTObject::NBFACE; ++face ) {
              for ( unsigned int module = 0; module < LHCb::Detector::UT::detail::DeUTObject::NBMODULE; ++module ) {
                for ( unsigned int sec = 0; sec < LHCb::Detector::UT::detail::DeUTObject::NBSUBSECTOR; ++sec ) {
                  auto sector = det.getSector( static_cast<LHCb::Detector::UT::DeUT::Side_t>( side ),
                                               static_cast<LHCb::Detector::UT::DeUT::Layer_t>( layer ),
                                               static_cast<LHCb::Detector::UT::DeUT::Stave_t>( stave ),
                                               static_cast<LHCb::Detector::UT::DeUT::Face_t>( face ),
                                               static_cast<LHCb::Detector::UT::DeUT::Module_t>( module ),
                                               static_cast<LHCb::Detector::UT::DeUT::Sector_t>( sec ) );
                  // Store the hardware name
                  m_map.insert( sector.getHardwareName() );
                }
              }
            }
          }
        }
      }
    }
  };

  // Definition of axis
  const GA::Axis<double> AxisADC   = GA::Axis<double>( 64, -32.5, 31.5, "nADC" ); //-32, 31
  const GA::Axis<double> AxisCalib = GA::Axis<double>( 257, -0.5, 256.5, "CalibStep" );
  const GA::Axis<double> AxisBCID  = GA::Axis<double>( 3565, -0.5, 3564.5, "BunchID" );

  const std::array<std::string, 4> s_HistNames = { "UTA_Tell_40_vs_ASIC_Flow0", "UTA_Tell_40_vs_ASIC_Flow1",
                                                   "UTC_Tell_40_vs_ASIC_Flow0", "UTC_Tell_40_vs_ASIC_Flow1" };

  const std::array<std::string, 4> s_HistBanksNames = { "UTA_BankType_vs_TELL40_Flow0", "UTA_BankType_vs_TELL40_Flow1",
                                                        "UTC_BankType_vs_TELL40_Flow0",
                                                        "UTC_BankType_vs_TELL40_Flow1" };

  /** ------------------------------------------------------------------------------------------------------------------------ */
  class UTNZSMonitor
      : public Algorithm::Consumer<void( UTDigits const&, UTDigits const&, DeUTDetector const&, ODIN const& ),
                                   Algorithm::Traits::usesConditions<DeUTDetector, GeomCache>> {
  public:
    UTNZSMonitor( const std::string& name, ISvcLocator* svcloc )
        : Consumer{ name,
                    svcloc,
                    { { "InputData", UTDigitLocation::UTDigits },
                      { "InputErrData", UTDigitLocation::UTDigits },
                      { "UTLocation", DeUTDetLocation::location() },
                      { "ODINLocation", ODINLocation::Default } } } {}
    ToolHandle<IUTReadoutTool> readoutTool{ this, "ReadoutTool", "UTReadoutTool" };

    std::once_flag m_alreadyInitialized;

    StatusCode initialize() override;
    void       operator()( const UTDigits&, const UTDigits&, DeUTDetector const&, ODIN const& ) const override;
    void       fillADC( const UTDigit*, ODIN const&, DeUTDetector const& det ) const;
    void       fillFlow( const UTDigit*, DeUTDetector const& ) const;
    void       incrementHistogram( GA::ProfileHistogram<2>&, double, double, double, double, float ) const;
    void       fillHeatMap_naturalunits( const UTDigit*, DeUTDetector const& ) const;

  private:
    UTCoordinatesMap UTMap;

    // TELL40 plots
    mutable GA::HistogramArray<GA::Histogram<2>, s_HistNames.size()> m_UT_Tell_40_vs_asic{
        this,
        []( int i ) { return fmt::format( "{}", s_HistNames[i] ); },
        []( int i ) { return fmt::format( "{}", s_HistNames[i] ); },
        { 54, 0, 54 },
        { 24, 0, 24 } };

    // The container for Mean_ADC per channel

    mutable std::map<std::string, GA::Histogram<2>> m_2d_ADCCounter;

    // 2D plots with MCMS value per chip
    mutable GA::HistogramArray<GA::ProfileHistogram<2>, 4> m_UT_mcm_strip{
        this,
        []( int i ) { return fmt::format( "MCM_strip_Layer{}", i ); },
        []( int i ) { return fmt::format( "MCM_strip_Layer{} (ASICs)", i ); },
        { 144, -9, 9, "Staves (ASICs)" },
        { 28, -7, 7, "Modules" } };

    // 2D plots with MCMS strip per chip
    mutable GA::HistogramArray<GA::ProfileHistogram<2>, 4> m_UT_mcm_mu{
        this,
        []( int i ) { return fmt::format( "MCM_mu_Layer{}", i ); },
        []( int i ) { return fmt::format( "MCM_mu_Layer{} (ASICs)", i ); },
        { 144, -9, 9, "Staves (ASICs)" },
        { 28, -7, 7, "Modules" } };

    // 1D plots: ADC
    mutable GA::Histogram<2> m_nADCvsBCID{ this, "ADCvsBuncId", "ADC vs BuncId", AxisBCID, AxisADC };
    mutable GA::Histogram<2> m_nADCvsCalibStep{ this, "ADCvsCalibStep", "ADC vs CalibStep", AxisCalib, AxisADC };

    mutable std::map<std::string, GA::SigmaCounter<>> RawADC;
    mutable std::map<std::string, GA::SigmaCounter<>> mcm;

  }; // End of class UTNZSMonitor
  DECLARE_COMPONENT( UTNZSMonitor )

  StatusCode UTNZSMonitor::initialize() {
    return Consumer::initialize().andThen( [&] {
      // Add condition callback to initialize the histograms on the first event.
      addConditionDerivation( { DeUTDetLocation::location() }, "GeomCache", [this]( const DeUTDetector& det ) {
        const GeomCache geometryinfo( det );
        std::call_once( this->m_alreadyInitialized,
                        [this, &geometryinfo] { // I hope the 'this' here refers to the captured
                                                // 'this', and not the 'this' of the synthesized
                                                // object corresponding to the outer lambda....
                          for ( const std::string& module : geometryinfo.m_map ) {
                            Utility::map_emplace( m_2d_ADCCounter, "ADC_" + module, this, "ADC_" + module,
                                                  { 512, 0, 512 }, { 64, -32.5, 31.5 } );
                          }
                          info() << "Histograms booked." << endmsg;
                        } );
        return geometryinfo;
      } );
    } );
  }

  void UTNZSMonitor::operator()( const UTDigits& digitsCont, const UTDigits& errdigitsCont, DeUTDetector const& det,
                                 ODIN const& odin ) const {

    // General counters
    // -------------------------------------------------------
    // Loop on digits
    // side==1 is Aside (left), side==0 is Cside (right)
    // -------------------------------------------------------
    for ( const UTDigit* d : digitsCont ) {
      fillFlow( d, det );
      fillADC( d, odin, det );
      fillHeatMap_naturalunits( d, det );
    }

    for ( const UTDigit* d : errdigitsCont ) {
      fillFlow( d, det );
      fillADC( d, odin, det );
      fillHeatMap_naturalunits( d, det );
    }
    // -------------------------------------------------------
    // end loop on digits
    // -------------------------------------------------------
  } // end execute

  // Fill ADC plots
  void UTNZSMonitor::fillADC( const UTDigit* aDigit, ODIN const& odin, DeUTDetector const& det ) const {

    // Summary plots
    ++m_nADCvsBCID[{ odin.bunchId(), aDigit->depositedCharge() }];
    ++m_nADCvsCalibStep[{ odin.calibrationStep(), aDigit->depositedCharge() }];

    // ADC counter
    Detector::UT::ChannelID channelID   = aDigit->channelID();
    std::string             module_name = det.findSector( channelID ).getHardwareName();

    RawADC[UT_layers[aDigit->layer()] + "_" + module_name + "_" + std::to_string( aDigit->strip() )] +=
        aDigit->depositedCharge();

    ++m_2d_ADCCounter.at( "ADC_" + module_name )[{ aDigit->strip(), aDigit->depositedCharge() }];
  }

  // Fill Pcie40 plots
  void UTNZSMonitor::fillFlow( const UTDigit* aDigit, DeUTDetector const& det ) const {
    Detector::UT::ChannelID channelID = aDigit->channelID();
    UTDAQID                 daqID     = readoutTool->channelIDToDAQID( channelID );
    unsigned int            board_ID  = daqID.board();
    unsigned int            sourceID  = UTMap.getSourceIDfromBoardNumber( board_ID );
    int                     SourceID_ = sourceID / 1000;

    unsigned int asic_times_lane = aDigit->asic() % 4 + 4 * extract<masks::lane>( aDigit->getdaqID() );

    // Get TELL40 and flavour only once

    int     tell40Bin = UTMap.getTELL40Bin( sourceID );
    Flavour flavour   = UTMap.stringToFlavour( det.findSector( channelID ).getFlavour() );

    // Check if the strip is aligned for filling
    if ( aDigit->strip() % 128 == 0 ) {
      // Handle UA sourceID_ == 10 and 11
      if ( SourceID_ == 10 ) {
        if ( flavour == Flavour::F2x4 ) {
          ++m_UT_Tell_40_vs_asic[0][{ tell40Bin, UTMap.get2_4_Flow0_Bin( asic_times_lane ) }];
        } else if ( flavour == Flavour::F2x5 ) {
          ++m_UT_Tell_40_vs_asic[0][{ tell40Bin, UTMap.get2_5_Flow0_Bin( asic_times_lane ) }];
        } else if ( flavour == Flavour::F4x3 || flavour == Flavour::F2x3 || flavour == Flavour::FJx3 ) {
          ++m_UT_Tell_40_vs_asic[0][{ tell40Bin, asic_times_lane }];
        }
      } else if ( SourceID_ == 11 ) {
        if ( flavour == Flavour::F2x4 ) {
          if ( asic_times_lane == 16 || asic_times_lane == 17 ) {
            ++m_UT_Tell_40_vs_asic[0][{ tell40Bin, UTMap.get2_4_Flow0_Bin( asic_times_lane ) }];
          } else {
            ++m_UT_Tell_40_vs_asic[1][{ tell40Bin, UTMap.get2_4_Flow1_Bin( asic_times_lane ) }];
          }
        } else if ( flavour == Flavour::F2x5 ) {
          ++m_UT_Tell_40_vs_asic[1][{ tell40Bin, UTMap.get2_5_Flow1_Bin( asic_times_lane ) }];
        } else if ( flavour == Flavour::F4x3 || flavour == Flavour::F2x3 || flavour == Flavour::FJx3 ) {
          ++m_UT_Tell_40_vs_asic[1][{ tell40Bin, asic_times_lane }];
        }
      }
      // Handle UC sourceID_ == 12 and 13
      if ( SourceID_ == 12 ) {
        if ( flavour == Flavour::F2x4 ) {
          ++m_UT_Tell_40_vs_asic[2][{ tell40Bin, UTMap.get2_4_Flow0_Bin( asic_times_lane ) }];
        } else if ( flavour == Flavour::F2x5 ) {
          ++m_UT_Tell_40_vs_asic[2][{ tell40Bin, UTMap.get2_5_Flow0_Bin( asic_times_lane ) }];
        } else if ( flavour == Flavour::F4x3 || flavour == Flavour::F2x3 || flavour == Flavour::FJx3 ) {
          ++m_UT_Tell_40_vs_asic[2][{ tell40Bin, asic_times_lane }];
        }
      } else if ( SourceID_ == 13 ) {
        if ( flavour == Flavour::F2x4 ) {
          if ( asic_times_lane == 16 || asic_times_lane == 17 ) {
            ++m_UT_Tell_40_vs_asic[2][{ tell40Bin, UTMap.get2_4_Flow0_Bin( asic_times_lane ) }];
          } else {
            ++m_UT_Tell_40_vs_asic[3][{ tell40Bin, UTMap.get2_4_Flow1_Bin( asic_times_lane ) }];
          }
        } else if ( flavour == Flavour::F2x5 ) {
          ++m_UT_Tell_40_vs_asic[3][{ tell40Bin, UTMap.get2_5_Flow1_Bin( asic_times_lane ) }];
        } else if ( flavour == Flavour::F4x3 || flavour == Flavour::F2x3 || flavour == Flavour::FJx3 ) {
          ++m_UT_Tell_40_vs_asic[3][{ tell40Bin, asic_times_lane }];
        }
      }
    }
  }

  // Increment 2D plots
  void UTNZSMonitor::incrementHistogram( GA::ProfileHistogram<2>& histogram, double x, double y,
                                         double positionasic_one, double positionasic_two, float val ) const {

    histogram[{ x + positionasic_one, y + 0.25 }] += val;
    histogram[{ x + positionasic_two, y + 0.25 }] += val;
    histogram[{ x + positionasic_one, y - 0.25 }] += val;
    histogram[{ x + positionasic_two, y - 0.25 }] += val;
  }

  // Fill heatmaps (2D) plots
  void UTNZSMonitor::fillHeatMap_naturalunits( const UTDigit* aDigit, DeUTDetector const& det ) const {

    if ( aDigit->strip() % 128 == 0 ) {

      // Let's find where we are
      Detector::UT::ChannelID channelID   = aDigit->channelID();
      std::array<double, 4>   Coordinates = det.findSector( channelID ).getMonitoringCoordinates();
      std::string             module_name = det.findSector( channelID ).getHardwareName();

      float x       = Coordinates[0];
      float y       = Coordinates[1];
      auto  chanID  = aDigit->channelID();
      auto  aSector = det.findSector( chanID );

      auto   trajStrip0   = aSector.createTraj( 0, 0 );
      auto   trajStrip511 = aSector.createTraj( 511, 0 );
      double stripx0      = -1 * ( trajStrip0.endPoint().x() + trajStrip0.beginPoint().x() ) * 0.5;
      double stripx511    = -1 * ( trajStrip511.endPoint().x() + trajStrip511.beginPoint().x() ) * 0.5;
      double stripwidth   = stripx511 - stripx0;
      int    chip;

      if ( stripwidth > 0 )
        chip = int( aDigit->strip() / 128 );
      else
        chip = 3 - int( aDigit->strip() / 128 );

      double positionasic_one = positionasic[chip][0];
      double positionasic_two = positionasic[chip][1];
      double positionasic_thr = positionasic[chip][2];

      // Sector map //
      if ( ( aDigit->stave() > 1 ) || ( fabs( y ) > 2 ) ) {
        incrementHistogram( m_UT_mcm_strip[aDigit->layer()], x, y, positionasic_one, positionasic_two,
                            aDigit->mcmStrip() );
        incrementHistogram( m_UT_mcm_mu[aDigit->layer()], x, y, positionasic_one, positionasic_two, aDigit->mcmVal() );
      }

      if ( ( ( aDigit->stave() == 1 ) && ( fabs( y ) < 2 ) ) || ( ( aDigit->stave() == 0 ) && ( fabs( y ) == 1.5 ) ) ) {
        m_UT_mcm_strip[aDigit->layer()][{ x + positionasic_thr, y + 0.25 }] += aDigit->mcmStrip();
        m_UT_mcm_strip[aDigit->layer()][{ x + positionasic_thr, y - 0.25 }] += aDigit->mcmStrip();
        m_UT_mcm_mu[aDigit->layer()][{ x + positionasic_thr, y + 0.25 }] += aDigit->mcmVal();
        m_UT_mcm_mu[aDigit->layer()][{ x + positionasic_thr, y - 0.25 }] += aDigit->mcmVal();
      }

      if ( ( aDigit->stave() == 0 ) && ( fabs( y ) < 1 ) ) {
        { m_UT_mcm_strip[aDigit->layer()][{ x + positionasic_thr, y }] += aDigit->mcmStrip(); }
        { m_UT_mcm_mu[aDigit->layer()][{ x + positionasic_thr, y }] += aDigit->mcmVal(); }
      }
    }
  }
} // namespace LHCb::UT
