/*****************************************************************************\
* (c) Copyright 2020 CERN for the benefit of the LHCb Collaboration           *
*                                                                             *
* This software is distributed under the terms of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/

#include "Event/ODIN.h"
#include "Event/TAEUtils.h"
#include "Gaudi/Accumulators/Histogram.h"
#include "GaudiKernel/GaudiException.h"
#include "GaudiKernel/PhysicalConstants.h"
#include "GaudiKernel/SystemOfUnits.h"
#include "Kernel/IUTReadoutTool.h"
#include "Kernel/STLExtensions.h"
#include "Kernel/UTDAQID.h"
#include "LHCbAlgs/MergingTransformer.h"
#include "UTDAQ/UTCoordinatesMap.h"
#include "UTDet/DeUTDetector.h"

/** @class UTSuperTAEMonitor UTSuperTAEMonitor.cpp
 * --------------------------------------------------
 * Counters for Data Online Monitoring : Super-TAE (UT in global)
 *  @author Wojciech Krupa (wokrupa@cern.ch) based on code by:
 *  @date   2024/10/07 / 2025/02
 * --------------------------------------------------
 **/

using ODIN        = LHCb::ODIN;
using ODINVector  = Gaudi::Functional::vector_of_const_<ODIN const*>;
using InputVector = Gaudi::Functional::vector_of_const_<LHCb::UTDigits const*>;
using namespace LHCb;
namespace GA = Gaudi::Accumulators;

namespace LHCb::UT {

  class UTSuperTAEMonitor final
      : public LHCb::Algorithm::MergingConsumer<void( ODINVector const&, InputVector const& )> {

  public:
    UTSuperTAEMonitor( const std::string&, ISvcLocator* );
    StatusCode initialize() override;
    void       operator()( ODINVector const&, InputVector const& ) const override;
    void       fillHistograms( const LHCb::UTDigit*, LHCb::ODIN const&, int ) const;

  private:
    LHCb::TAE::Handler m_taeHandler{ this };
    UTCoordinatesMap   UTMap;
    using Axis = GA::Axis<float>;

    unsigned int m_HalfWindow = 3;

    // Definition of axis
    const GA::Axis<double> AxisADC        = GA::Axis<double>( 64, -32.5, 31.5, "nADC" );       //-32-31
    const GA::Axis<double> AxisADC_signed = GA::Axis<double>( 32, -0.5, 31.5, "nADC" );        // 0-31
    const GA::Axis<double> AxisBCID       = GA::Axis<double>( 3565, -0.5, 3564.5, "BunchID" ); //
    const GA::Axis<double> AxisTAE =
        GA::Axis<double>( 2 * m_HalfWindow + 1, -1 * m_HalfWindow - 0.5, m_HalfWindow + 0.5, "TAE index" );

    // General TAE counters
    mutable GA::Histogram<1> m_bxtype{ this, "BXType", "BXType", { 4, -0.5, 3.5 } };
    mutable GA::Histogram<1> m_trgtype{ this, "TrgType", "TrgType", { 16, -0.5, 15.5 } };
    mutable GA::Histogram<2> m_bcid_bxtype{ this, "BXTypeVsBXID", "BXTypeVsBXID", AxisBCID, { 4, -0.5, 3.5 } };
    mutable GA::Histogram<2> m_bcid_trgtype{ this, "TrgTypeVsBXID", "TrgTypeVsBXID", AxisBCID, { 16, -0.5, 15.5 } };
    mutable GA::WeightedHistogram<2> m_bcid_taesummary{
        this, "TAESummary", "TAE summary", AxisBCID, { 14, -0.5, -0.5 + 14 } };
    mutable GA::Histogram<2> m_bcid_taeindex{ this, "TAEIndex", "TAE index", AxisBCID, { 64, -0.5, -0.5 + 64 } };

    // Counters TAE step vs ADC per DCB or per quadrant or UT
    mutable std::map<std::string, GA::Histogram<2>> m_2d_DCB_tae_adc;
    mutable std::map<std::string, GA::Histogram<2>> m_2d_DCBQuadrant_tae_adc;
    mutable GA::Histogram<2>                        m_2d_UT_tae_adc{ this, "UTTAE", "UTTAE", AxisTAE, AxisADC_signed };
  };

  // Declaration of the Algorithm Factory
  DECLARE_COMPONENT( UTSuperTAEMonitor )

  UTSuperTAEMonitor::UTSuperTAEMonitor( const std::string& name, ISvcLocator* pSvcLocator )
      : LHCb::Algorithm::MergingConsumer<void( ODINVector const&, InputVector const& )>(
            name, pSvcLocator, { KeyValues{ "ODINVector", {} }, KeyValues{ "InputVector", {} } } ){};

  StatusCode UTSuperTAEMonitor::initialize() {
    return base_class::initialize().andThen( [&] {
      // Set the top directory to UT
      // Initialise counters
      for ( const auto& DCB : UTMap.getDCBsAside() ) {
        Utility::map_emplace( m_2d_DCB_tae_adc, toString( DCB ), this, toString( DCB ), AxisTAE, AxisADC_signed );
      }
      for ( const auto& DCB : UTMap.getDCBsCside() ) {
        Utility::map_emplace( m_2d_DCB_tae_adc, toString( DCB ), this, toString( DCB ), AxisTAE, AxisADC_signed );
      }
      for ( const auto& DCBQuadrant : UT_DCBQuadrants ) {
        Utility::map_emplace( m_2d_DCBQuadrant_tae_adc, DCBQuadrant, this, DCBQuadrant, AxisTAE, AxisADC_signed );
      }
    } );
  }

  void UTSuperTAEMonitor::operator()( ODINVector const& odinVector, InputVector const& inputVector ) const {

    auto taeEvents = m_taeHandler.arrangeTAE( odinVector, inputVector, 3 );

    if ( taeEvents.empty() ) { return; }

    for ( auto const& [offset, data] : taeEvents ) {
      // offset 0 is central, negative for Prev, positive for Next
      const auto& [odin, digitsCont] = data;

      for ( auto const& d : digitsCont ) { fillHistograms( d, odin, offset ); }
    }
  }

  // Fill histograms
  void UTSuperTAEMonitor::fillHistograms( const LHCb::UTDigit* aDigit, LHCb::ODIN const& odin, int offset ) const {

    // Let's find where we are
    Detector::UT::ChannelID channelID = aDigit->channelID();
    std::string             dcb_name  = UTMap.getDCB( channelID / 512 );
    int                     TAEindex  = offset; //

    // Let's find where we are in the LHCb
    auto bcid    = odin.bunchId();
    auto bx      = static_cast<uint8_t>( odin.bunchCrossingType() );
    auto trgtype = odin.triggerType();

    // simple counters
    ++m_bxtype[bx];
    ++m_trgtype[trgtype];
    ++m_bcid_bxtype[{ bcid, bx }];
    ++m_bcid_trgtype[{ bcid, trgtype }];

    // summary plots this should be replaced by odin.eventType soon as below
    double weight  = 1;
    int    offset2 = 0;
    m_bcid_taesummary[{ bcid, offset2 + bx }] += weight;
    offset2 += 4;
    m_bcid_taesummary[{ bcid, offset2++ }] += weight * odin.isTAE();
    m_bcid_taesummary[{ bcid, offset2++ }] += weight * odin.timeAlignmentEventCentral();
    m_bcid_taesummary[{ bcid, offset2++ }] += weight * odin.timeAlignmentEventFirst();

    ++m_bcid_taeindex[{ bcid, odin.timeAlignmentEventIndex() }];
    ++m_2d_UT_tae_adc[{ TAEindex, aDigit->depositedCharge() }];
    ++m_2d_DCB_tae_adc.at( dcb_name )[{ TAEindex, aDigit->depositedCharge() }];
    ++m_2d_DCBQuadrant_tae_adc.at( dcb_name.substr( 0, 6 ) )[{ TAEindex, aDigit->depositedCharge() }];
  };
}; // namespace LHCb::UT
