/*****************************************************************************\
* (c) Copyright 2000-2018 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/

#include "Event/ODIN.h"
#include "Event/UTDigit.h"
#include "Kernel/IUTReadoutTool.h"
#include "Kernel/STLExtensions.h"
#include "Kernel/UTDAQID.h"
#include "LHCbAlgs/Consumer.h"
#include "UTDAQ/UTCoordinatesMap.h"

/** @class UTTAEMonitor UTTAEMonitor.h
 * --------------------------------------------------
 *  Counters for Data Online Monitoring : TAE (UT in local)
 *  @author Wojciech Krupa (wokrupa@cern.ch) based on code by:
 *  @author C. Hadjivasiliou
 *  @date   2024/03/19 / 2025/02
 * --------------------------------------------------
 **/

namespace GA     = Gaudi::Accumulators;
using EventTypes = LHCb::ODIN::EventTypes;

namespace LHCb::UT {

  class UTTAEMonitor : public Algorithm::Consumer<void( UTDigits const&, UTDigits const&, ODIN const& ),
                                                  Algorithm::Traits::usesConditions<DeUTDetector>> {

  public:
    UTTAEMonitor( const std::string&, ISvcLocator* );
    StatusCode initialize() override;
    void       operator()( const UTDigits&, const UTDigits&, ODIN const& ) const override;
    void       fillHistograms( const UTDigit*, ODIN const& ) const;

    ToolHandle<IUTReadoutTool> readoutTool{ this, "ReadoutTool", "UTReadoutTool" };
    UTCoordinatesMap           UTMap;

    unsigned int m_HalfWindow = 3;

    // Definition of axis
    const GA::Axis<double> AxisADC        = GA::Axis<double>( 64, -32.5, 31.5, "nADC" );       //-32-31
    const GA::Axis<double> AxisADC_signed = GA::Axis<double>( 32, -0.5, 31.5, "nADC" );        // 0-31
    const GA::Axis<double> AxisBCID       = GA::Axis<double>( 3565, -0.5, 3564.5, "BunchID" ); //
    const GA::Axis<double> AxisTAE = GA::Axis<double>( 7, -3.5, 3.5, "TAE index" ); // for windows 3 - central is 4

  private:
    // General counters
    mutable GA::Histogram<1> m_bxtype{ this, "BXType", "BXType", { 4, -0.5, 3.5 } };
    mutable GA::Histogram<1> m_trgtype{ this, "TrgType", "TrgType", { 16, -0.5, 15.5 } };
    mutable GA::Histogram<2> m_bcid_bxtype{ this, "BXTypeVsBXID", "BXTypeVsBXID", AxisBCID, { 4, -0.5, 3.5 } };
    mutable GA::Histogram<2> m_bcid_trgtype{ this, "TrgTypeVsBXID", "TrgTypeVsBXID", AxisBCID, { 16, -0.5, 15.5 } };
    mutable GA::WeightedHistogram<2> m_bcid_taesummary{
        this, "TAESummary", "TAE summary", AxisBCID, { 14, -0.5, -0.5 + 14 } };
    mutable GA::Histogram<2> m_bcid_taeindex{ this, "TAEIndex", "TAE index", AxisBCID, { 64, -0.5, -0.5 + 64 } };

    // Counters TAE step vs ADC per DCB or per quadrant or UT
    mutable std::map<std::string, GA::Histogram<2>> m_2d_DCB_tae_adc;
    mutable std::map<std::string, GA::Histogram<2>> m_2d_DCBQuadrant_tae_adc;
    mutable GA::Histogram<2>                        m_2d_UT_tae_adc{ this, "UTTAE", "UTTAE", AxisTAE, AxisADC_signed };
    mutable GA::Histogram<1> m_1d_UT_tae_adc{ this, "UTTAE_projection", "UTTAE_projection", { 7, -3.5, 3.5 } };
  };
  DECLARE_COMPONENT( UTTAEMonitor )

  UTTAEMonitor::UTTAEMonitor( const std::string& name, ISvcLocator* svcloc )
      : Consumer{ name,
                  svcloc,
                  { { "InputData", UTDigitLocation::UTDigits },
                    { "InputErrorData", UTDigitLocation::UTDigits },
                    { "ODINLocation", ODINLocation::Default } } } {}

  StatusCode UTTAEMonitor::initialize() {
    return Consumer::initialize().andThen( [&] {
      for ( const auto& DCB : UTMap.getDCBsAside() ) {
        Utility::map_emplace( m_2d_DCB_tae_adc, toString( DCB ), this, toString( DCB ), AxisTAE, AxisADC_signed );
      }
      for ( const auto& DCB : UTMap.getDCBsCside() ) {
        Utility::map_emplace( m_2d_DCB_tae_adc, toString( DCB ), this, toString( DCB ), AxisTAE, AxisADC_signed );
      }
      for ( const auto& DCBQuadrant : UT_DCBQuadrants ) {
        Utility::map_emplace( m_2d_DCBQuadrant_tae_adc, DCBQuadrant, this, DCBQuadrant, AxisTAE, AxisADC_signed );
      }
    } );
  }

  void UTTAEMonitor::operator()( const UTDigits& digitsCont, const UTDigits& errdigitsCont, ODIN const& odin ) const {

    // Loop over UTDigits
    for ( const auto& d : digitsCont ) { fillHistograms( d, odin ); }
    // Loop over UTErrDigits
    for ( const auto& d : errdigitsCont ) { fillHistograms( d, odin ); }
  }

  // Fill histograms
  void UTTAEMonitor::fillHistograms( const UTDigit* aDigit, ODIN const& odin ) const {

    // Let's find where we are
    Detector::UT::ChannelID channelID = aDigit->channelID();
    std::string             dcb_name  = UTMap.getDCB( channelID / 512 );
    unsigned int            TAEindex  = odin.timeAlignmentEventIndex();

    // Let's find where we are in the LHCb
    auto bcid    = odin.bunchId();
    auto bx      = static_cast<uint8_t>( odin.bunchCrossingType() );
    auto trgtype = odin.triggerType();

    // simple counters
    ++m_bxtype[bx];
    ++m_trgtype[trgtype];
    ++m_bcid_bxtype[{ bcid, bx }];
    ++m_bcid_trgtype[{ bcid, trgtype }];

    // summary plots this should be replace by odin.eventType soon as below
    double weight = 1;
    int    offset = 0;
    m_bcid_taesummary[{ bcid, offset + bx }] += weight;
    offset += 4;
    m_bcid_taesummary[{ bcid, offset++ }] += weight * odin.isTAE();
    m_bcid_taesummary[{ bcid, offset++ }] += weight * odin.timeAlignmentEventCentral();
    m_bcid_taesummary[{ bcid, offset++ }] += weight * odin.timeAlignmentEventFirst();

    // for ( auto const& et : TAEEventTypes ) { m_bcid_taesummary[{bcid, offset++}] += weight * odin.eventTypeBit( et
    // );}

    ++m_bcid_taeindex[{ bcid, odin.timeAlignmentEventIndex() }];

    if ( TAEindex != 0 ) {
      int offset = TAEindex - m_HalfWindow - 1; // let's alligne that with superTAE
      ++m_2d_UT_tae_adc[{ offset, aDigit->depositedCharge() }];
      ++m_2d_DCB_tae_adc.at( dcb_name )[{ offset, aDigit->depositedCharge() }];
      ++m_1d_UT_tae_adc[offset];
      ++m_2d_DCBQuadrant_tae_adc.at( dcb_name.substr( 0, 6 ) )[{ offset, aDigit->depositedCharge() }];
    }
  };
} // namespace LHCb::UT
