/***************************************************************************** \
* (c) Copyright 2000-2018 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the GNU General Public      *
* Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/

#include "Event/ODIN.h"
#include "Event/RawBank.h"
#include "Event/RawEvent.h"
#include "Event/UTDigit.h"
#include "Kernel/IUTReadoutTool.h"
#include "Kernel/UTDAQBoard.h"
#include "Kernel/UTDAQDefinitions.h"
#include "Kernel/UTDAQID.h"
#include "LHCbAlgs/Consumer.h"
#include "UTDAQ/UTCoordinatesMap.h"
#include "UTDAQ/UTDAQHelper.h"
#include "UTDAQ/UTInfo.h"

// Needed for VELO
#include "Event/VPFullCluster.h"
#include "Kernel/VPConstants.h"

/** @class UTVeloUTCorrelationsMonitor UTVeloUTCorrelationsMonitor.cpp
 * --------------------------------------------------
 *  Counters for Data Online Monitoring : VP-UT correlation monitors
 *  The algorithm produces correlation plots for hits
 *  @author Paolo Gandini
 *  @author Wojciech Krupa
 *  @date  2024/04/20 / 2025/02
 * --------------------------------------------------
 **/

namespace GA = Gaudi::Accumulators;

namespace LHCb::UT {

  // Definition of axis
  const GA::Axis<double> AxisDIGIT     = GA::Axis<double>( 1001, -0.5, 10000.5, "# UT" );
  const GA::Axis<double> AxisVeloDIGIT = GA::Axis<double>( 1001, -0.5, 10000.5, "# Velo" );

  /** ------------------------------------------------------------------------------------------------------------------------ */
  class UTVeloUTCorrelationsMonitor : public Algorithm::Consumer<void( UTDigits const&, UTDigits const&, ODIN const&,
                                                                       std::vector<VPFullCluster> const& ),
                                                                 Algorithm::Traits::usesConditions<DeUTDetector>> {
  public:
    UTVeloUTCorrelationsMonitor( const std::string& name, ISvcLocator* svcloc )
        : Consumer{ name,
                    svcloc,
                    {
                        { "InputData", UTDigitLocation::UTDigits },
                        { "InputErrorData", UTDigitLocation::UTDigits },
                        { "ODINLocation", ODINLocation::Default },
                        { "VeloClusterLocation", VPFullClusterLocation::Default },
                    } } {}
    ToolHandle<IUTReadoutTool> readoutTool{ this, "ReadoutTool", "UTReadoutTool" };

    void operator()( const UTDigits&, const UTDigits&, ODIN const&, const std::vector<VPFullCluster>& ) const override;

  private:
    // General plots
    mutable GA::Histogram<1> m_nUTDigits{ this, "nUTDigits", "Total UT Digits", AxisDIGIT };
    mutable GA::Histogram<1> m_nVeloDigits{ this, "nVeloDigits", "Total Velo Digits", AxisVeloDIGIT };
    mutable GA::Histogram<2> m_nUTnVeloDigits{ this, "nUTnVeloDigits", "nUT vs nVELO", AxisDIGIT, AxisVeloDIGIT };
    mutable GA::Histogram<2> m_nUTnVeloDigits_bb{ this, "nUTnVeloDigits_bb", "nUT vs nVELO - bb", AxisDIGIT,
                                                  AxisVeloDIGIT };
    mutable GA::Histogram<2> m_nUTnVeloDigits_be{ this, "nUTnVeloDigits_be", "nUT vs nVELO - be", AxisDIGIT,
                                                  AxisVeloDIGIT };
    mutable GA::Histogram<2> m_nUTnVeloDigits_eb{ this, "nUTnVeloDigits_eb", "nUT vs nVELO - eb", AxisDIGIT,
                                                  AxisVeloDIGIT };
    mutable GA::Histogram<2> m_nUTnVeloDigits_ee{ this, "nUTnVeloDigits_ee", "nUT vs nVELO - ee", AxisDIGIT,
                                                  AxisVeloDIGIT };

    mutable GA::HistogramArray<GA::Histogram<2>, 4> m_nUTnVeloDigits_bx{
        this,
        []( int i ) { return fmt::format( "nUTnVeloDigits_bx{}", i ); },
        []( int i ) { return fmt::format( "nUTnVeloDigits BXID{}", i ); },
        { 201, -0.5, 400.5, "# UT" },
        { 201, -0.5, 400.5, "# Velo" } };

  }; // End of class UTVeloUTCorrelationsMonitor
  DECLARE_COMPONENT( UTVeloUTCorrelationsMonitor )

  void UTVeloUTCorrelationsMonitor::operator()( const UTDigits& digitsCont, const UTDigits& errdigitsCont,
                                                ODIN const& odin, const std::vector<VPFullCluster>& vpClusters ) const {

    // General counters
    const int bunchCrossingType = (int)odin.bunchCrossingType();
    ++m_nUTDigits[digitsCont.size() + errdigitsCont.size()];
    ++m_nVeloDigits[vpClusters.size()];
    ++m_nUTnVeloDigits[{ digitsCont.size() + errdigitsCont.size(), vpClusters.size() }];
    ++m_nUTnVeloDigits_bx[bunchCrossingType][{ digitsCont.size() + errdigitsCont.size(), vpClusters.size() }];

    if ( odin.bunchCrossingType() == ODIN::BXTypes::BeamCrossing )
      ++m_nUTnVeloDigits_bb[{ digitsCont.size() + errdigitsCont.size(), vpClusters.size() }];
    else if ( odin.bunchCrossingType() == ODIN::BXTypes::Beam1 )
      ++m_nUTnVeloDigits_be[{ digitsCont.size() + errdigitsCont.size(), vpClusters.size() }];
    else if ( odin.bunchCrossingType() == ODIN::BXTypes::Beam2 )
      ++m_nUTnVeloDigits_eb[{ digitsCont.size() + errdigitsCont.size(), vpClusters.size() }];
    else if ( odin.bunchCrossingType() == ODIN::BXTypes::NoBeam )
      ++m_nUTnVeloDigits_ee[{ digitsCont.size() + errdigitsCont.size(), vpClusters.size() }];
  }
} // namespace LHCb::UT
